/*
 * Created on 16.02.2007
 */
package lgmsimulation;

import silvers.Statik;

/**
 * A method to fit a LGCM of one variable for slope variance and covariance of level and slope
 * 
 * @author timo
 */
public class CholeskyFit 
{
    public static double[][] procent;
    public static int procentCount;
    
    public static double[] power;
    public static int powerCount;
    
    public static final double EPS = 0.001;
    
    public static double lastTwoll;
    public static boolean heywoodFlag;
    
    public static double[] choleskyEstimation(double[][] data, double[] initial, double lvar, double err)
    {
        heywoodFlag = false;
        int anzOcc = data[0].length;
        int anzPer = data.length;
        int anzPar = 2;
        double lvarChol = Math.sqrt(lvar);
        
        double[][] structure = new double[anzOcc][anzPar];
        for (int i=0; i<anzOcc; i++) {structure[i][0] = 1; structure[i][1] = (double)i/(double)anzOcc;}
        double[][] structureTrans = Statik.transpose(structure);
        
        double[] pos = new double[anzPar];
        for (int i=0; i<anzPar; i++) pos[i] = initial[i];
        
        double[][] moments = new double[anzOcc][anzOcc]; 
        for (int i=0; i<anzOcc; i++)
            for (int j=i; j<anzOcc; j++)
            {
                moments[i][j] = 0;
                for (int k=0; k<anzPer; k++) moments[i][j] += data[k][i]*data[k][j];
                moments[j][i] = moments[i][j];
            }
        
        double[][] mat = new double[][]{{lvarChol*lvarChol, lvarChol*pos[0]},{lvarChol*pos[0],pos[0]*pos[0]+pos[1]*pos[1]}};
        double[][] sigma = Statik.multiply(structure, Statik.multiply(mat, structureTrans));
        double[][] sigmaA = new double[anzOcc][anzOcc], sigmaB = new double[anzOcc][anzOcc], sigmaAA = new double[anzOcc][anzOcc];
        for (int i=0; i<anzOcc; i++)
            for (int j=0; j<anzOcc; j++)
            {
                sigmaA[i][j] = ((i+j)*lvarChol / (double)anzOcc) + ((2*i*j)*pos[0] / (double)(anzOcc*anzOcc));
                sigmaB[i][j] = ((2*i*j)*pos[1] / (double)(anzOcc*anzOcc));
                sigmaAA[i][j]= ((2*i*j) / (double)(anzOcc*anzOcc));
            } // sigmaBB ist = sigmaAA, und sigmaAB ist berall null.
        
        double[][] sigmaInv = new double[anzOcc][anzOcc], work = new double[anzOcc][anzOcc];
        double[][] sigmaInvSigmaDevA =  new double[anzOcc][anzOcc], sigmaInvSigmaDevB = new double[anzOcc][anzOcc],
            sigmaInvSigmaDevASigmaInv = new double[anzOcc][anzOcc], sigmaInvSigmaDevBSigmaInv = new double[anzOcc][anzOcc],
	    	sigmaInvSigmaDevAA = new double[anzOcc][anzOcc], sigmaInvSigmaDevAASigmaInv = new double[anzOcc][anzOcc],
        	sigmaInvSigmaDevASigmaInvSigmaDevA = new double[anzOcc][anzOcc], sigmaInvSigmaDevASigmaInvSigmaDevASigmaInv = new double[anzOcc][anzOcc],
	    	sigmaInvSigmaDevASigmaInvSigmaDevB = new double[anzOcc][anzOcc], sigmaInvSigmaDevASigmaInvSigmaDevBSigmaInv = new double[anzOcc][anzOcc],
	    	sigmaInvSigmaDevBSigmaInvSigmaDevB = new double[anzOcc][anzOcc], sigmaInvSigmaDevBSigmaInvSigmaDevBSigmaInv = new double[anzOcc][anzOcc];
        
        double val = Double.MAX_VALUE, dist = 1, diff = 1;
        double[] grad = new double[anzPar], work2 = new double[2];
        double[][] jacobi = new double[anzPar][anzPar], jacobiInv = new double[anzPar][anzPar], work22 = new double[anzPar][anzPar];
        
        while ((diff > EPS*EPS) && (dist > EPS)) 
        {
            dist = val;
            mat[0][0] = lvarChol*lvarChol; mat[0][1] = mat[1][0] = lvarChol*pos[0]; mat[1][1] = pos[0]*pos[0]+pos[1]*pos[1];
            for (int i=0; i<anzOcc; i++)
                for (int j=0; j<anzOcc; j++)
                {
                    double zwerg = 0;
                    for (int k=0; k<2; k++)
                        for (int l=0; l<2; l++)
                            zwerg += structure[i][k]*structure[j][l]*mat[k][l];
                    sigma[i][j] = zwerg;
                }
            for (int i=0; i<anzOcc; i++) sigma[i][i] += err;
            Statik.invert(sigma, sigmaInv, work);

            for (int i=0; i<anzOcc; i++)
                for (int j=0; j<anzOcc; j++)
                {
                    sigmaA[i][j] = ((i+j)*lvarChol / (double)anzOcc) + ((2*i*j)*pos[0] / (double)(anzOcc*anzOcc));
                    sigmaB[i][j] = ((2*i*j)*pos[1] / (double)(anzOcc*anzOcc));
                }
            Statik.multiply(sigmaInv, sigmaA, sigmaInvSigmaDevA); Statik.multiply(sigmaInvSigmaDevA, sigmaInv, sigmaInvSigmaDevASigmaInv);
            Statik.multiply(sigmaInv, sigmaB, sigmaInvSigmaDevB); Statik.multiply(sigmaInvSigmaDevB, sigmaInv, sigmaInvSigmaDevBSigmaInv);
            Statik.multiply(sigmaInvSigmaDevASigmaInv, sigmaA, sigmaInvSigmaDevASigmaInvSigmaDevA); 
            Statik.multiply(sigmaInvSigmaDevASigmaInvSigmaDevA, sigmaInv, sigmaInvSigmaDevASigmaInvSigmaDevASigmaInv);
            Statik.multiply(sigmaInvSigmaDevASigmaInv, sigmaB, sigmaInvSigmaDevASigmaInvSigmaDevB); 
            Statik.multiply(sigmaInvSigmaDevASigmaInvSigmaDevB, sigmaInv, sigmaInvSigmaDevASigmaInvSigmaDevBSigmaInv);
            Statik.multiply(sigmaInvSigmaDevBSigmaInv, sigmaB, sigmaInvSigmaDevBSigmaInvSigmaDevB); 
            Statik.multiply(sigmaInvSigmaDevBSigmaInvSigmaDevB, sigmaInv, sigmaInvSigmaDevBSigmaInvSigmaDevBSigmaInv);
            Statik.multiply(sigmaInv, sigmaAA, sigmaInvSigmaDevAA); Statik.multiply(sigmaInvSigmaDevAA, sigmaInv, sigmaInvSigmaDevAASigmaInv);
            
            val = Math.log(Statik.determinantOfPositiveDefiniteMatrix(sigma))*anzPer;
            for (int i=0; i<2; i++) {grad[i] = 0; 
            	for (int j=0; j<2; j++) jacobi[i][j] = 0;
            }
            for (int i=0; i<anzOcc; i++)
            {
                grad[0] += anzPer*sigmaInvSigmaDevA[i][i];
                grad[1] += anzPer*sigmaInvSigmaDevB[i][i];
                jacobi[0][0] += anzPer*(-sigmaInvSigmaDevASigmaInvSigmaDevA[i][i] + sigmaInvSigmaDevAA[i][i]);
                jacobi[0][1] += anzPer*(-sigmaInvSigmaDevASigmaInvSigmaDevB[i][i]);
                jacobi[1][1] += anzPer*(-sigmaInvSigmaDevBSigmaInvSigmaDevB[i][i] + sigmaInvSigmaDevAA[i][i]);
                for (int j=0; j<anzOcc; j++)
                {
                    val += sigmaInv[i][j] * moments[i][j];
                    grad[0] += -sigmaInvSigmaDevASigmaInv[i][j] * moments[i][j];
                    grad[1] += -sigmaInvSigmaDevBSigmaInv[i][j] * moments[i][j];
                    jacobi[0][0] += (2*sigmaInvSigmaDevASigmaInvSigmaDevASigmaInv[i][j] - sigmaInvSigmaDevAASigmaInv[i][j])  * moments[i][j]; 
                    jacobi[0][1] += (2*sigmaInvSigmaDevASigmaInvSigmaDevBSigmaInv[i][j])  									 * moments[i][j]; 
                    jacobi[1][1] += (2*sigmaInvSigmaDevBSigmaInvSigmaDevBSigmaInv[i][j] - sigmaInvSigmaDevAASigmaInv[i][j])  * moments[i][j]; 
                }
                
            }
            dist = dist - val;
            if (dist > 0)
            {
	            jacobi[1][0] = jacobi[0][1];
	            Statik.invert(jacobi, jacobiInv, work22);
	            work2[0] = - jacobiInv[0][0]*grad[0] - jacobiInv[0][1]*grad[1];
	            work2[1] = - jacobiInv[0][1]*grad[0] - jacobiInv[1][1]*grad[1];
	            diff = work2[0]*work2[0] + work2[1]*work2[1];
//	            System.out.println("Position a = "+pos[0]+", b = "+pos[1]+", diff = "+diff+", value = "+val);
	            pos[0] += work2[0]; pos[1] += work2[1];
            } else {
//	            System.out.println("Position a = "+pos[0]+", b = "+pos[1]+", diff = "+diff+", value = "+val+", breaking because of increase!");
                pos[0] -= work2[0]; pos[1] -= work2[1];
                heywoodFlag = true;
            }
        }
        
        return pos;
    }
    
    public static void simulate(int runs, double lvar, double err, int anzPer, int anzOcc)
    {
        SimpleLGM lgm = new SimpleLGM();
        double cov = 0, svar = 0;
        double[][] structure = new double[anzOcc][2];
        for (int i=0; i<anzOcc; i++) {structure[i][0] = 1; structure[i][1] = (double)i/(double)anzOcc;}
        int[] errPar = new int[anzOcc]; for (int i=0; i<anzOcc; i++) errPar[i] = -1;
        double[] errVal = new double[anzOcc]; for (int i=0; i<anzOcc; i++) errVal[i] = err;
        LinearModel model = new LinearModel(
                structure,
                new int[][]{{-1,0},{0,1}},
                new double[][]{{lvar,cov},{cov,svar}},
                new int[]{-1,-1},
                new double[]{0,0},
                errPar,
                errVal
                );
        double[] val = new double[]{cov,svar};
        int usual = 0, varNegative = 0, detNegative = 0;
        double[] distNoAction = new double[runs], distSetVarianceToZero = new double[runs], distOptimizeCholesky = new double[runs], 
        	distSetToOriginIfNPD = new double[runs], distSetToOriginIfNeg = new double[runs]; 
        int heywoodcount = 0;
        double[][] usualErg = new double[100][2];
        double[][] choleskyErg = new double[100][2];
        
        for (int i=0; i<runs; i++)
        {
            val[0] = cov; val[1] = svar;
            model.createData(anzPer);
            double base = lgm.twollEvaluate(model, val);
	        val = lgm.twollAproxLinearModel(model, val);
	        if (i < 100) {usualErg[i][0] = val[0]; usualErg[i][1] = val[1];}
	        double det = lvar*val[1]-val[0]*val[0];

	        distNoAction[i] = base - lgm.twollEvaluate(model, val);
	        if (val[1] < 0) 
	        {
	            varNegative++;
	            val[1] = 0;
	            distSetVarianceToZero[i] = base - lgm.twollEvaluate(model, val);
	            distSetToOriginIfNeg[i] = 0;
	        } else 
	        {
	            distSetToOriginIfNeg[i] = distNoAction[i];
	            distSetVarianceToZero[i] = distNoAction[i];
	            if (det < 0)
	            {
	                detNegative++;
	                distSetToOriginIfNPD[i] = 0; 
	            } else {distSetToOriginIfNPD[i] = distNoAction[i]; usual++;}
	        }
	        
//	        System.out.println("Erg usual: Var = "+val[1]+", Cov = "+val[0]+" (Det = "+det+")");
	        if (det > 0) {val[0] = val[0] / Math.sqrt(lvar); val[1] = Math.sqrt(val[1] - val[0]*val[0]);}
	        else {val[0] = 1; val[1]=1;}
	        
	        val = CholeskyFit.choleskyEstimation(model.data, val, lvar, err);
            val[1] = val[1]*val[1]+val[0]*val[0]; val[0] = val[0] * Math.sqrt(lvar);
	        if (i < 100) {choleskyErg[i][0] = val[0]; choleskyErg[i][1] = val[1];}
	        distOptimizeCholesky[i] = base - lgm.twollEvaluate(model, val);
            if ((heywoodFlag) || (distOptimizeCholesky[i]<0))
            {
//                System.out.println("Heywood / Inoptimal Estimation: "+distOptimizeCholesky[i]);
//    	        System.out.println("Erg cholesky: Var = "+val[1]+", Cov = "+val[0]+" (Det = "+
//                (lvar*val[1]-val[0]*val[0])+" compared to "+det+"), improvment = "+(distOptimizeCholesky[i]));
    	        distOptimizeCholesky[i] = -999;
    	        heywoodcount++;
            }
        }
//        System.out.println("Usual = "+(usual/(double)runs));
//        System.out.println("Negative Variance = "+(varNegative/(double)runs));
//        System.out.println("Not positive definite = "+(detNegative/(double)runs));

        procent[procentCount][0] = (varNegative/(double)runs);
        procent[procentCount][1] = ((varNegative+detNegative)/(double)runs);
        
        /*
        int maplen = 240; int negOffset = 40;
        double[][] distributionMap = new double[maplen][5];
        System.out.println("Heywoods = "+heywoodcount);
        heywoodcount = 0;
        for (int i=0; i<runs; i++)
        {
            int ix = negOffset + (int)Math.round(distNoAction[i]*20); distributionMap[(ix>=maplen?maplen-1:(ix<0?ix=0:ix))][0]++; 
            	ix = negOffset + (int)Math.round(distSetVarianceToZero[i]*20); distributionMap[(ix>=maplen?maplen-1:(ix<0?ix=0:ix))][1]++; 
            	ix = negOffset + (int)Math.round(distSetToOriginIfNPD[i]*20); distributionMap[(ix>=maplen?maplen-1:(ix<0?ix=0:ix))][3]++; 
            	ix = negOffset + (int)Math.round(distSetToOriginIfNeg[i]*20); distributionMap[(ix>=maplen?maplen-1:(ix<0?ix=0:ix))][4]++; 
            if (distOptimizeCholesky[i]!=-999) 
               {ix = negOffset + (int)Math.round(distOptimizeCholesky[i]*20); distributionMap[(ix>=maplen?maplen-1:(ix<0?ix=0:ix))][2]++;}
            else heywoodcount++;
        }
        System.out.println("Heywoods found in Cholesky Optimization: "+heywoodcount);
        for (int j=0; j<5; j++) distributionMap[maplen-1][j] /= (double)(j==2?runs-heywoodcount:runs);
        for (int i=maplen-2; i>=0; i--)
            for (int j=0; j<5; j++) distributionMap[i][j] = distributionMap[i][j] / (double)(j==2?runs-heywoodcount:runs) + distributionMap[i+1][j];
        
        Statik.writeMatrix(distributionMap, "typeoneDistrubitons.txt", '\t');
        Statik.writeMatrix(usualErg,"usualResults.txt", '\t');
        Statik.writeMatrix(choleskyErg,"choleskyResults.txt", '\t');
        */
    }
    
    public static void simulateIfH0off(int runs, double svar, int anzPer, int anzOcc)
    {
        double err = 10;
        double lvar = 100;
        SimpleLGM lgm = new SimpleLGM();
        double cov = 0;
        double[][] structure = new double[anzOcc][2];
        for (int i=0; i<anzOcc; i++) {structure[i][0] = 1; structure[i][1] = (double)i/(double)anzOcc;}
        int[] errPar = new int[anzOcc]; for (int i=0; i<anzOcc; i++) errPar[i] = -1;
        double[] errVal = new double[anzOcc]; for (int i=0; i<anzOcc; i++) errVal[i] = err;
        LinearModel model = new LinearModel(
                structure,
                new int[][]{{-1,0},{0,1}},
                new double[][]{{lvar,cov},{cov,svar}},
                new int[]{-1,-1},
                new double[]{0,0},
                errPar,
                errVal
                );
        double[] val = new double[]{cov,svar};
        int usual = 0, varNegative = 0, detNegative = 0;
        int rejected = 0;
        int heywoodcount = 0;
        
        for (int i=0; i<runs; i++)
        {
            val[0] = cov; val[1] = svar;
            model.createData(anzPer);
            double base = lgm.twollEvaluate(model, val);
	        val = lgm.twollAproxLinearModel(model, val);
	        double det = lvar*val[1]-val[0]*val[0];

	        if (val[1] < 0) 
	        {
	            varNegative++;
	            val[1] = 0;
	        } else 
	        {
	            if (det < 0)
	            {
	                detNegative++;
	            } else usual++;
	        }
	        
	        if (det > 0) {val[0] = val[0] / Math.sqrt(lvar); val[1] = Math.sqrt(val[1] - val[0]*val[0]);}
	        else {val[0] = 1; val[1]=1;}
	        
	        val = CholeskyFit.choleskyEstimation(model.data, val, lvar, err);
            val[1] = val[1]*val[1]+val[0]*val[0]; val[0] = val[0] * Math.sqrt(lvar);
	        double ratio = base - lgm.twollEvaluate(model, val);
            if ((heywoodFlag) || (ratio<0))
    	        heywoodcount++;
            else if (ratio > 5.991) rejected++;
        }
        
        power[powerCount++] = ((double)rejected/(double)(runs-heywoodcount));        
    }

    public static void main(String[] args)
    {
        /*
        procent = new double[20][2]; procentCount = 0;
        for (int i=2; i<10; i++)
        {
            System.out.println("Starting N = "+i*10);
            simulate(10000, 100, 10, i*10, 10);
            procentCount++;
        }
        for (int i=1; i<13; i++)
        {
            System.out.println("Starting N = "+i*100);
            simulate(100000, 100, 10, i*100, 10);
            procentCount++;
        }
        Statik.writeMatrix(procent,"out.txt",'\t');
        */
        powerCount = 0;
        power = new double[20];
        for (int i=2; i<10; i++)
        {
            System.out.println("Starting N = "+i*10);
            simulateIfH0off(100000, 1, i*10, 10);
        }
        for (int i=1; i<13; i++)
        {
            System.out.println("Starting N = "+i*100);
            simulateIfH0off(100000, 1, i*100, 10);
        }
        Statik.writeMatrix(power,"out.txt");
        
    }
}
