/*
 * Created on 26.04.2006
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package lgmsimulation;
import java.util.Arrays;
import java.util.Random;

import lgmsimulation.SingletonModel.parameterType;

import silvers.Statik;


/**
 * @author timo
 * 
 * Operates like a Singleton Model, but optimizing cholesky parameter. Variances can be fixed to zero by fixing all 
 * cholesky parameters in the same row and column of the cholesky matrix. The i,j - th covariance is the sum of the
 * i-th and the j-th row of the Cholesky matrix. 
 * 
 */
public class CholeskyModel extends Model
{
    public static enum parameterType {mean, structure, facChol, err};
    
    int anzFac;
  
    double[] meanVal;
    int[] meanPar;
    double[][] structureVal;
    int[][] structurePar;
    double[][] facCholVal;
    int[][] facCholPar;
    double errVal;
    int errPar;
  
    double[][] facCov;
  
    // gives the position of the parameter (row, column); second is unused for mean, both unused for error, 1st <= 2nd for facCov
    int[][] parPos;
  
    // parameter types
    parameterType[] parType;
  
    // Some working variables
    double[][] facCovWork, sigmaWork, sigmaWork2, sigmaWork3, sigmaWork4, structureWork, structureTransWork;
    double[] muWork;
    int[] sDevRow, sDevCol, cDevRow, cDevCol, mDev;
  
    public CholeskyModel(int[] meanPar, double[] meanVal, int[][] structurePar, double[][] structureVal, 
              int[][] facCholPar, double[][] facCholVal, int errPar, double errVal)
    {
      this.meanPar = meanPar; this.meanVal = meanVal;
      this.structurePar = structurePar; this.structureVal = structureVal; 
      this.facCholVal = facCholVal; this.facCholPar = facCholPar;  
      this.errVal = errVal; this.errPar = errPar;
      
      anzFac = meanPar.length;
      anzVar = structurePar.length;
      anzPar = getActiveParameters().length;
      collectParameterTypesAndPos();
      
      anzPer = 0; data = new double[0][];
      
      fisherInformationMatrix = new double[anzPar][anzPar];
      
    }

    public CholeskyModel(CholeskyModel toCopy)
    {
      anzFac = toCopy.anzFac;
      anzVar = toCopy.anzVar;
      anzPar = toCopy.anzPar;
      anzPer = toCopy.anzPer;
    
      meanPar = new int[toCopy.meanPar.length];
      for (int i=0; i<meanPar.length; i++) meanPar[i] = toCopy.meanPar[i];
      meanVal = Statik.copy(toCopy.meanVal);
      structurePar = new int[toCopy.structurePar.length][toCopy.structurePar[0].length];
      for (int i=0; i<structurePar.length; i++) for (int j=0; j<structurePar[i].length; j++) 
          structurePar[i][j] = toCopy.structurePar[i][j];
      structureVal = Statik.copy(toCopy.structureVal);
      facCholPar = new int[toCopy.anzFac][toCopy.anzFac];
      for (int i=0; i<anzFac; i++) for (int j=0; j<=i; j++) facCholPar[i][j] = toCopy.facCholPar[i][j]; 
      facCholVal = Statik.copy(toCopy.facCholVal);
      errVal = toCopy.errVal; errPar = toCopy.errPar;
      collectParameterTypesAndPos();
      
      data = Statik.copy(toCopy.data);
      computeMoments();
      fisherInformationMatrix = new double[anzPar][anzPar];
    }
    
    public CholeskyModel copy() {return new CholeskyModel(this);}
      
    public void collectParameterTypesAndPos()
    {
      int maxPar = maxParNumber();
      parType = new parameterType[maxPar+1];
      parPos = new int[maxPar+1][];
  
      for (int i=0; i<meanPar.length; i++) if (meanPar[i]!=NOPARAMETER) {parType[meanPar[i]] = parameterType.mean; parPos[meanPar[i]] = new int[]{i,0};}
      for (int i=0; i<structurePar.length; i++) for (int j=0; j<structurePar[i].length; j++) 
          if (structurePar[i][j]!=NOPARAMETER) {parType[structurePar[i][j]] = parameterType.structure; parPos[structurePar[i][j]] = new int[]{i,j};}
      for (int i=0; i<anzFac; i++) for (int j=0; j<=i; j++) 
          if (facCholPar[i][j]!=NOPARAMETER) {parType[facCholPar[i][j]] = parameterType.facChol; parPos[facCholPar[i][j]] = new int[]{i,j};}
      if (errPar != NOPARAMETER) {parType[errPar] = parameterType.err; parPos[errPar] = new int[]{0,0};}
    }
      
    public int maxParNumber()
    {
      int erg = NOPARAMETER;
      for (int i=0; i<meanPar.length; i++) erg = Math.max(meanPar[i], erg);
      for (int i=0; i<structurePar.length; i++)
          for (int j=0; j<structurePar[i].length; j++) erg = Math.max(structurePar[i][j], erg);
      for (int i=0; i<anzFac; i++)
          for (int j=0; j<=i; j++) erg = Math.max(facCholPar[i][j], erg);
      erg = Math.max(errPar, erg);
      return erg;
    }
      
    public int[] getActiveParameters()
    {
      int toPar = maxParNumber()+1;
      int[] ergO = new int[toPar];
      for (int i=0; i<meanPar.length; i++)
          if (meanPar[i] != NOPARAMETER) ergO[meanPar[i]]=1;
      for (int i=0; i<structurePar.length; i++)
          for (int j=0; j<structurePar[i].length; j++)
              if (structurePar[i][j] != NOPARAMETER) ergO[structurePar[i][j]]=1;
      for (int i=0; i<anzFac; i++)
          for (int j=0; j<=i; j++)
              if (facCholPar[i][j] != NOPARAMETER) ergO[facCholPar[i][j]]=1;
      if (errPar != NOPARAMETER) ergO[errPar]=1;
      int anz = 0; for (int i=0; i<toPar; i++) if (ergO[i]!=0) anz++;
      int[] erg = new int[anz]; int j=0; 
      for (int i=0; i<toPar; i++) if (ergO[i]!=0) erg[j++] = i;
      Arrays.sort(erg);
      return erg;
    }
      
    public void renumberParameter(int oldNumber, int newNumber)
    {
      for (int i=0; i<meanPar.length; i++)
          if (meanPar[i] == oldNumber) meanPar[i] = newNumber;
      for (int i=0; i<structurePar.length; i++)
          for (int j=0; j<structurePar[i].length; j++)
              if (structurePar[i][j] == oldNumber) structurePar[i][j] = newNumber;
      for (int i=0; i<anzFac; i++)
          for (int j=0; j<=i; j++)
              if (facCholPar[i][j] == oldNumber) facCholPar[i][j] = newNumber;
      if (errPar == oldNumber) errPar = newNumber;
      collectParameterTypesAndPos();
    }
    
    public int getAnzFac() {return anzFac;}

    protected void computeMatrixTimesSigmaDev(int par, double[][] matrix, double[][] erg)
    {
        if (parType[par]==parameterType.mean) 
            for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) erg[i][j] = 0;
        if (parType[par]==parameterType.structure)
        {
            for (int i=0; i<anzVar; i++)
                for (int j=0; j<anzVar; j++)
                {
                    erg[i][j] = 0;
                    if (j==parPos[par][0])
                    {
                        if (matrix != null)
                            for (int l=0; l<anzVar; l++)
                            {
                                double s = 0;
                                for (int m=0; m<anzFac; m++) s += facCov[m][parPos[par][1]] * structureVal[l][m];
                                erg[i][j] += matrix[i][l]*s;
                            } 
                        else {
                            for (int m=0; m<anzFac; m++) erg[i][j] += facCov[m][parPos[par][1]] * structureVal[i][m];
                        }
                    }
                    if ((matrix!=null) || (i==parPos[par][0]))
                    {
                        double s = 0;
                        for (int m=0; m<anzFac; m++) s += facCov[parPos[par][1]][m] * structureVal[j][m];
                        erg[i][j] += (matrix!=null?matrix[i][parPos[par][0]]:1)*s;
                    }
                }
        }
        if (parType[par]==parameterType.facChol)
        {
            if ((sigmaWork4==null) || (sigmaWork4.length!=anzVar) || (sigmaWork4[0].length!=anzVar)) sigmaWork4 = new double[anzVar][anzVar];
            
            for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) 
            {
                sigmaWork4[i][j] = 0;
                for (int k=parPos[par][1]; k<anzFac; k++) 
                    sigmaWork4[i][j] += structureVal[i][parPos[par][0]] * structureVal[j][k] * facCholVal[k][parPos[par][1]];
                
//                sigmaWork4[i][j] = sigmaWork4[j][i] = structureVal[i][parPos[par][0]] * structureVal[j][parPos[par][1]] + 
//                      (parPos[par][0]!=parPos[par][1]?structureVal[j][parPos[par][0]] * structureVal[i][parPos[par][1]]:0);
            }
            for (int i=0; i<anzVar; i++) for (int j=0; j<=i; j++) {sigmaWork4[i][j] += sigmaWork4[j][i]; sigmaWork4[j][i] = sigmaWork4[i][j];}
            if (matrix!=null) Statik.multiply(matrix, sigmaWork4, erg); else Statik.copy(sigmaWork4, erg);            
        }
        if (parType[par]==parameterType.err)
        {
            for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) erg[i][j] = (matrix!=null?sigInv[i][j]:(i==j?1:0));
        }
    }
    
    protected void computeMatrixTimesSigmaDevDev(int par1, int par2, double[][] matrix, double[][] erg)
    {
        for (int i=0; i<anzVar; i++) 
            for (int j=0; j<anzVar; j++) erg[i][j] = 0;
        if ((parType[par1] == parameterType.structure) && (parType[par2] == parameterType.structure))
        {
            if (matrix != null) 
                for (int i=0; i<anzVar; i++)
                {
                    erg[i][parPos[par2][0]] += facCov[parPos[par1][1]][parPos[par2][1]]*matrix[i][parPos[par1][0]]; 
                    erg[i][parPos[par1][0]] += facCov[parPos[par1][1]][parPos[par2][1]]*matrix[i][parPos[par2][0]]; 
                } else {
                    erg[parPos[par1][0]][parPos[par2][0]] += facCov[parPos[par1][1]][parPos[par2][1]]; 
                    erg[parPos[par2][0]][parPos[par1][0]] += facCov[parPos[par1][1]][parPos[par2][1]]; 
                }
        }
        if (((parType[par1]==parameterType.structure) && (parType[par2]==parameterType.facChol)) ||
                ((parType[par2]==parameterType.structure) && (parType[par1]==parameterType.facChol))) 
            {
                if (parType[par1]==parameterType.structure) {int t=par1; par1 = par2; par2 = t;}
                if ((sigmaWork4==null) || (sigmaWork4.length!=anzVar) || (sigmaWork4[0].length!=anzVar)) sigmaWork4 = new double[anzVar][anzVar];
                for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) sigmaWork4[i][j] = 0;
                
                for (int i=0; i<anzVar; i++)
                {
                    if (parPos[par2][1]==parPos[par1][0]) 
                        for (int k=0; k<anzFac; k++) sigmaWork4[parPos[par2][0]][i] += facCholVal[k][parPos[par1][1]]*structureVal[i][k];
                    sigmaWork4[parPos[par2][0]][i] += facCholVal[parPos[par2][1]][parPos[par1][1]]*structureVal[i][parPos[par1][0]];
                    sigmaWork4[i][parPos[par2][0]] += sigmaWork4[parPos[par2][0]][i];
                }
                if (matrix != null) Statik.multiply(sigInv, sigmaWork4, erg);
                else Statik.copy(sigmaWork4, erg);
            }
        if ((parType[par1]==parameterType.facChol) && (parType[par2]==parameterType.facChol))
        {
            if (parPos[par1][1]==parPos[par2][1])
            {
                if ((sigmaWork4==null) || (sigmaWork4.length!=anzVar) || (sigmaWork4[0].length!=anzVar)) sigmaWork4 = new double[anzVar][anzVar];
    
                for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) 
                    sigmaWork4[i][j] = structureVal[i][parPos[par1][0]] * structureVal[j][parPos[par2][0]];

                for (int i=0; i<anzVar; i++) for (int j=0; j<=i; j++) {sigmaWork4[i][j] += sigmaWork4[j][i]; sigmaWork4[j][i] = sigmaWork4[i][j];}
                if (matrix != null) Statik.multiply(sigInv, sigmaWork4, erg);
                else Statik.copy(sigmaWork4, erg);
            }
        }
    }
 
    protected void computeMatrixTimesMuDev(int par, double[][] matrix, double[] erg)
    {
        for (int i=0; i<anzVar; i++) erg[i] = 0;
        if (parType[par]==parameterType.structure)
        {
            double v = meanVal[parPos[par][1]];
            if (matrix==null) erg[parPos[par][0]] = v;
            else for (int i=0; i<anzVar; i++) erg[i] = matrix[i][parPos[par][0]]*v;
        } else if (parType[par]==parameterType.mean) {
            for (int i=0; i<anzVar; i++)
            {
                erg[i] = 0;
                if (matrix==null) erg[i] = structureVal[i][parPos[par][0]];
                else for (int j=0; j<anzVar; j++) erg[i] += matrix[i][j] * structureVal[j][parPos[par][0]];
            }
        }
    }

    public void setParameter(int nr, double value)
    {
        if (parType[nr]==parameterType.mean)
            for (int i=0; i<anzFac; i++)
                if (meanPar[i]==nr) meanVal[i] = value; 
        if (parType[nr]==parameterType.structure)
            for (int i=0; i<anzVar; i++)
                for (int j=0; j<anzFac; j++)
                    if (structurePar[i][j]==nr) structureVal[i][j] = value; 
        if (parType[nr]==parameterType.facChol)
            for (int i=0; i<anzFac; i++)
                for (int j=0; j<=i; j++)
                    if (facCholPar[i][j]==nr) facCholVal[i][j] = value;
        if (parType[nr]==parameterType.err) errVal = value;
    }
    
    public double getParameter(int nr)
    {
        if (parType[nr]==parameterType.mean) return meanVal[parPos[nr][0]];
        if (parType[nr]==parameterType.err) return errVal;
        if (parType[nr]==parameterType.structure) return structureVal[parPos[nr][0]][parPos[nr][1]];
        if (parType[nr]==parameterType.facChol) return facCholVal[parPos[nr][0]][parPos[nr][1]];
        return -999;
    }
    
    public boolean isErrorParameter(int nr) {return parType[nr]==parameterType.err;}
    
    public static CholeskyModel fixParameter(CholeskyModel in, int nr, double value) {CholeskyModel erg = new CholeskyModel(in); erg.fixParameter(nr,value); return erg;}
    public void fixParameter(int nr, double value)
    {
        anzPar--;
        for (int i=0; i<anzFac; i++)
        {
            if (meanPar[i]==nr) { meanVal[i] = value; meanPar[i] = NOPARAMETER;}
            if (meanPar[i] >nr)  meanPar[i]--;
        }
        for (int i=0; i<anzVar; i++)
            for (int j=0; j<anzFac; j++)
            {
                if (structurePar[i][j]==nr) { structureVal[i][j] = value; structurePar[i][j] = -1;}
                if (structurePar[i][j] >nr)  structurePar[i][j]--;
            }
        for (int i=0; i<anzFac; i++)
            for (int j=0; j<=i; j++)
            {
                if (facCholPar[i][j]==nr) { facCholVal[i][j] = value; facCholPar[i][j] = -1;}
                if (facCholPar[i][j] >nr)  facCholPar[i][j]--;
            }
        if (errPar==nr) { errVal = value; errPar = -1;}
        if (errPar >nr)  errPar--;
        collectParameterTypesAndPos();
    }

    public void evaluateMuAndSigma(double[] values)
    {
        if ((sigma==null) || (sigma.length!=anzVar) || (sigma[0].length!=anzVar)) sigma = new double[anzVar][anzVar];
        if ((facCov==null) || (facCov.length!=anzFac) || (facCov[0].length!=anzFac)) facCov = new double[anzFac][anzFac];
        if ((structureWork==null) || (structureWork.length!=anzVar) || (structureWork[0].length!=anzFac)) structureWork = new double[anzVar][anzFac];
        if ((structureTransWork==null) || (structureTransWork.length!=anzFac) || (structureTransWork[0].length!=anzVar)) structureTransWork = new double[anzFac][anzVar];
        if ((mu==null) || (mu.length!=anzVar)) mu = new double[anzVar];
        if (values!=null) setParameter(values);

        for (int i=0; i<anzFac; i++)
            for (int j=i; j<anzFac; j++)
            {
                facCov[i][j] = 0; 
                for (int k=0; k<=i; k++) facCov[i][j] += facCholVal[i][k]*facCholVal[j][k];
                facCov[j][i] = facCov[i][j];
            }
        
        Statik.transpose(structureVal, structureTransWork);
        Statik.multiply(structureVal, facCov, structureWork);
        Statik.multiply(structureWork, structureTransWork, sigma);
        
        Statik.multiply(structureVal, meanVal, mu);

        for (int i=0; i<anzVar; i++) 
            sigma[i][i] += errVal;
    }
    
    public void computeLeastSquaresDerivatives(double[] value, boolean recomputeMuAndSigma)
    {
        if ((sigmaWork==null) || (sigmaWork.length!=anzVar) || (sigmaWork[0].length!=anzVar)) sigmaWork = new double[anzVar][anzVar];
        if ((sigmaWork2==null) || (sigmaWork2.length!=anzVar) || (sigmaWork2[0].length!=anzVar)) sigmaWork2 = new double[anzVar][anzVar];
        if ((sigmaWork3==null) || (sigmaWork3.length!=anzVar) || (sigmaWork3[0].length!=anzVar)) sigmaWork3 = new double[anzVar][anzVar];
        if ((muWork==null) || (muWork.length!=anzVar)) muWork = new double[anzVar];
        if ((lsD==null) || (lsD.length!=anzPar)) lsD = new double[anzPar];
        if ((lsDD==null) || (lsDD.length!=anzVar) || (lsDD[0].length!=anzVar)) lsDD = new double[anzPar][anzPar];
        
        if (value != null) setParameter(value);
        if (recomputeMuAndSigma) evaluateMuAndSigma(value);
        
        ls = 0;
        for (int i=0; i<anzVar; i++)
            for (int j=0; j<anzVar; j++) {sigmaWork[i][j] = sigma[i][j] - dataCov[i][j]; ls += 0.5*sigmaWork[i][j]*sigmaWork[i][j];}
        for (int i=0; i<anzVar; i++) {muWork[i] = mu[i] - dataMean[i]; ls += 0.5*muWork[i]*muWork[i];}
        
        for (int p=0; p<anzPar; p++)
        {
            lsD[p] = 0;
            if (parType[p] == parameterType.mean)
                for (int i=0; i<anzVar; i++) lsD[p] += structureVal[i][parPos[p][0]]*muWork[i];
            if (parType[p] == parameterType.structure) lsD[p] += meanVal[parPos[p][1]]*muWork[parPos[p][0]];
            if ((parType[p] == parameterType.structure) || (parType[p] == parameterType.facChol))
            {
                computeMatrixTimesSigmaDev(p, null, sigmaWork2);
                for (int i=0; i<anzVar; i++)
                    for (int j=0; j<anzVar; j++) 
                        lsD[p] += sigmaWork2[i][j]*sigmaWork[i][j];
            }
            if (parType[p] == parameterType.err)
                for (int i=0; i<anzVar; i++) lsD[p] += sigmaWork[i][i];
            for (int p2l=p; p2l<anzPar; p2l++)
            {
                int p1 = p, p2 = p2l;
                if (((parType[p1]!=parameterType.structure) && (parType[p2]==parameterType.structure)) ||
                    ((parType[p1]!=parameterType.structure) && (parType[p1]!=parameterType.facChol) && (parType[p2]==parameterType.facChol)) ||
                    ((parType[p1]==parameterType.mean) &&       (parType[p2]==parameterType.err))) {p1 = p2l; p2 = p;}
                
                lsDD[p1][p2] = 0;
                if ((parType[p1] == parameterType.mean) && (parType[p2] == parameterType.mean))
                    for (int i=0; i<anzVar; i++)
                        lsDD[p1][p2] += structureVal[i][parPos[p1][0]]*structureVal[i][parPos[p2][0]];
                else if ((parType[p1] == parameterType.structure) && (parType[p2] == parameterType.mean))
                {
                    if(parPos[p2][0]==parPos[p1][1]) lsDD[p1][p2] += muWork[parPos[p1][0]];
                    lsDD[p1][p2] += structureVal[parPos[p1][0]][parPos[p2][0]]*meanVal[parPos[p1][1]];
                }
                else if ((parType[p1] == parameterType.structure) && (parType[p2] == parameterType.err))
                    for (int l=0; l<anzFac; l++) 
                        lsDD[p1][p2] += 2*(structureVal[parPos[p1][0]][l]*facCov[parPos[p1][1]][l]);
                else if ((parType[p1] == parameterType.err) && (parType[p2] == parameterType.err))
                    lsDD[p1][p2] += anzVar;
                else {
                    computeMatrixTimesSigmaDev(p1, null, sigmaWork2);
                    computeMatrixTimesSigmaDev(p2, null, sigmaWork3);
                    for (int i=0; i<anzVar; i++)
                        for (int j=0; j<anzVar; j++) 
                            lsDD[p1][p2] += sigmaWork2[i][j]*sigmaWork3[i][j];
                    computeMatrixTimesSigmaDevDev(p1, p2, null, sigmaWork2);
                    for (int i=0; i<anzVar; i++)
                        for (int j=0; j<anzVar; j++) 
                            lsDD[p1][p2] += sigmaWork[i][j]*sigmaWork2[i][j];
                    
                    if (((parType[p1] == parameterType.structure) && (parType[p2] == parameterType.structure))
                        && (parPos[p1][0] == parPos[p2][0])) lsDD[p1][p2] += meanVal[parPos[p1][1]]*meanVal[parPos[p2][1]];
                }

                lsDD[p2][p1] = lsDD[p1][p2];
            }
        }
    }

    public double[] getFreeBasisStartingValues()
    {
        CholeskyModel copy = copy();
        copy.meanVal[0] = copy.dataMean[0];
        if (copy.meanPar[1]!=NOPARAMETER) copy.meanVal[1] = copy.dataMean[1] - copy.meanVal[0];
        for (int i=0; i<structureVal.length; i++) 
            if (copy.structurePar[i][1] != NOPARAMETER) {
                copy.structureVal[i][1] = (copy.dataMean[i]-copy.dataMean[0])/copy.meanVal[1]; copy.fixParameter(copy.structurePar[i][1]);}
        
        copy().estimateML(0.01);
        double[] erg = new double[anzPar];
        int j=0; 
        for (int i=0; i<anzPar; i++) erg[i] = (parType[i] == parameterType.structure?copy.structureVal[1][parPos[i][1]]:copy.getParameter(j++));
        
        return erg;
    }

    public double[] getArbitraryStartingValues()
    {
        double[] erg = new double[anzPar];
        
        for (int i=0; i<anzPar; i++)
        {
            if (parType[i] == parameterType.mean) erg[i] = 0;
            if (parType[i] == parameterType.facChol) 
                if (parPos[i][0] == parPos[i][1]) erg[i] = 1; else erg[i] = 0;
            if (parType[i] == parameterType.err) erg[i] = 3.0;
            if (parType[i] == parameterType.structure) 
                if (parPos[i][0] < parPos[i][1]) erg[i] = 0.0; else erg[i] = 1.0;
        }
        
        return erg;
    }

    public double[] estimateLSOfLinearModel()
    {
        CholeskyModel copy = new CholeskyModel(this);
        double basefak = copy.structureVal[1][1]-copy.structureVal[0][1];
        for (int i=2; i<copy.anzVar; i++) copy.structureVal[i][1] = copy.structureVal[0][1] + i*basefak;
        for (int i=0; i<copy.anzPar; i++) if (copy.parType[i] == parameterType.structure) {copy.fixParameter(i,copy.parPos[i][0]); i--;}
        double[] zwerg = copy.estimateLS();
        double[] erg = new double[anzPar];
        int i=0; 
        for (int j=0; j<anzPar; j++) 
            if (parType[j]!=parameterType.structure) erg[j] = zwerg[i++]; else erg[j] = parPos[j][0];
        
        return erg;
    }
    
    public double[] estimateMLOfLinearModel()
    {
        CholeskyModel copy = new CholeskyModel(this);
        double basefak = copy.structureVal[1][1]-copy.structureVal[0][1];
        for (int i=2; i<copy.anzVar; i++) copy.structureVal[i][1] = copy.structureVal[0][1] + i*basefak;
        for (int i=0; i<copy.anzPar; i++) if (copy.parType[i] == parameterType.structure) {copy.fixParameter(i); i--;}
        double[] zwerg = copy.estimateML(copy.estimateLS(),0.0001);
        double[] erg = new double[anzPar];
        int i=0; 
        for (int j=0; j<anzPar; j++) 
            if (parType[j]!=parameterType.structure) erg[j] = zwerg[i++]; else erg[j] = copy.structureVal[parPos[j][0]][parPos[j][1]];
        
        return erg;
    }
    
    public void computeLogLikelihoodDerivatives(double[] value, boolean recomputeMuAndSigma)
    {
        
        if ((sigmaWork==null) || (sigmaWork.length!=anzVar) || (sigmaWork[0].length!=anzVar)) sigmaWork = new double[anzVar][anzVar];
        if ((sigmaWork2==null) || (sigmaWork2.length!=anzVar) || (sigmaWork2[0].length!=anzVar)) sigmaWork2 = new double[anzVar][anzVar];
        if ((sigmaWork3==null) || (sigmaWork3.length!=anzVar) || (sigmaWork3[0].length!=anzVar)) sigmaWork3 = new double[anzVar][anzVar];
        if ((sigmaWork4==null) || (sigmaWork4.length!=anzVar) || (sigmaWork4[0].length!=anzVar)) sigmaWork4 = new double[anzVar][anzVar];
        if ((sigInv==null) || (sigInv.length!=anzVar) || (sigInv[0].length!=anzVar)) sigInv = new double[anzVar][anzVar];
        if ((llD==null) || (llD.length!=anzPar)) llD = new double[anzPar];
        if ((llDD==null) || (llDD.length!=anzVar) || (llDD[0].length!=anzVar)) llDD = new double[anzPar][anzPar];
        
        if (value != null) setParameter(value);
        if (recomputeMuAndSigma) evaluateMuAndSigma(value);
        
        sigmaDet = Statik.invert(sigma,sigInv, sigmaWork);
        if ((Statik.determinantOfPositiveDefiniteMatrix(sigma)==-1) && (sigmaDet > 0))
        {
//            System.out.println("Not positive defnite Sigma!");
            sigmaDet = Double.NaN;
        }
        
        ll = anzPer*anzVar*LNTWOPI + anzPer*Math.log(sigmaDet);
        for (int i=0; i<anzVar; i++) 
            for (int j=0; j<anzVar; j++)
                ll += sigInv[i][j]*(xBiSum[i][j] - anzPer*mu[i]*dataMean[j] - anzPer*mu[j]*dataMean[i] + anzPer*mu[i]*mu[j]);
        
        for (int p=0; p<anzPar; p++)
        {
            llD[p] = 0.0;
            if (parType[p] == parameterType.mean)
            {
                for (int i=0; i<anzVar; i++)
                    for (int j=0; j<anzVar; j++) llD[p] -= 2*structureVal[i][parPos[p][0]]*sigInv[i][j]*anzPer*(dataMean[j] - mu[j]);
            } 
            if (parType[p] == parameterType.structure)
            {
                for (int j=0; j<anzVar; j++) llD[p] -= 2*meanVal[parPos[p][1]]*sigInv[parPos[p][0]][j]*anzPer*(dataMean[j] - mu[j]);
            }
            if (parType[p] != parameterType.mean)
            {
                computeSigmaInvDev(p, sigmaWork);
                double trace = 0;
                for (int i=0; i<anzVar; i++) trace += sigmaWork[i][i];
                llD[p] += trace*anzPer;
                Statik.multiply(sigmaWork, sigInv, sigmaWork2);
                for (int i=0; i<anzVar; i++) 
                    for (int j=0; j<anzVar; j++) llD[p] -= sigmaWork2[i][j]*(xBiSum[i][j] - anzPer*mu[i]*dataMean[j] - anzPer*mu[j]*dataMean[i] + anzPer*mu[i]*mu[j]);
            }
        }
        
        for (int p1=0; p1<anzPar; p1++)
            for (int p2=0; p2<anzPar; p2++)
                llDD[p1][p2] = 0;
        for (int p1=0; p1<anzPar; p1++)
            for (int p2=0; p2<anzPar; p2++)
            {
                // Term 4 in Eq. (13) of von Oertzen, Ghisletta, Lindenberger (2009)
                if ((parType[p1]==parameterType.mean) && (parType[p2]==parameterType.mean))
                {
                    for (int i=0; i<anzVar; i++)
                        for (int j=0; j<anzVar; j++)
                            llDD[p1][p2] += 2*anzPer*structureVal[i][parPos[p1][0]]*structureVal[j][parPos[p2][0]]*sigInv[i][j];
                }
                if (((parType[p1]==parameterType.mean) && (parType[p2]==parameterType.structure))
                ||  ((parType[p1]==parameterType.structure) && (parType[p2]==parameterType.mean)))
                {
                    int q1 = (parType[p1]==parameterType.mean?p1:p2), q2 = (parType[p1]==parameterType.mean?p2:p1); 
                    for (int i=0; i<anzVar; i++)
                        llDD[p1][p2] += 2*anzPer*structureVal[i][parPos[q1][0]]*meanVal[parPos[q2][1]]*sigInv[i][parPos[q2][0]];
                }
                if ((parType[p1]==parameterType.structure) && (parType[p2]==parameterType.structure))
                {
                    llDD[p1][p2] += 2*anzPer*meanVal[parPos[p1][1]]*meanVal[parPos[p2][1]]*sigInv[parPos[p1][0]][parPos[p2][0]];
                }
                // Term 2 in Eq. (13) of von Oertzen, Ghisletta, Lindenberger (2009)
                if (((parType[p1]==parameterType.mean) && (parType[p2]==parameterType.structure))
                ||  ((parType[p1]==parameterType.structure) && (parType[p2]==parameterType.mean)))
                {
                    int q1 = (parType[p1]==parameterType.mean?p1:p2), q2 = (parType[p1]==parameterType.mean?p2:p1);
                    
                    if (parPos[q1][0] == parPos[q2][1]) 
                        for (int i=0; i<anzVar; i++) llDD[p1][p2] -= 2*anzPer*(dataMean[i]-mu[i])*sigInv[parPos[q2][0]][i];
                }
                // Term 3 and 5 in Eq. (13) of von Oertzen, Ghisletta, Lindenberger (2009)
                if ((parType[p1]==parameterType.structure) && (parType[p2]==parameterType.structure))
                {
                    computeSigmaInvDev(p1, sigmaWork);
                    Statik.multiply(sigmaWork, sigInv, sigmaWork2);
                    computeSigmaInvDev(p2, sigmaWork);
                    Statik.multiply(sigmaWork, sigInv, sigmaWork3);
                    for (int j=0; j<anzVar; j++)
                    {
                        llDD[p1][p2] += 2*meanVal[parPos[p1][1]]*sigmaWork3[parPos[p1][0]][j]*anzPer*(dataMean[j]-mu[j]);
                        llDD[p1][p2] += 2*meanVal[parPos[p2][1]]*sigmaWork2[parPos[p2][0]][j]*anzPer*(dataMean[j]-mu[j]);
                    }
                } 
                else if (((parType[p1]!=parameterType.facChol) && (parType[p1]!=parameterType.err) && (parType[p2]!=parameterType.mean))
                ||  ((parType[p1]!=parameterType.mean) && (parType[p2]!=parameterType.facChol) && (parType[p2]!=parameterType.err)))
                {
                    boolean exchange = (parType[p1]==parameterType.facChol) || (parType[p1]==parameterType.err)
                                        || (parType[p2]==parameterType.mean);
                    int q1 = (exchange?p2:p1), q2 = (exchange?p1:p2); 
                    computeSigmaInvDev(q2, sigmaWork2);
                    Statik.multiply(sigmaWork2, sigInv, sigmaWork);
                    if (parType[q1]==parameterType.mean)
                    {
                        for (int i=0; i<anzVar; i++)
                            for (int j=0; j<anzVar; j++)
                                llDD[p1][p2] += 2*structureVal[i][parPos[q1][0]]*sigmaWork[i][j]*anzPer*(dataMean[j]-mu[j]);
                    } else {
                        for (int j=0; j<anzVar; j++)
                            llDD[p1][p2] += 2*meanVal[parPos[q1][1]]*sigmaWork[parPos[q1][0]][j]*anzPer*(dataMean[j]-mu[j]);
                    }
                }
                
                // Term 1,6,7 and 8 in Eq. (13) of von Oertzen, Ghisletta, Lindenberger (2009)
                if ((parType[p1]!=parameterType.mean) && (parType[p2]!=parameterType.mean))
                {
                    computeSigmaInvDev(p1, sigmaWork);
                    computeSigmaInvDev(p2, sigmaWork2);
                    computeSigmaInvDevDev(p1, p2, sigmaWork3);
                    Statik.multiply(sigmaWork, sigmaWork2, sigmaWork4);
                    double trace = 0;
                    for (int i=0; i<anzVar; i++) trace += sigmaWork3[i][i] - sigmaWork4[i][i];
                    llDD[p1][p2] += trace*anzPer;
                    Statik.multiply(sigmaWork4, sigInv, sigmaWork);
                    Statik.multiply(sigmaWork3, sigInv, sigmaWork2);
                    for (int i=0; i<anzVar; i++)
                        for (int j=0; j<anzVar; j++)
                            llDD[p1][p2] += (2*sigmaWork[i][j] - sigmaWork2[i][j])*(xBiSum[i][j] - anzPer*mu[i]*dataMean[j] - anzPer*mu[j]*dataMean[i] + anzPer*mu[i]*mu[j]);
                }
            }
    }

    public static void testPathtracking()
    {
        CholeskyModel model = new CholeskyModel(new int[]{0,1}, new double[]{2,3}, 
                new int[][]{{-1,-1},{-1,-1},{-1,6},{-1,7}}, new double[][]{{1,0},{1,1},{1,2},{1,3}}, 
                new int[][]{{2,-1},{3,4}}, new double[][]{{Math.sqrt(2),0},{0.5/Math.sqrt(2),Math.sqrt(15.0/8.0)}}, 
                5, 0.5);

//        model.fixParameter(new int[]{6,7});
        double[] pop = model.getParameter();
        model.createData(100);
        model.computeLogLikelihoodDerivatives(model.getArbitraryStartingValues(), true);
        double popLL = model.ll;
        int anzVar = model.anzVar;
        
        model.logStream = System.out;
        double EPS = 0.001;
        int anzRep = 1, ptFailed = 0, lsFailed = 0, abFailed = 0, ptSteps = 0, lsSteps = 0, abSteps = 0; 
        for (int i=0; i<anzRep; i++)
        {
            double varL = 1.0;                                                      // variance level chi^2 with mean 1., shifted by 1.
            double varS = rand.nextGaussian()/2; varS *= varS;                      // variance slope chi^2 with mean 1/4.
            double covLS = Math.sqrt(varL*varS)*(rand.nextDouble()-0.5)/0.5;        // correlation uniform from -0.25 to 0.25
            double err = 0.1 + rand.nextDouble()*0.4;                               // error uniform from 0.1 to 0.5
            pop[0] = rand.nextGaussian();                                           // means and path weights Gaussian with stdv 1, mean 0
            pop[1] = rand.nextGaussian();
            model.facCov[0][0] = varL; model.facCov[0][1] = model.facCov[1][0] = covLS; model.facCov[1][1] = varS;
            model.facCholVal = Statik.choleskyDecompose(model.facCov);
            pop[2] = model.getParameter(2); pop[3] = model.getParameter(3); pop[4] = model.getParameter(4);
            pop[5] = err;
            pop[6] = rand.nextGaussian();
            pop[7] = rand.nextGaussian();
            
            model.evaluateMuAndSigma(pop);
            model.createData(100);
            System.out.println("Population = "+Statik.matrixToString(pop));
            double[] popEstimate = model.estimateML(pop, EPS); popLL = model.ll;
            System.out.println("Pop Est    = "+Statik.matrixToString(popEstimate)+"\t steps = "+model.steps+" ll = "+model.ll);
            double[] estimate = model.estimateML(model.getArbitraryStartingValues(),EPS);
            System.out.println("AB Estimate= "+Statik.matrixToString(estimate)+"\t steps = "+model.steps+" ll = "+model.ll);
            abFailed += (model.ll - popLL > 1.0?1:0); abSteps += model.steps;
            estimate = model.estimateML(EPS);
            System.out.println("LS Estimate= "+Statik.matrixToString(estimate)+"\t steps = "+model.steps+" ll = "+model.ll);
            lsFailed += (model.ll - popLL > 1.0?1:0); lsSteps += model.steps;
            double[] pathEstimate = model.estimateMLByPathtracking(model.getArbitraryStartingValues(), EPS);
            System.out.println("Path Track = "+Statik.matrixToString(pathEstimate)+"\t steps = "+model.steps+" ll = "+model.ll);
            ptFailed += (model.ll - popLL > 1.0?1:0); ptSteps += model.steps;
            System.out.println();
        }
        System.out.println("Arbitrary Start = ("+100*(abFailed/(double)anzRep)+" %, "+abSteps/(double)anzRep+")");
        System.out.println("LS Start        = ("+100*(lsFailed/(double)anzRep)+" %, "+lsSteps/(double)anzRep+")");
        System.out.println("Path Tracking   = ("+100*(ptFailed/(double)anzRep)+" %, "+ptSteps/(double)anzRep+")");
        
        /*
        double[][] baseCov = new double[anzVar][anzVar], diffCov =new double[anzVar][anzVar]; 
        double[] baseMu = new double[anzVar], diffMu =new double[anzVar]; 
        for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) {baseCov[i][j] = model.sigma[i][j]; diffCov[i][j] = model.dataCov[i][j] - baseCov[i][j];}
        for (int i=0; i<anzVar; i++) {baseMu[i] = model.mu[i]; diffMu[i] = model.dataMean[i] - baseMu[i];}
        
        double lambda = 0.0;
        for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) model.dataCov[i][j] = baseCov[i][j] + lambda*diffCov[i][j];
        for (int i=0; i<anzVar; i++) model.dataMean[i] = baseMu[i] + lambda*diffMu[i];
        model.computeMomentsFromDataCovarianceAndMean();
        model.computeLogLikelihoodDerivatives(model.getArbitraryStartingValues(), true);
        model.computePathDerivatives(baseCov, diffCov, baseMu, diffMu, true);
        double orgLL = model.ll;
        double[] orgLLD = Statik.copy(model.llD);
        double EPS = 0.0001;
        lambda += EPS;
        for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) model.dataCov[i][j] = baseCov[i][j] + lambda*diffCov[i][j];
        for (int i=0; i<anzVar; i++) model.dataMean[i] = baseMu[i] + lambda*diffMu[i];
        model.computeMomentsFromDataCovarianceAndMean();
        model.computeLogLikelihoodDerivatives(model.getArbitraryStartingValues(), true);
        double numLLD = (model.ll - orgLL)/EPS;
        double[] numLLDD = new double[model.anzPar];
        for (int i=0; i<model.anzPar; i++) numLLDD[i] = (model.llD[i] - orgLLD[i])/EPS;
        System.out.println("Theoretical   = "+model.llPathD+", gradient = \t\t"+Statik.matrixToString(model.llPathDD));
        System.out.println("Numerical     = "+numLLD+", gradient = \t\t"+Statik.matrixToString(numLLDD));
        System.out.println("Actual llD   = "+Statik.matrixToString(model.llD));
        System.out.println("Previous llD = "+Statik.matrixToString(orgLLD));
        for (int i=0; i<model.anzPar; i++) model.setParameter(i, model.getParameter(i) + EPS*model.pathDirection[i]);
        model.computeLogLikelihoodDerivatives(null, true);
        System.out.println("Path     llD = "+Statik.matrixToString(orgLLD));

        model.computeMoments();
        double[] pathEstimate = model.estimateMLByPathtracking(0.001);
        */
        
        System.exit(0);
    }
    
    
    public static void main(String args[])
    {
        testPathtracking();
        
        CholeskyModel model = new CholeskyModel(new int[]{0,1}, new double[]{2,3}, 
                new int[][]{{-1,-1},{-1,-1},{-1,2},{-1,3},{-1,4}}, new double[][]{{1,0},{1,1},{1,2},{1,3},{1,4}}, 
                new int[][]{{5,7},{7,6}}, new double[][]{{2,0.5},{0.5,2}}, 
                8, 0.5);
        model.fixParameter(new int[]{3});
        
        model = new CholeskyModel(new int[]{0,1}, new double[]{2,3}, 
                new int[][]{{-1,-1},{-1,-1},{-1,6},{-1,7}}, new double[][]{{1,0},{1,1},{1,2},{1,3}}, 
                new int[][]{{2,-1},{3,4}}, new double[][]{{Math.sqrt(2),0},{0.5/Math.sqrt(2),Math.sqrt(15.0/8.0)}}, 
                5, 0.5);

        model.logStream = System.out;
//        model.fixParameter(new int[]{6,7});
        double[] pop = new double[model.anzPar];
        for (int i=0; i<model.anzPar; i++) pop[i] = model.getParameter(i);
        model.createData(100);

        
        model.computeLeastSquaresDerivatives(pop, true);
        double ls = model.ls;
        double[] derOrg = Statik.copy(model.lsD);
        double[] der = new double[pop.length];
        double[][] d2 = new double[pop.length][pop.length];
        double eps = 0.0000001;
        for (int i=0; i<pop.length; i++)
        {
            pop[i] += eps;
            model.computeLeastSquaresDerivatives(pop);
            der[i] = (model.ls - ls)/eps;
            pop[i] -= eps;
        }
        for (int i=0; i<pop.length; i++)
        {
            pop[i] += eps;
            model.computeLeastSquaresDerivatives(pop);
            for (int j=0; j<pop.length; j++)
                d2[i][j] = (model.lsD[j] - derOrg[j])/eps;
            pop[i] -= eps;
        }
        model.computeLeastSquaresDerivatives(pop, true);
        System.out.println("Numerical Values = ");
        System.out.println(Statik.matrixToString(der));
        System.out.println(Statik.matrixToString(d2));
        System.out.println("Symbolic  Values = ");
        System.out.println("True LS = "+model.ls);
        System.out.println(Statik.matrixToString(model.lsD));
        System.out.println(Statik.matrixToString(model.lsDD));
        
        for (int i=0; i<pop.length; i++) der[i] -= model.lsD[i];
        for (int i=0; i<pop.length; i++) for (int j=0; j<pop.length; j++) d2[i][j] -= model.lsDD[i][j];
        System.out.println("\r\nDifference = ");
        System.out.println(Statik.matrixToString(der));
        System.out.println(Statik.matrixToString(d2));

        model.setParameter(pop);
        int anzRep = 100; double[] estSum = new double[pop.length], estSSum = new double[pop.length], erg = null;
        for (int i=0; i<pop.length; i++) estSum[i] = estSSum[i] = 0;
        System.out.println("Population    = "+Statik.matrixToString(pop));
        int anzOK = 0;
        for (int i=0; i<anzRep; i++)
        {
            model.setParameter(pop);
            model.createData(1000);
            model.evaluateMuAndSigma(pop);
            model.estimateLS(pop, 0.001);
            System.out.print(model.ls+", ");
            erg = model.estimateLS(0.001);
            System.out.println(model.ls+", error = "+model.errVal);
            if (model.warningFlag != warningFlagTypes.OK) System.out.println("Susp. Estimate = "+Statik.matrixToString(erg));
            if (model.warningFlag != warningFlagTypes.FAILED) {anzOK++; for (int j=0; j<erg.length; j++) {estSum[j] += erg[j]; estSSum[j] += erg[j]*erg[j];}}
            
            if (i==0)
            {
                int p1 = 6, p2 = 7;
                double s1 = 0.12, s2 = 0.12;
                double[][] surface = new double[42][42], surface2 = new double[42][42];
                for (int k=0; k<41; k++) surface[k+1][0] = surface2[k+1][0] = erg[p1]+(k-20)*s1;
                for (int l=0; l<41; l++) surface[0][l+1] = surface2[0][l+1] = erg[p2]+(l-20)*s2;
                for (int k=0; k<41; k++)
                    for (int l=0; l<41; l++)
                    {
                        model.setParameter(erg);
                        model.setParameter(p1, model.getParameter(p1)+(k-20)*s1);
                        model.setParameter(p2, model.getParameter(p2)+(l-20)*s2);
                        model.computeLeastSquaresDerivatives(null, true);
                        surface[k+1][l+1] = model.ls;
                        surface2[k+1][l+1] = model.lsD[p1];
                    }
                Statik.writeMatrix(surface, "surface.txt", '\t');
                Statik.writeMatrix(surface2, "surface2.txt", '\t');
            }
        }
        for (int j=0; j<estSum.length; j++) {estSum[j] /= (double)anzOK; estSSum[j] = Math.sqrt(estSSum[j]/(double)anzOK - estSum[j]*estSum[j]);}
        System.out.println("Population         = "+Statik.matrixToString(pop));
        System.out.println("Mean Estimate (LS) = "+Statik.matrixToString(estSum));
        System.out.println("STDV Estimate (LS) = "+Statik.matrixToString(estSSum));
        System.out.println((anzRep-anzOK)+" of "+anzRep+" runs failed.");

        model.computeLogLikelihoodDerivatives(pop, true);
        double ll = model.ll;
        derOrg = Statik.copy(model.llD);
        der = new double[pop.length];
        d2 = new double[pop.length][pop.length];
        eps = 0.0000001;
        for (int i=0; i<pop.length; i++)
        {
            pop[i] += eps;
            model.computeLogLikelihoodDerivatives(pop);
            der[i] = (model.ll - ll)/eps;
            pop[i] -= eps;
        }
        for (int i=0; i<pop.length; i++)
        {
            pop[i] += eps;
            model.computeLogLikelihoodDerivatives(pop);
            for (int j=0; j<pop.length; j++)
                d2[i][j] = (model.llD[j] - derOrg[j])/eps;
            pop[i] -= eps;
        }
        model.computeLogLikelihoodDerivatives(pop, true);
        System.out.println("Numerical Values = ");
        System.out.println(Statik.matrixToString(der));
        System.out.println(Statik.matrixToString(d2));
        System.out.println("Symbolic  Values = ");
        System.out.println("True LL = "+model.ll);
        System.out.println(Statik.matrixToString(model.llD));
        System.out.println(Statik.matrixToString(model.llDD));
        
        for (int i=0; i<pop.length; i++) der[i] -= model.llD[i];
        for (int i=0; i<pop.length; i++) for (int j=0; j<pop.length; j++) d2[i][j] -= model.llDD[i][j];
        System.out.println("\r\nDifference = ");
        System.out.println(Statik.matrixToString(der));
        System.out.println(Statik.matrixToString(d2));
        
        
        double[] starting = new double[pop.length];
        model.setParameter(pop);
        System.out.println("Population        = "+Statik.matrixToString(pop));
        for (int i=0; i<pop.length; i++) estSum[i] = estSSum[i] = 0;
        anzOK = 0; int sus = 0; double sumll = 0;
        for (int i=0; i<anzRep; i++)
        {
            model.setParameter(pop);
            model.createData(100);
            model.evaluateMuAndSigma(pop);
            model.warningFlag = warningFlagTypes.OK;

//            starting = pop;
            starting = model.getArbitraryStartingValues();
//            starting = model.estimateLS(0.001);
            
            if (model.warningFlag != warningFlagTypes.FAILED) erg = model.estimateML(starting, 0.001);
            if (model.warningFlag != warningFlagTypes.OK) System.out.println("Susp. Estimate = "+Statik.matrixToString(erg));
            if (model.errVal > 0.8) 
            {
                System.out.println("High error: "+model.errVal+", ll = "+model.ll+", Warning = "+model.warningFlag);
                if (model.warningFlag != warningFlagTypes.FAILED) erg = model.estimateML(starting, 0.001);
            }
            if (model.warningFlag != warningFlagTypes.FAILED) {anzOK++; for (int j=0; j<erg.length; j++) {estSum[j] += erg[j]; estSSum[j] += erg[j]*erg[j];} sumll+=model.ll;}
            if (model.warningFlag == warningFlagTypes.SUSPICIOUS) sus++;
        }
        for (int j=0; j<estSum.length; j++) {estSum[j] /= (double)anzOK; estSSum[j] = Math.sqrt(estSSum[j]/(double)anzOK - estSum[j]*estSum[j]);}
        System.out.println("Population        = "+Statik.matrixToString(pop));
        System.out.println("Mean Estimate (ML)= "+Statik.matrixToString(estSum));
        System.out.println("STDV Estimate (ML)= "+Statik.matrixToString(estSSum));
        System.out.println("Average LL        = "+(sumll/(double)anzOK));
        System.out.println((anzRep-anzOK)+" of "+anzRep+" runs failed, "+sus+" were suspicious.");
        
        /*
        System.out.println("Mu = "+Statik.matrixToString(model.mu));
        System.out.println("Sigma = \r\n"+Statik.matrixToString(model.sigma));
        
        double[][] work = new double[model.anzVar][model.anzVar];
        model.sigInv = new double[model.anzVar][model.anzVar];
        Statik.invert(model.sigma, model.sigInv, work);  
        double[][] res = new double[3][3];
        for (int p = 0; p<8; p++)
        {
            model.computeSigmaInvDev(p, res);
            Statik.multiply(model.sigma, res, work);
            System.out.println("SigmaDev = \r\n"+Statik.matrixToString(work));
        }
        
        */
    }

}
