/*
 * Created on 21.04.2008
 */
package lgmsimulation;

import java.util.Random;
import silvers.Statik;

/**
 * Playclass to set up a univariate dual change score model with 5 occasions. Basis Sigma is given by a structure matrix of
 * 
 * Lambda := [[1,beta,beta^2,beta^3,beta^4],[0,alpha beta,alpha beta^2, alpha beta^3]]
 * 
 * and a 2x2 variance matrix of [[sigma, tau],[tau, nu]]. The error matrix is represented as five independent errors with values
 * 1/e1 ... 1/e5, the inverse to express infinite errors.
 *  
 * 
 * @author timo
 */
public class DCSM5 {
    
    public static double[] lastParameterDistribution;
    public static double[] lastLLRDistribution;
    public static double LASTBIAS = 0;
    
    public final static int ONREAL = 0, ONLATENT = 1, ONCHANGE = 2;
    public final static int GAMMAY = 0, BETAX = 1;

    public static double[][] computeStructure(double betaX, double betaY, double gammaX, double gammaY, int meas) 
        {double[][] erg = new double[meas][4]; computeStructure(erg,betaX,betaY,gammaX,gammaY); return erg;} 
    public static void computeStructure(double[][] erg, double betaX, double betaY, double gammaX, double gammaY)
    {
        erg[0][0] = 1; erg[0][1] = 0; erg[0][2] = 0; erg[0][3] = 0;
        erg[1][0] = 0; erg[1][1] = 0; erg[1][2] = 1; erg[1][3] = 0;
        for (int i=1; i<erg.length/2; i++)
        {
            for (int j=0; j<4; j++) erg[2*i][j] = 0;
            erg[2*i][1] = 1;
            for (int j=0; j<4; j++) erg[2*i][j] += betaX*erg[2*i-2][j] + gammaX*erg[2*i-1][j];
            for (int j=0; j<4; j++) erg[2*i+1][j] = 0;
            erg[2*i+1][3] = 1;
            for (int j=0; j<4; j++) erg[2*i+1][j] += gammaY*erg[2*i-2][j] + betaY*erg[2*i-1][j];
        }
    }
    
    /**
     * The first index of erg separates the value, the derivative wrt gammaY and the 2nd derivative wrt gammaY
     * @param erg
     * @param betaX
     * @param betaY
     * @param gammaX
     * @param gammaY
     */
    public static void computeStructureWRTGamma(double[][][] erg, double betaX, double betaY, double gammaX, double gammaY)
    {
        for (int i=0; i<3; i++) 
            for (int j=0; j<4; j++) erg[i][0][j] = erg[i][1][j] = 0;
        erg[0][0][0] = erg[0][1][2] = 1; 
        for (int i=1; i<erg[0].length/2; i++)
        {
            for (int k=0; k<3; k++) 
                for (int j=0; j<4; j++) erg[k][2*i][j] = betaX*erg[k][2*i-2][j] + gammaX*erg[k][2*i-1][j];
            erg[0][2*i][1] += 1;
            for (int j=0; j<4; j++) erg[0][2*i+1][j] = gammaY*erg[0][2*i-2][j] + betaY*erg[0][2*i-1][j];
            for (int j=0; j<4; j++) erg[1][2*i+1][j] = gammaY*erg[1][2*i-2][j] + betaY*erg[1][2*i-1][j] + erg[0][2*i-2][j];
            for (int j=0; j<4; j++) erg[2][2*i+1][j] = gammaY*erg[2][2*i-2][j] + betaY*erg[2][2*i-1][j] + 2*erg[1][2*i-2][j];
            erg[0][2*i+1][3] += 1;
        }
    }
    /**
     * The first index of erg separates the value, the derivative wrt betaX and the 2nd derivative wrt betaX
     * @param erg
     * @param betaX
     * @param betaY
     * @param gammaX
     * @param gammaY
     */
    public static void computeStructureWRTBeta(double[][][] erg, double betaX, double betaY, double gammaX, double gammaY)
    {
        for (int i=0; i<3; i++) 
            for (int j=0; j<4; j++) erg[i][0][j] = erg[i][1][j] = 0;
        erg[0][0][0] = erg[0][1][2] = 1; 
        for (int i=1; i<erg[0].length/2; i++)
        {
            for (int k=0; k<3; k++) 
                for (int j=0; j<4; j++) erg[k][2*i+1][j] = betaY*erg[k][2*i-1][j] + gammaY*erg[k][2*i-2][j];
            erg[0][2*i+1][3] += 1;
            for (int j=0; j<4; j++) erg[0][2*i][j] = gammaX*erg[0][2*i-1][j] + betaX*erg[0][2*i-2][j];
            for (int j=0; j<4; j++) erg[1][2*i][j] = gammaX*erg[1][2*i-1][j] + betaX*erg[1][2*i-2][j] + erg[0][2*i-2][j];
            for (int j=0; j<4; j++) erg[2][2*i][j] = gammaX*erg[2][2*i-1][j] + betaX*erg[2][2*i-1][j] + 2*erg[1][2*i-2][j];
            erg[0][2*i][1] += 1;
        }
    }

    public static void computeModelCovAndMean(double[][] Sigma, double[] mu, 
            double betaX, double betaY, double gammaX, double gammaY, double[][] facCov, double[] facMean, double[] error)
    {
        int meas = mu.length;
        double[][] structure = new double[meas][4];
        double[][] work = new double[meas][4], work2 = new double[4][meas];
        computeStructure(structure, betaX, betaY, gammaX, gammaY);
        Statik.multiply(structure, facCov, work);
        Statik.transpose(structure, work2);
        Statik.multiply(work, work2, Sigma);
        for (int i=0; i<meas; i++) Sigma[i][i] += error[i];
        Statik.multiply(structure, facMean, mu);
    }
    
    public static void computeModelCovAndMeanWRTParameter(double[][][] Sigma, double[][] mu, 
            double[][] facCov, double[] facMean, double[] error, double[][][] structure)
    {
        int meas = mu[0].length;
        double[][] work = new double[meas][4], work2 = new double[4][meas], work3 = new double[meas][meas];        
        Statik.multiply(structure[0], facCov, work);
        Statik.transpose(structure[0], work2);
        Statik.multiply(work, work2, Sigma[0]);
        for (int i=0; i<meas; i++) Sigma[0][i][i] += error[i];

        Statik.transpose(structure[2], work2);
        Statik.multiply(work, work2, Sigma[2]);
        Statik.transpose(structure[1], work2);
        Statik.multiply(structure[1], facCov, work);
        Statik.multiply(work, work2, work3);         
        for (int i=0; i<meas; i++) for (int j=0; j<meas; j++) Sigma[2][i][j] += work3[i][j];
        Statik.transpose(Sigma[2], work3);
        for (int i=0; i<meas; i++) for (int j=0; j<meas; j++) Sigma[2][i][j] += work3[i][j];

        Statik.transpose(structure[0], work2);
        Statik.multiply(work, work2, Sigma[1]);
        Statik.transpose(Sigma[1], work3);
        for (int i=0; i<meas; i++) for (int j=0; j<meas; j++) Sigma[1][i][j] += work3[i][j];

        Statik.multiply(structure[0], facMean, mu[0]);
        Statik.multiply(structure[1], facMean, mu[1]);
        Statik.multiply(structure[2], facMean, mu[2]);
    }

    // centMoments and sum real data!
    public static double[] computeMinusTwoLogLikelihoodWRTParameter(double[] erg, double[][][] Sigma, double[][] mu, double[][] centMoment, double[] sum, int N)
    {
        int meas = centMoment.length;

        double[][] work1 = new double[meas][meas];
        double[][] work2 = new double[meas][meas];
        double[][] invSig = new double[meas][meas];
        double[][] centMomentCorr = new double[meas][meas];
        double det = Statik.invert(Sigma[0], invSig, work1);
        for (int i=0; i<meas; i++)
            for (int j=0; j<meas; j++) centMomentCorr[i][j] = centMoment[i][j] - mu[0][i]*sum[j] - mu[0][j]*sum[i] + N*mu[0][i]*mu[0][j];
        double[] lokMu = new double[meas];  for (int i=0; i<meas; i++) lokMu[i] = sum[i] - N*mu[0][i];
        erg[0] = N*Math.log(det) + Statik.dotMultiply(centMomentCorr,invSig);
        erg[1] = -2*Statik.multiply(mu[1], invSig, lokMu);
        erg[2] = -2*Statik.multiply(mu[2], invSig, lokMu) + 2*Statik.multiply(mu[1],invSig,mu[1]);
        
        Statik.multiply(invSig,Sigma[1],work1);
        erg[1] += N*Statik.trace(work1);
        
        Statik.multiply(work1,invSig,work2);
        erg[1] -= Statik.dotMultiply(work2, centMomentCorr); 
        erg[2] += 4*Statik.multiply(mu[1], work2, lokMu);
        
        Statik.multiply(work2,Sigma[1],work1);
        erg[2] -= N*Statik.trace(work1);
        Statik.multiply(work1,invSig,work2);
        erg[2] += 2*Statik.dotMultiply(work2, centMomentCorr);
        Statik.multiply(invSig,Sigma[2],work1);
        erg[2] += N*Statik.trace(work1);
        Statik.multiply(work1, invSig, work2);
        erg[2] -= Statik.dotMultiply(work2, centMomentCorr);
        
        return erg;
    }
    // centMoments and sum real data!
    public static double[] computeMinusTwoLogLikelihoodWRTParameter(double[] erg, double betaX, double betaY, double gammaX, double gammaY, double[][] facCov, double[] facMean, double[] error, 
            double[][] centMoment, double[] sum, int N, int parameter)
    {
        int meas = centMoment.length;

        double[][][] Sigma = new double[3][meas][meas];
        double[][] mu = new double[3][meas];
        double[][][] structure = new double[3][meas][4];
        if (parameter == GAMMAY) computeStructureWRTGamma(structure, betaX, betaY, gammaX, gammaY);
        if (parameter == BETAX) computeStructureWRTBeta(structure, betaX, betaY, gammaX, gammaY);
        computeModelCovAndMeanWRTParameter(Sigma, mu, facCov, facMean, error, structure);
        return computeMinusTwoLogLikelihoodWRTParameter(erg, Sigma, mu, centMoment, sum, N);
    }
    // centMoments and sum real data!
    public static double computeMinusTwoLogLikelihood2(double[][] Sigma, double[] mu, double[][] centMoment, double[] sum, int N)
    {
        int meas = centMoment.length;

        double[][] work1 = new double[meas][meas];
        double[][] invSig = new double[meas][meas];
        double[][] centMomentCorr = new double[meas][meas];
        double det = Statik.invert(Sigma, invSig, work1);
        for (int i=0; i<meas; i++)
            for (int j=0; j<meas; j++) centMomentCorr[i][j] = centMoment[i][j] - mu[i]*sum[j] - mu[j]*sum[i] + N*mu[i]*mu[j];
        return N*Math.log(det) + Statik.dotMultiply(centMomentCorr,invSig);
    }
    
    // using the pregiven gammaY value as starting value.
    public static double optimizeGammaY(double betaX, double betaY, double gammaX, double gammaY, double[][] facCov, double[] facMean, double[] error, 
            int N, double[][] centMoment, double[] mean, double eps)
    {   
        final int MAXITER = 100;
        double[] gvals = new double[3];
        double step = Double.MAX_VALUE;
        double gy = gammaY;
        int c = 0;
        while ((Math.abs(step) > eps) && (c < MAXITER))
        {
            computeMinusTwoLogLikelihoodWRTParameter(gvals, betaX, betaY, gammaX, gy, facCov, facMean, error, centMoment, mean, N, GAMMAY);
            step = (-gvals[1]/gvals[2])*(c<5?0.5:1);
            gy += step;
            c++;
        }
        
        if (Math.abs(step) > eps) 
            return gammaY;
        return gy;
    }
    
    // using the pregiven gammaY or BetaX value as starting value.
    public static double optimizeParameterAttrition(double betaX, double betaY, double gammaX, double gammaY, double[][] facCov, double[] facMean, double[] error, 
            int[] N, double[][][] centMoment, double[][] mean, int[][] measuredTimePoints, double eps, int parameter)
    {   
        final int MAXITER = 100;
        double[] gvals = new double[3], work = new double[3];
        double[][][][] workSigma = new double[measuredTimePoints.length][3][][];
        double[][][] workMu = new double[measuredTimePoints.length][3][];
        for (int i=0; i<measuredTimePoints.length; i++) for (int j=0; j<3; j++) workSigma[i][j] = new double[measuredTimePoints[i].length][measuredTimePoints[i].length];
        for (int i=0; i<measuredTimePoints.length; i++) for (int j=0; j<3; j++) workMu[i][j] = new double[measuredTimePoints[i].length];
        
        double step = Double.MAX_VALUE;
        double x = (parameter==GAMMAY?gammaY:betaX);
        int c = 0;
        int meas = error.length;
        double[][][] Sigma = new double[3][meas][meas];
        double[][] mu = new double[3][meas];
        double[][][] structure = new double[3][meas][4];
        while ((Math.abs(step) > eps) && (c < MAXITER))
        {
            if (parameter==GAMMAY) computeStructureWRTGamma(structure, betaX, betaY, gammaX, x);
            else computeStructureWRTBeta(structure, x, betaY, gammaX, gammaY);
            computeModelCovAndMeanWRTParameter(Sigma, mu, facCov, facMean, error, structure);
            for (int i=0; i<3; i++) gvals[i] = 0;
            for (int pat = 0; pat<measuredTimePoints.length; pat++)
            {
                for (int i=0; i<3; i++) {Statik.submatrix(Sigma[i], measuredTimePoints[pat], measuredTimePoints[pat], workSigma[pat][i]); Statik.subvector(mu[i], measuredTimePoints[pat], workMu[pat][i]);}
                try {
                    computeMinusTwoLogLikelihoodWRTParameter(work, workSigma[pat], workMu[pat], centMoment[pat], mean[pat], N[pat]);
                } catch (Exception e) {System.out.println("Warning: Optimization failed (Matrix has no inverse)!"); return gammaY;}
                for (int i=0; i<3; i++) gvals[i] += work[i];
            }

            step = (-gvals[1]/gvals[2])*(c<5?0.8:1);
            x += step;
            c++;
        }
        
        if ((Math.abs(step) > eps) || (Double.isNaN(x))) 
        {System.out.println("Warning: Optimization failed (NaN)!"); return (parameter==GAMMAY?gammaY:betaX);}
        return x;
    }

    public static double computeMinusTwoLogLikelihood(double betaX, double betaY, double gammaX, double gammaY, double[][] facCov, double[] facMean, double[] error, 
            double[][] centMoment, double[] mean)
    {
        int meas = centMoment.length;
        double[][] Sigma = new double[meas][meas];
        double[] mu = new double[meas];
        computeModelCovAndMean(Sigma, mu, betaX, betaY, gammaX, gammaY, facCov, facMean, error);
        return computeMinusTwoLogLikelihood(Sigma, mu, centMoment, mean);
    }
    public static double computeMinusTwoLogLikelihood(double[][] Sigma, double[] mu, double[][] centMoment, double[] mean)
    {
        int meas = centMoment.length;
        double[][] invSig = new double[meas][meas];
        double[][] work3 = new double[meas][meas];
        double det = Statik.invert(Sigma, invSig, work3);
        double val = Math.log(det);
        for (int i=0; i<meas; i++) mu[i] = mean[i] - mu[i];
        for (int i=0; i<meas; i++)
            for (int j=0; j<meas; j++)
                val += invSig[i][j] * (centMoment[i][j] + mu[i] * mu[j]);
        return val;
    }
    
    public static double sarisSatorraPower(double betaX, double betaY, double gammaX, double gammaY, double[][] facCov, double[] facMean, double[] error, 
                  double betaXH, double betaYH, double gammaXH, double gammaYH, double[][] facCovH, double[] facMeanH, double[] errorH, 
                  int meas, int N, int df)
    {
        double[][] Sigma = new double[meas][meas];
        double[] mu = new double[meas];
        computeModelCovAndMean(Sigma, mu, betaX, betaY, gammaX, gammaY, facCov, facMean, error);
        double noncentrality = computeMinusTwoLogLikelihood(betaXH, betaYH, gammaXH, gammaYH, facCovH, facMeanH, errorH, Sigma, mu);
        noncentrality -= computeMinusTwoLogLikelihood(betaX, betaY, gammaX, gammaY, facCov, facMean, error, Sigma, mu);
        System.out.println(N*noncentrality+1);
        double power = Statik.chiSquareDensity(df, N*noncentrality, Statik.FIVEPERCENTTHRESHOLD[df-1]);
        return power;
    }
    
    public static double sarisSatorraPowerAttrition(double betaX, double betaY, double gammaX, double gammaY, double[][] facCov, double[] facMean, double[] error, 
            double betaXH, double betaYH, double gammaXH, double gammaYH, double[][] facCovH, double[] facMeanH, double[] errorH, 
            int[][] measuredTimePoints, double[] patternProb, int meas, int N, int df)
    {
        double[][][] Sigma = new double[meas][][], SigmaH = new double[meas][][];
        double[][] mu = new double[meas][], muH = new double[meas][];
        double[][] selErr = new double[meas][], selErrH = new double[meas][];
        for (int i=0; i<meas; i++) 
        {
            Sigma[i] = new double[i+1][i+1]; SigmaH[i] = new double[i+1][i+1]; 
            mu[i] = new double[i+1]; muH[i] = new double[i+1]; 
            selErr[i] = new double[i+1]; selErrH[i] = new double[i+1];
        }
        computeModelCovAndMean(Sigma[meas-1], mu[meas-1], betaX, betaY, gammaX, gammaY, facCov, facMean, error);
        computeModelCovAndMean(SigmaH[meas-1], muH[meas-1], betaXH, betaYH, gammaXH, gammaYH, facCovH, facMeanH, errorH);

        double power = 0; double snc = 0;
        for (int pat=0; pat<patternProb.length; pat++)
        {
            int selmeas = measuredTimePoints[pat].length;
            Statik.submatrix(Sigma[meas-1], measuredTimePoints[pat], measuredTimePoints[pat], Sigma[selmeas-1]);
            Statik.subvector(mu[meas-1], measuredTimePoints[pat], mu[selmeas-1]);
            Statik.submatrix(SigmaH[meas-1], measuredTimePoints[pat], measuredTimePoints[pat], SigmaH[selmeas-1]);
            Statik.subvector(muH[meas-1], measuredTimePoints[pat], muH[selmeas-1]);
            double noncentrality = computeMinusTwoLogLikelihood(SigmaH[selmeas-1], muH[selmeas-1], Sigma[selmeas-1], mu[selmeas-1]);
            noncentrality -= computeMinusTwoLogLikelihood(Sigma[selmeas-1], mu[selmeas-1], Sigma[selmeas-1], mu[selmeas-1]);
            snc += patternProb[pat] * noncentrality;
            power += patternProb[pat] * Statik.chiSquareDensity(df, N*patternProb[pat]*noncentrality, Statik.FIVEPERCENTTHRESHOLD[df-1]);
            // TODO : Ich traue dieser Summe nicht; die Verteilung ist eine Summe von gewichteten noncentral chisquares, hier wird eine Mixture genommen. 
        }
        return Statik.chiSquareDensity(df, N*snc, Statik.FIVEPERCENTTHRESHOLD[df-1]);
//        System.out.println("Sum SS Power = "+Statik.chiSquareDensity(df, N*snc, Statik.FIVEPERCENTTHRESHOLD[df-1]));
//        return Statik.chiSquareDensity(df, N*snc, Statik.FIVEPERCENTTHRESHOLD[df-1]);
//        return power;
    }

    public static double gammaPower(double betaX, double betaY, double gammaX, double gammaY, double[][] facCov, double[] facMean, double[] error, 
            double gammaYH, int N, int trials, int df)
    {
        int meas = error.length;
        Random r = new Random();
        double[][] Sigma = new double[meas][meas];
        double[] mu = new double[meas];
        double[][] dataCov = new double[meas][meas];
        double[] dataMean = new double[meas];
        double[] data = new double[meas];
        computeModelCovAndMean(Sigma, mu, betaX, betaY, gammaX, gammaY, facCov, facMean, error);
        double[][] chol = Statik.choleskyDecompose(facCov);
        double[][] structure = computeStructure(betaX, betaY, gammaX, gammaY, meas);
        double[][] strucChol = Statik.multiply(structure, chol);
        double[] obsMean = Statik.multiply(structure, facMean);
        double[] errStdv = new double[meas]; for (int i=0; i<meas; i++) errStdv[i] = Math.sqrt(error[i]);
        double[] work = new double[4];
          
        int count = 0;
        double meanDiff = 0;
        lastParameterDistribution = new double[trials];
        for (int trial=0; trial < trials; trial++)
        {
            for (int i=0; i<meas; i++) 
                for (int j=0; j<meas; j++) dataCov[i][j] = 0;
            for (int i=0; i<meas; i++) dataMean[i] = 0; 
//            r = new Random(1234567);        // DEBUG
            for (int i=0; i<N; i++)
            {
                for (int j=0; j<4; j++) work[j] = r.nextGaussian();
                Statik.multiply(strucChol, work, data);
                for (int j=0; j<meas; j++) data[j] += obsMean[j] + errStdv[j]*r.nextGaussian(); 
                for (int j=0; j<meas; j++) dataMean[j] += data[j]; 
                for (int j=0; j<meas; j++) 
                    for (int k=0; k<meas; k++) dataCov[j][k] += data[j]*data[k];
            }
            
            double gammaYE = optimizeGammaY(betaX, betaY, gammaX, gammaY, facCov, facMean, error, N, dataCov, dataMean, 0.00001);
            lastParameterDistribution[trials] = gammaYE;
            for (int i=0; i<meas; i++) dataMean[i] /= N; 
            for (int i=0; i<meas; i++) 
                for (int j=0; j<meas; j++) dataCov[i][j] = dataCov[i][j]/N - dataMean[i]*dataMean[j];
            double val = computeMinusTwoLogLikelihood(betaX, betaY, gammaX, gammaYH, facCov, facMean, error, dataCov, dataMean);
            double val2 = computeMinusTwoLogLikelihood(betaX, betaY, gammaX, gammaYE, facCov, facMean, error, dataCov, dataMean);
            val -= val2;
            val *= N;
            meanDiff += val;
            if (val > Statik.FIVEPERCENTTHRESHOLD[df-1]) count++;
        }
      
        // Debugging reasons:
        double noncentrality = computeMinusTwoLogLikelihood(betaX, betaY, gammaX, gammaYH, facCov, facMean, error, Sigma, mu);
        noncentrality -= computeMinusTwoLogLikelihood(betaX, betaY, gammaX, gammaY, facCov, facMean, error, Sigma, mu);
        System.out.println("Avarage Difference = "+meanDiff / (double)trials);
        System.out.println("Noncentrality = "+noncentrality*N);
        return (double)count / (double)trials;
    }
    
    // computes Power for gamma or beta, dependent on <code>parameter</code>
    public static double parameterPowerAttrition(double betaX, double betaY, double gammaX, double gammaY, double[][] facCov, double[] facMean, double[] error, 
            double parameterH, int N, int trials, int df, double meanReturnProb, double returnProbIncreasePerStdv, int baseNonrandomAttritionOn, int parameter)
    {
        int meas = error.length;
        Random r = new Random();
        double[][] Sigma = new double[meas][meas];
        double[] mu = new double[meas];
        double[][][] dataCov = new double[meas/2][][]; for (int i=0; i<meas/2; i++) dataCov[i] = new double[2*(i+1)][2*(i+1)];
        double[][] dataMean = new double[meas/2][]; for (int i=0; i<meas/2; i++) dataMean[i] = new double[2*(i+1)];
        double[] data = new double[meas], latentData = new double[meas];
        computeModelCovAndMean(Sigma, mu, betaX, betaY, gammaX, gammaY, facCov, facMean, error);
        double[][] chol = Statik.choleskyDecompose(facCov);
        double[][] structure = computeStructure(betaX, betaY, gammaX, gammaY, meas);
        double[][] strucChol = Statik.multiply(structure, chol);
        double[] obsMean = Statik.multiply(structure, facMean);
        double[] errStdv = new double[meas]; for (int i=0; i<meas; i++) errStdv[i] = Math.sqrt(error[i]);
        double[] totStdv = new double[meas]; for (int i=0; i<meas; i++) totStdv[i] = Math.sqrt(Sigma[i][i]);
        double[] work = new double[4];
        int[][] measuredTimePoints = new int[meas/2][];
        for (int i=0; i<meas/2; i++) {measuredTimePoints[i] = new int[2*(i+1)]; for (int j=0; j<2*(i+1); j++) measuredTimePoints[i][j] = j;}
        int[] Ns = new int[meas/2];
        double[][] SigmaH = new double[meas][meas];
        double[] muH = new double[meas];
        double[][][] SigmaHSub = new double[meas][][];
        double[][] muHSub = new double[meas][];
        if (parameter==GAMMAY) computeModelCovAndMean(SigmaH, muH, betaX, betaY, gammaX, parameterH, facCov, facMean, error);
        if (parameter==BETAX) computeModelCovAndMean(SigmaH, muH,  parameterH, betaY, gammaX, gammaY, facCov, facMean, error);
        for (int pat=0; pat<meas/2; pat++) 
        {
            SigmaHSub[pat] = new double[2*(pat+1)][2*(pat+1)]; muHSub[pat] = new double[2*(pat+1)];
            Statik.submatrix(SigmaH, measuredTimePoints[pat], measuredTimePoints[pat], SigmaHSub[pat]); Statik.subvector(muH, measuredTimePoints[pat], muHSub[pat]);
        }
        double[][] SigmaE = new double[meas][meas];
        double[] muE = new double[meas];
        double[][][] workSigma = new double[measuredTimePoints.length][][];
        double[][] workMu = new double[measuredTimePoints.length][];
        for (int i=0; i<measuredTimePoints.length; i++) workSigma[i] = new double[measuredTimePoints[i].length][measuredTimePoints[i].length];
        for (int i=0; i<measuredTimePoints.length; i++) workMu[i] = new double[measuredTimePoints[i].length];
        double[] Nss = new double[meas/2]; 
        Nss[0] = 1-meanReturnProb; for (int i=1; i<Nss.length-1; i++) Nss[i] = meanReturnProb * Nss[i-1];
        Nss[Nss.length-1] = 1.0; for (int i=0; i<Nss.length-1; i++) Nss[Nss.length-1] -= Nss[i];
        Nss[1] = Nss[1] + Nss[0]; Nss[0] = 0;

        double meanBias = 0;
        int obs = 0;
        int count = 0;
        double meanDiff = 0;
        lastParameterDistribution = new double[trials];
        lastLLRDistribution = new double[trials];
        for (int trial=0; trial < trials; trial++)
        {
            for (int k=0; k<meas/2; k++) for (int i=0; i<dataCov[k].length; i++) for (int j=0; j<dataCov[k][i].length; j++) dataCov[k][i][j] = 0;
            for (int k=0; k<meas/2; k++) for (int i=0; i<dataMean[k].length; i++) dataMean[k][i] = 0;
            for (int i = 0; i < meas/2; i++) Ns[i] = 0;
//            r = new Random(1234567);
            for (int i=0; i<N; i++)
            {
                for (int j=0; j<4; j++) work[j] = r.nextGaussian();
                Statik.multiply(strucChol, work, latentData);
                for (int j=0; j<meas; j++) data[j] = latentData[j] + obsMean[j] + errStdv[j]*r.nextGaussian();
                // selection
                int grp = 1; while (grp+1 < meas/2) 
                {
                    double returnProb = meanReturnProb;
                    if (baseNonrandomAttritionOn == ONREAL) returnProb += returnProbIncreasePerStdv*((data[2*grp]+data[2*grp+1]-mu[2*grp]-mu[2*grp+1])/
                            Math.sqrt(totStdv[2*grp]*totStdv[2*grp]+totStdv[2*grp+1]*totStdv[2*grp+1]));
                    if (baseNonrandomAttritionOn == ONLATENT) returnProb += returnProbIncreasePerStdv*((latentData[2*grp]+latentData[2*grp+1]-mu[2*grp]-mu[2*grp+1])/
                            Math.sqrt(totStdv[2*grp]*totStdv[2*grp]+totStdv[2*grp+1]*totStdv[2*grp+1]));
                    if ((baseNonrandomAttritionOn == ONCHANGE) && (grp>=1)) returnProb += returnProbIncreasePerStdv*
                        ((latentData[2*grp]+latentData[2*grp+1]-latentData[2*grp-1]-latentData[2*grp-2]-mu[2*grp]-mu[2*grp+1]+mu[2*grp-1]+mu[2*grp-2])/
                        Math.sqrt(totStdv[2*grp-2]*totStdv[2*grp-2]+totStdv[2*grp-1]*totStdv[2*grp-1]+totStdv[2*grp]*totStdv[2*grp]+totStdv[2*grp+1]*totStdv[2*grp+1]));
                    if (r.nextDouble() > returnProb) break; 
                    grp++;
                }
                if (grp==0) grp++;
                int jmp=1; while ((grp >= meas/2) || (grp <=0) || (Ns[grp]>=Nss[grp]*N)) {grp += jmp; jmp = (jmp<0?-jmp+1:-jmp-1);} 
                Ns[grp]++; obs += (grp+1); 
                for (int j=0; j<2*(grp+1); j++) dataMean[grp][j] += data[j]; 
                for (int j=0; j<2*(grp+1); j++) 
                    for (int k=0; k<2*(grp+1); k++) dataCov[grp][j][k] += data[j]*data[k];
            }
            
            double parameterE = optimizeParameterAttrition(betaX, betaY, gammaX, gammaY, facCov, facMean, error, Ns, dataCov, dataMean, measuredTimePoints, 0.0001, parameter);
            lastParameterDistribution[trial] = parameterE;
            
            if (parameter==GAMMAY) computeModelCovAndMean(SigmaE, muE, betaX, betaY, gammaX, parameterE, facCov, facMean, error);
            if (parameter==BETAX) computeModelCovAndMean(SigmaE, muE, parameterE, betaY, gammaX, gammaY, facCov, facMean, error);
            double val = 0;
            for (int pat = 0; pat<measuredTimePoints.length; pat++)
            {
                Statik.submatrix(SigmaE, measuredTimePoints[pat], measuredTimePoints[pat], workSigma[pat]); Statik.subvector(muE, measuredTimePoints[pat], workMu[pat]);
                val -= computeMinusTwoLogLikelihood2(workSigma[pat], workMu[pat], dataCov[pat], dataMean[pat], Ns[pat]);
                val += computeMinusTwoLogLikelihood2(SigmaHSub[pat], muHSub[pat], dataCov[pat], dataMean[pat], Ns[pat]);
            }
            lastLLRDistribution[trial] = val;
            
            meanBias += parameterE;
            meanDiff += val;
            if (val > Statik.FIVEPERCENTTHRESHOLD[df-1]) count++;
        }
      
        LASTBIAS = ((double)meanBias / (double)trials) - (parameter==GAMMAY?gammaY:betaX);
        return (double)count / (double)trials;
    }

    // computes Power for gamma or beta, dependent on <code>parameter</code>
    public static double parameterPowerAttritionFixedPattern(double betaX, double betaY, double gammaX, double gammaY, double[][] facCov, double[] facMean, double[] error, 
            double parameterH, int N, int trials, int df, int[][] pattern, double[] patternProb, int parameter)
    {
        int meas = error.length;
        Random r = new Random();
        double[][] Sigma = new double[meas][meas];
        double[] mu = new double[meas];
        double[][][] dataCov = new double[pattern.length][][]; for (int i=0; i<pattern.length; i++) dataCov[i] = new double[pattern[i].length][pattern[i].length];
        double[][] dataMean = new double[pattern.length][]; for (int i=0; i<pattern.length; i++) dataMean[i] = new double[pattern[i].length];
        double[] data = new double[meas];
        computeModelCovAndMean(Sigma, mu, betaX, betaY, gammaX, gammaY, facCov, facMean, error);
        double[][] chol = Statik.choleskyDecompose(facCov);
        double[][] structure = computeStructure(betaX, betaY, gammaX, gammaY, meas);
        double[][] strucChol = Statik.multiply(structure, chol);
        double[] obsMean = Statik.multiply(structure, facMean);
        double[] errStdv = new double[meas]; for (int i=0; i<meas; i++) errStdv[i] = Math.sqrt(error[i]);
        double[] work = new double[4];
        int[] Ns = new int[pattern.length];
        for (int i=0; i<pattern.length; i++) Ns[i] = (int)Math.round(patternProb[i] * N);
        double[][] SigmaH = new double[meas][meas];
        double[] muH = new double[meas];
        double[][][] SigmaHSub = new double[pattern.length][][];
        double[][] muHSub = new double[pattern.length][];
        if (parameter==GAMMAY) computeModelCovAndMean(SigmaH, muH, betaX, betaY, gammaX, parameterH, facCov, facMean, error);
        if (parameter==BETAX) computeModelCovAndMean(SigmaH, muH,  parameterH, betaY, gammaX, gammaY, facCov, facMean, error);
        for (int pat=0; pat<pattern.length; pat++) 
        {
            SigmaHSub[pat] = new double[pattern[pat].length][pattern[pat].length]; muHSub[pat] = new double[pattern[pat].length];
            Statik.submatrix(SigmaH, pattern[pat], pattern[pat], SigmaHSub[pat]); Statik.subvector(muH, pattern[pat], muHSub[pat]);
        }
        double[][] SigmaE = new double[meas][meas];
        double[] muE = new double[meas];
        double[][][] workSigma = new double[pattern.length][][];
        double[][] workMu = new double[pattern.length][];
        for (int i=0; i<pattern.length; i++) workSigma[i] = new double[pattern[i].length][pattern[i].length];
        for (int i=0; i<pattern.length; i++) workMu[i] = new double[pattern[i].length];

        double meanBias = 0;
        int obs = 0;
        int count = 0;
        double meanDiff = 0;
        lastParameterDistribution = new double[trials];
        lastLLRDistribution = new double[trials];
        for (int trial=0; trial < trials; trial++)
        {
            for (int k=0; k<pattern.length; k++) for (int i=0; i<dataCov[k].length; i++) for (int j=0; j<dataCov[k][i].length; j++) dataCov[k][i][j] = 0;
            for (int k=0; k<pattern.length; k++) for (int i=0; i<dataMean[k].length; i++) dataMean[k][i] = 0;
            
            for (int i=0; i<pattern.length; i++)
                for (int k=0; k<Ns[i]; k++)
                {
                    for (int j=0; j<4; j++) work[j] = r.nextGaussian();
                    Statik.multiply(strucChol, work, data);
                    for (int j=0; j<meas; j++) data[j] += obsMean[j] + errStdv[j]*r.nextGaussian();
                    for (int j=0; j<pattern[i].length; j++) dataMean[i][j] += data[pattern[i][j]]; 
                    for (int j=0; j<pattern[i].length; j++) 
                        for (int l=0; l<pattern[i].length; l++) dataCov[i][j][l] += data[pattern[i][j]]*data[pattern[i][l]];
                }
            
            double parameterE = optimizeParameterAttrition(betaX, betaY, gammaX, gammaY, facCov, facMean, error, Ns, dataCov, dataMean, pattern, 0.0001, parameter);
            lastParameterDistribution[trial] = parameterE;
            
            if (parameter==GAMMAY) computeModelCovAndMean(SigmaE, muE, betaX, betaY, gammaX, parameterE, facCov, facMean, error);
            if (parameter==BETAX) computeModelCovAndMean(SigmaE, muE, parameterE, betaY, gammaX, gammaY, facCov, facMean, error);
            double val = 0;
            for (int pat = 0; pat<pattern.length; pat++)
            {
                Statik.submatrix(SigmaE, pattern[pat], pattern[pat], workSigma[pat]); Statik.subvector(muE, pattern[pat], workMu[pat]);
                val -= computeMinusTwoLogLikelihood2(workSigma[pat], workMu[pat], dataCov[pat], dataMean[pat], Ns[pat]);
                val += computeMinusTwoLogLikelihood2(SigmaHSub[pat], muHSub[pat], dataCov[pat], dataMean[pat], Ns[pat]);
            }
            lastLLRDistribution[trial] = val;
            
            meanBias += parameterE;
            meanDiff += val;
            if (val > Statik.FIVEPERCENTTHRESHOLD[df-1]) count++;
        }
      
        LASTBIAS = ((double)meanBias / (double)trials) - (parameter==GAMMAY?gammaY:betaX);
        return (double)count / (double)trials;
    }

    // Determinant of Sigma without the denominator
    public static double numerDet(double alpha, double beta, double sigma, double tau, double nu, double[] error)
    {
        double[] coeff = new double[9]; numerDetCoeffsBeta(coeff, alpha, sigma, tau, nu, error);
        double val = coeff[8];
        val = beta*val + coeff[7];
        val = beta*val + coeff[6];
        val = beta*val + coeff[5];
        val = beta*val + coeff[4];
        val = beta*val + coeff[3];
        val = beta*val + coeff[2];
        val = beta*val + coeff[1];
        val = beta*val + coeff[0];
        
        return val;
    }
   
    // Determinant of Sigma without the denominator, coefficients of beta
    public static void numerDetCoeffsBeta(double[] erg, double alpha, double sigma, double tau, double nu, double[] error)
    {
        double twoAlphaTau = 2*alpha*tau;
        double alphaSq = alpha*alpha;
        double alphaSqDet = alphaSq*(sigma*nu-tau*tau);
        erg[8] = error[4]*sigma;
        erg[7] = twoAlphaTau*error[4];
        erg[6] = alphaSqDet*error[4]*(error[0]+error[1]+error[2]+error[3])+ error[4]*(alphaSq*nu + 2*alpha*tau) + error[3]*sigma;
        erg[5] = alphaSqDet*(2*error[0]*error[4]+2*error[1]*error[4]+2*error[2]*error[4])+2*alphaSq*nu*error[4]+2*alpha*tau*(error[3]+error[4]);
        erg[4] = alphaSqDet*(error[0]*error[3]+3*error[0]*error[4]+error[1]*error[3]+3*error[1]*error[4]+error[2]*error[3]+error[2]*error[4])+alphaSq*nu*(error[3]+3*error[4])+alpha*tau*(2*error[3]+2*error[4])+error[2]*sigma;
        erg[3] = alphaSqDet*(2*error[0]*error[3]+4*error[0]*error[4]+2*error[1]*error[3]+2*error[1]*error[4])+alphaSq*nu*(2*error[3]+4*error[4])+alpha*tau*(2*error[2]+2*error[3]);
        erg[2] = alphaSqDet*(error[0]*error[2]+3*error[0]*error[3]+3*error[0]*error[4]+error[1]*error[2]+error[1]*error[3]+error[1]*error[4])+alphaSq*nu*(error[2]+3*error[3]+3*error[4])+alpha*tau*2*error[2]+error[1]*sigma;
        erg[1] = alphaSqDet*(2*error[0]*error[2]+2*error[0]*error[3]+2*error[0]*error[4])+alphaSq*nu*(2*error[2]+2*error[3]+2*error[4])+2*alpha*tau*error[1];
        erg[0] = alphaSqDet*(error[0]*error[1]+error[0]*error[2]+error[0]*error[3]+error[0]*error[4])+alphaSq*nu*(error[1]+error[2]+error[3]+error[4])+error[0]*sigma+1;
    }
    
    // Determinant of Sigma without the denominator, coefficients of beta and then alpha
    public static void numerDetCoeffs(double[][] erg, double sigma, double tau, double nu, double[] error)
    {
        for (int i=0; i<9; i++) 
            for (int j=0; j<3; j++)
                erg[i][j] = 0;
        
//        double twoAlphaTau = 2*alpha*tau;
//        double alphaSq = alpha*alpha;
        double det = (sigma*nu-tau*tau);
        erg[8][0] = error[4]*sigma;
        erg[7][1] = 2*tau*error[4];
        erg[6][2] = det*error[4]*(error[0]+error[1]+error[2]+error[3])+error[4]*nu; erg[6][1] = error[4]*2*tau; erg[6][0] = error[3]*sigma;
        erg[5][2] = det*(2*error[0]*error[4]+2*error[1]*error[4]+2*error[2]*error[4])+2*nu*error[4]; erg[5][1] = 2*tau*(error[3]+error[4]);
        erg[4][2] = det*(error[0]*error[3]+3*error[0]*error[4]+error[1]*error[3]+3*error[1]*error[4]+error[2]*error[3]+error[2]*error[4])+nu*(error[3]+3*error[4]); erg[4][1] = tau*(2*error[3]+2*error[4]); erg[4][0] = error[2]*sigma;
        erg[3][2] = det*(2*error[0]*error[3]+4*error[0]*error[4]+2*error[1]*error[3]+2*error[1]*error[4])+nu*(2*error[3]+4*error[4]); erg[3][1] = tau*(2*error[2]+2*error[3]);
        erg[2][2] = det*(error[0]*error[2]+3*error[0]*error[3]+3*error[0]*error[4]+error[1]*error[2]+error[1]*error[3]+error[1]*error[4])+nu*(error[2]+3*error[3]+3*error[4]); erg[2][1] = tau*2*error[2]; erg[2][0] = error[1]*sigma;
        erg[1][2] = det*(2*error[0]*error[2]+2*error[0]*error[3]+2*error[0]*error[4])+nu*(2*error[2]+2*error[3]+2*error[4]); erg[1][1] = 2*tau*error[1];
        erg[0][2] = det*(error[0]*error[1]+error[0]*error[2]+error[0]*error[3]+error[0]*error[4])+nu*(error[1]+error[2]+error[3]+error[4]); erg[0][0] = error[0]*sigma+1;
    }

    /**
     *  Computes the adjoint of Sigma multiplied by e1*...*e5.  
     */
    public static void adjointE(double[][] erg, double alpha, double beta, double sigma, double tau, double nu, double[] error)
    {
        double[][][] ergPol = new double[5][5][9];
        adjointECoeffsBeta(ergPol, alpha, sigma, tau, nu, error);
        for (int i=0; i<5; i++)
            for (int j=0; j<5; j++)
                erg[i][j] = Statik.evaluate(ergPol[i][j], beta);
    }
    
    /**
     *  Computes the adjoint of Sigma in the coefficients of beta, multiplied by e1*...*e5.
     *  The exponent of beta is the last index of erg, exponent zero first.  
     */
    public static void adjointECoeffsBeta(double[][][] erg, double alpha, double sigma, double tau, double nu, double[] error)
    {
        for (int i=0; i<5; i++)
            for (int j=0; j<5; j++)
                for (int k=0; k<9; k++)
                    erg[i][j][k] = 0;
        
        double alphaSqNu = alpha*alpha*nu;
        double alphaSqDet = alpha*alpha*(sigma*nu-tau*tau);
        double alphaTau = alpha*tau;
        double e1 = error[0], e2 = error[1], e3 = error[2], e4 = error[3], e5 = error[4];
        
        erg[0][0][0] = alphaSqNu*(e2+e3+e4+e5)+1;
        erg[0][0][1] = 2*alphaSqNu*(e3+e4+e5)+2*alphaTau*e2;
        erg[0][0][2] = alphaSqDet*(e2*e3+e2*e4+e2*e5)+alphaSqNu*(e3+3*e4+3*e5)+2*alphaTau*e3+sigma*e2;
        erg[0][0][3] = alphaSqDet*(2*e2*e4+2*e2*e5)+alphaSqNu*(2*e4+4*e5)+alphaTau*(2*e3+2*e4);
        erg[0][0][4] = alphaSqDet*(e2*e4+3*e2*e5+e3*e4+e3*e5)+alphaSqNu*(e4+3*e5)+alphaTau*(2*e4+2*e5)+sigma*e3;
        erg[0][0][5] = alphaSqDet*(2*e2*e5+2*e3*e5)+alphaSqNu*(2*e5)+alphaTau*(2*e4+2*e5);
        erg[0][0][6] = alphaSqDet*(e2*e5+e3*e5+e4*e5)+alphaSqNu*(e5)+alphaTau*(2*e5)+sigma*e4;
        erg[0][0][7] = 2*alphaTau*e5;
        erg[0][0][8] = sigma*e5;
        
        erg[0][1][0] = alphaTau;
        erg[0][1][1] = alphaSqDet*(e3+e4+e5)+sigma;
        erg[0][1][2] = alphaSqDet*(e3+2*e4+2*e5);
        erg[0][1][3] = alphaSqDet*(2*e4+3*e5);
        erg[0][1][4] = alphaSqDet*(e4+3*e5);
        erg[0][1][5] = alphaSqDet*2*e5;
        erg[0][1][6] = alphaSqDet*e5;
        
        erg[0][2][0] = alphaTau;
        erg[0][2][1] = alphaSqDet*(-e2)+alphaTau;
        erg[0][2][2] = alphaSqDet*(e4+e5)+sigma;
        erg[0][2][3] = alphaSqDet*(e4+2*e5);
        erg[0][2][4] = alphaSqDet*(e4+2*e5);
        erg[0][2][5] = alphaSqDet*(2*e5);
        erg[0][2][6] = alphaSqDet*e5;
        
        erg[0][3][0] = alphaTau;
        erg[0][3][1] = alphaSqDet*(-e2)+alphaTau;
        erg[0][3][2] = alphaSqDet*(-e2-e3)+alphaTau;
        erg[0][3][3] = alphaSqDet*(-e3+e5)+sigma;
        erg[0][3][4] = alphaSqDet*(e5);
        erg[0][3][5] = alphaSqDet*(e5);
        erg[0][3][6] = alphaSqDet*(e5);
        
        erg[0][4][0] = alphaTau; 
        erg[0][4][1] = alphaSqDet*(-e2)+alphaTau;
        erg[0][4][2] = alphaSqDet*(-e2-e3)+alphaTau;
        erg[0][4][3] = alphaSqDet*(-e2-2*e3-e4)+alphaTau;
        erg[0][4][4] = alphaSqDet*(-e3-e4)+sigma;
        erg[0][4][5] = alphaSqDet*(-e4);
        
        erg[1][1][0] = alphaSqDet*(e1*e3+e1*e4+e1*e5)+alphaSqNu*(e3+e4+e5)+e1*sigma+1; 
        erg[1][1][1] = alphaSqDet*(2*e1*e3+2*e1*e4+2*e1*e5)+alphaSqNu*(2*e3+2*e4+2*e5);
        erg[1][1][2] = alphaSqDet*(e1*e3+3*e1*e4+3*e1*e5)+alphaSqNu*(e3+3*e4+3*e5)+2*alphaTau*e3;
        erg[1][1][3] = alphaSqDet*(2*e1*e4+4*e1*e5)+alphaSqNu*(2*e4+4*e5)+alphaTau*(2*e3+2*e4);
        erg[1][1][4] = alphaSqDet*(e1*e4+3*e1*e5+e3*e4+e3*e5)+alphaSqNu*(e4+3*e5)+alphaTau*(2*e4+2*e5)+e3*sigma;
        erg[1][1][5] = alphaSqDet*(2*e1*e5+2*e3*e5)+alphaSqNu*(2*e5)+alphaTau*(2*e4+2*e5);
        erg[1][1][6] = alphaSqDet*(e1*e5+e3*e5+e4*e5)+alphaSqNu*(e5)+alphaTau*(2*e5)+sigma*e4;
        erg[1][1][7] = 2*alphaTau*e5;
        erg[1][1][8] = sigma*e5;
        
        erg[1][2][0] = alphaSqDet*e1+alphaSqNu;
        erg[1][2][1] = alphaSqDet*e1+alphaSqNu+alphaTau;
        erg[1][2][2] = 2*alphaTau;
        erg[1][2][3] = alphaSqDet*(e4+e5)+sigma;
        erg[1][2][4] = alphaSqDet*(e4+2*e5);
        erg[1][2][5] = alphaSqDet*2*e5;
        erg[1][2][6] = alphaSqDet*e5;

        erg[1][3][0] = alphaSqDet*e1+alphaSqNu;
        erg[1][3][1] = alphaSqDet*e1+alphaSqNu+alphaTau;
        erg[1][3][2] = alphaSqDet*e1+alphaSqNu+alphaTau;
        erg[1][3][3] = alphaSqDet*(-e3)+2*alphaTau; 
        erg[1][3][4] = alphaSqDet*e5+sigma;
        erg[1][3][5] = alphaSqDet*e5;
        erg[1][3][6] = alphaSqDet*e5;

        erg[1][4][0] = alphaSqDet*e1+alphaSqNu; 
        erg[1][4][1] = alphaSqDet*e1+alphaSqNu+alphaTau;
        erg[1][4][2] = alphaSqDet*e1+alphaSqNu+alphaTau;
        erg[1][4][3] = alphaSqDet*(e1-e3)+alphaSqNu+alphaTau;
        erg[1][4][4] = alphaSqDet*(-e3-e4)+2*alphaTau;
        erg[1][4][5] = alphaSqDet*(-e4)+sigma;

        erg[2][2][0] = alphaSqDet*(e1*e2+e1*e4+e1*e5)+alphaSqNu*(e2+e4+e5)+e1*sigma+1; 
        erg[2][2][1] = alphaSqDet*(2*e1*e4+2*e1*e5)+alphaSqNu*(2*e4+2*e5)+2*alphaTau*e2;
        erg[2][2][2] = alphaSqDet*(3*e1*e4+3*e1*e5+e2*e4+e2*e5)+alphaSqNu*(3*e4+3*e5)+e2*sigma;
        erg[2][2][3] = alphaSqDet*(2*e1*e4+4*e1*e5+2*e2*e4+2*e2*e5)+alphaSqNu*(2*e4+4*e5)+alphaTau*(2*e4);
        erg[2][2][4] = alphaSqDet*(e1*e4+3*e1*e5+e2*e4+3*e2*e5)+alphaSqNu*(e4+3*e5)+alphaTau*(2*e4+2*e5);
        erg[2][2][5] = alphaSqDet*(2*e1*e5+2*e2*e5)+alphaSqNu*(2*e5)+alphaTau*(2*e4+2*e5);
        erg[2][2][6] = alphaSqDet*(e1*e5+e2*e5+e4*e5)+alphaSqNu*(e5)+alphaTau*(2*e5)+sigma*e4;
        erg[2][2][7] = 2*alphaTau*e5;
        erg[2][2][8] = sigma*e5;
        
        erg[2][3][0] = alphaSqDet*(e1)+alphaSqNu;
        erg[2][3][1] = alphaSqDet*(2*e1)+2*alphaSqNu;
        erg[2][3][2] = alphaSqDet*(2*e1+e2)+2*alphaSqNu+alphaTau;
        erg[2][3][3] = alphaSqDet*(e1+e2)+alphaSqNu+2*alphaTau; 
        erg[2][3][4] = 2*alphaTau;
        erg[2][3][5] = alphaSqDet*(e5)+sigma;
        erg[2][3][6] = alphaSqDet*e5;
        
        erg[2][4][0] = alphaSqDet*(e1)+alphaSqNu;
        erg[2][4][1] = alphaSqDet*(2*e1)+2*alphaSqNu;
        erg[2][4][2] = alphaSqDet*(2*e1+e2)+2*alphaSqNu+alphaTau;
        erg[2][4][3] = alphaSqDet*(2*e1+e2)+2*alphaSqNu+alphaTau;
        erg[2][4][4] = alphaSqDet*(e1+e2)+alphaSqNu+2*alphaTau;
        erg[2][4][5] = alphaSqDet*(-e4)+2*alphaTau;
        erg[2][4][6] = sigma;

        erg[3][3][0] = alphaSqDet*(e1*e2+e1*e3+e1*e5)+alphaSqNu*(e2+e3+e5)+e1*sigma+1; 
        erg[3][3][1] = alphaSqDet*(2*e1*e3+2*e1*e5)+alphaSqNu*(2*e3+2*e5)+2*alphaTau*e2;
        erg[3][3][2] = alphaSqDet*(e1*e3+3*e1*e5+e2*e3+e2*e5)+alphaSqNu*(e3+3*e5)+2*alphaTau*e3+e2*sigma;
        erg[3][3][3] = alphaSqDet*(4*e1*e5+2*e2*e5)+alphaSqNu*(4*e5)+alphaTau*(2*e3);
        erg[3][3][4] = alphaSqDet*(3*e1*e5+3*e2*e5+e3*e5)+alphaSqNu*(3*e5)+alphaTau*(2*e5)+sigma*e3;
        erg[3][3][5] = alphaSqDet*(2*e1*e5+2*e2*e5+2*e3*e5)+alphaSqNu*(2*e5)+alphaTau*(2*e5);
        erg[3][3][6] = alphaSqDet*(e1*e5+e2*e5+e3*e5)+alphaSqNu*(e5)+alphaTau*(2*e5);
        erg[3][3][7] = 2*alphaTau*e5;
        erg[3][3][8] = sigma*e5;
        
        erg[3][4][0] = alphaSqDet*(e1)+alphaSqNu;
        erg[3][4][1] = alphaSqDet*(2*e1)+2*alphaSqNu;
        erg[3][4][2] = alphaSqDet*(3*e1+e2)+3*alphaSqNu;
        erg[3][4][3] = alphaSqDet*(3*e1+2*e2)+3*alphaSqNu+alphaTau;
        erg[3][4][4] = alphaSqDet*(2*e1+2*e2+e3)+2*alphaSqNu+2*alphaTau;
        erg[3][4][5] = alphaSqDet*(e1+e2+e3)+alphaSqNu+2*alphaTau;
        erg[3][4][6] = 2*alphaTau;
        erg[3][4][7] = sigma;

        erg[4][4][0] = alphaSqDet*(e1*e2+e1*e3+e1*e4)+alphaSqNu*(e2+e3+e4)+e1*sigma+1; 
        erg[4][4][1] = alphaSqDet*(2*e1*e3+2*e1*e4)+alphaSqNu*(2*e3+2*e4)+2*alphaTau*e2;
        erg[4][4][2] = alphaSqDet*(e1*e3+3*e1*e4+e2*e3+e2*e4)+alphaSqNu*(e3+3*e4)+2*alphaTau*e3+e2*sigma;
        erg[4][4][3] = alphaSqDet*(2*e1*e4+2*e2*e4)+alphaSqNu*(2*e4)+alphaTau*(2*e3+2*e4);
        erg[4][4][4] = alphaSqDet*(e1*e4+e2*e4+e3*e4)+alphaSqNu*(e4)+alphaTau*(2*e4)+sigma*e3;
        erg[4][4][5] = 2*alphaTau*e4;
        erg[4][4][6] = sigma*e4;
        
        
        for (int i=0; i<5; i++)
            for (int k=0; k<9; k++)
            {
                for (int j=i+1; j<5; j++)
                    {erg[i][j][k] *= -error[i]*error[j]; erg[j][i][k] = erg[i][j][k];}
                erg[i][i][k] *= error[i];
            }
    }
    
    /**
     *  Computes the adjoint of Sigma in the coefficients of beta, multiplied by e1*...*e5.
     *  The exponent of beta is the one before last index of erg, exponent zero first, and alpha is the last index.  
     */
    public static void adjointEBothCoeffs(double[][][][] erg, double sigma, double tau, double nu, double[] error)
    {
        for (int i=0; i<5; i++)
            for (int j=0; j<5; j++)
                for (int k=0; k<9; k++)
                    for (int l=0; l<3; l++)
                        erg[i][j][k][l] = 0;
        
        double det = sigma*nu-tau*tau;
        double e1 = error[0], e2 = error[1], e3 = error[2], e4 = error[3], e5 = error[4];
        
        erg[0][0][0][2] = nu*(e2+e3+e4+e5); erg[0][0][0][0] = 1;
        erg[0][0][1][2] = 2*nu*(e3+e4+e5); erg[0][0][1][1] = 2*tau*e2;
        erg[0][0][2][2] = det*(e2*e3+e2*e4+e2*e5)+nu*(e3+3*e4+3*e5); erg[0][0][2][1]=2*tau*e3; erg[0][0][2][0] = sigma*e2;
        erg[0][0][3][2] = det*(2*e2*e4+2*e2*e5)+nu*(2*e4+4*e5); erg[0][0][3][1] = tau*(2*e3+2*e4);
        erg[0][0][4][2] = det*(e2*e4+3*e2*e5+e3*e4+e3*e5)+nu*(e4+3*e5); erg[0][0][4][1] = tau*(2*e4+2*e5); erg[0][0][4][0] = sigma*e3;
        erg[0][0][5][2] = det*(2*e2*e5+2*e3*e5)+nu*(2*e5); erg[0][0][5][1] = tau*(2*e4+2*e5);
        erg[0][0][6][2] = det*(e2*e5+e3*e5+e4*e5)+nu*(e5); erg[0][0][6][1] = tau*(2*e5); erg[0][0][6][0] = sigma*e4;
        erg[0][0][7][1] = 2*tau*e5;
        erg[0][0][8][0] = sigma*e5;
        
        erg[0][1][0][1] = tau;
        erg[0][1][1][2] = det*(e3+e4+e5); erg[0][1][1][0] = sigma;
        erg[0][1][2][2] = det*(e3+2*e4+2*e5);
        erg[0][1][3][2] = det*(2*e4+3*e5);
        erg[0][1][4][2] = det*(e4+3*e5);
        erg[0][1][5][2] = det*2*e5;
        erg[0][1][6][2] = det*e5;
        
        erg[0][2][0][1] = tau;
        erg[0][2][1][2] = det*(-e2); erg[0][2][1][1] = tau;
        erg[0][2][2][2] = det*(e4+e5); erg[0][2][2][0] = sigma;
        erg[0][2][3][2] = det*(e4+2*e5);
        erg[0][2][4][2] = det*(e4+2*e5);
        erg[0][2][5][2] = det*(2*e5);
        erg[0][2][6][2] = det*e5;
        
        erg[0][3][0][1] = tau;
        erg[0][3][1][2] = det*(-e2); erg[0][3][1][1] = tau;
        erg[0][3][2][2] = det*(-e2-e3); erg[0][3][2][1] = tau;
        erg[0][3][3][2] = det*(-e3+e5); erg[0][3][3][0] = sigma;
        erg[0][3][4][2] = det*(e5);
        erg[0][3][5][2] = det*(e5);
        erg[0][3][6][2] = det*(e5);
        
        erg[0][4][0][1] = tau; 
        erg[0][4][1][2] = det*(-e2); erg[0][4][1][1] = tau;
        erg[0][4][2][2] = det*(-e2-e3); erg[0][4][2][1] = tau;
        erg[0][4][3][2] = det*(-e2-2*e3-e4); erg[0][4][3][1] = tau;
        erg[0][4][4][2] = det*(-e3-e4); erg[0][4][4][0] = sigma;
        erg[0][4][5][2] = det*(-e4);
        
        erg[1][1][0][2] = det*(e1*e3+e1*e4+e1*e5)+nu*(e3+e4+e5); erg[1][1][0][0] = e1*sigma+1; 
        erg[1][1][1][2] = det*(2*e1*e3+2*e1*e4+2*e1*e5)+nu*(2*e3+2*e4+2*e5);
        erg[1][1][2][2] = det*(e1*e3+3*e1*e4+3*e1*e5)+nu*(e3+3*e4+3*e5); erg[1][1][2][1] = 2*tau*e3;
        erg[1][1][3][2] = det*(2*e1*e4+4*e1*e5)+nu*(2*e4+4*e5); erg[1][1][3][1] = tau*(2*e3+2*e4);
        erg[1][1][4][2] = det*(e1*e4+3*e1*e5+e3*e4+e3*e5)+nu*(e4+3*e5); erg[1][1][4][1] = tau*(2*e4+2*e5); erg[1][1][4][0] = e3*sigma;
        erg[1][1][5][2] = det*(2*e1*e5+2*e3*e5)+nu*(2*e5); erg[1][1][5][1] = tau*(2*e4+2*e5);
        erg[1][1][6][2] = det*(e1*e5+e3*e5+e4*e5)+nu*(e5); erg[1][1][6][1] = tau*(2*e5); erg[1][1][6][0] = sigma*e4;
        erg[1][1][7][1] = 2*tau*e5;
        erg[1][1][8][0] = sigma*e5;
        
        erg[1][2][0][2] = det*e1+nu;
        erg[1][2][1][2] = det*e1+nu; erg[1][2][1][1] = tau;
        erg[1][2][2][1] = 2*tau;
        erg[1][2][3][2] = det*(e4+e5); erg[1][2][3][0] = sigma;
        erg[1][2][4][2] = det*(e4+2*e5);
        erg[1][2][5][2] = det*2*e5;
        erg[1][2][6][2] = det*e5;

        erg[1][3][0][2] = det*e1+nu;
        erg[1][3][1][2] = det*e1+nu; erg[1][3][1][1] = tau;
        erg[1][3][2][2] = det*e1+nu; erg[1][3][2][1] = tau;
        erg[1][3][3][2] = det*(-e3); erg[1][3][3][1] = 2*tau; 
        erg[1][3][4][2] = det*e5; erg[1][3][4][0] = sigma;
        erg[1][3][5][2] = det*e5;
        erg[1][3][6][2] = det*e5;

        erg[1][4][0][2] = det*e1+nu; 
        erg[1][4][1][2] = det*e1+nu; erg[1][4][1][1] = tau;
        erg[1][4][2][2] = det*e1+nu; erg[1][4][2][1] = tau;
        erg[1][4][3][2] = det*(e1-e3)+nu; erg[1][4][3][1] = tau;
        erg[1][4][4][2] = det*(-e3-e4); erg[1][4][4][1] = 2*tau;
        erg[1][4][5][2] = det*(-e4); erg[1][4][5][0] = sigma;

        erg[2][2][0][2] = det*(e1*e2+e1*e4+e1*e5)+nu*(e2+e4+e5); erg[2][2][0][0] = e1*sigma+1; 
        erg[2][2][1][2] = det*(2*e1*e4+2*e1*e5)+nu*(2*e4+2*e5); erg[2][2][1][1] = 2*tau*e2;
        erg[2][2][2][2] = det*(3*e1*e4+3*e1*e5+e2*e4+e2*e5)+nu*(3*e4+3*e5); erg[2][2][2][0] = e2*sigma;
        erg[2][2][3][2] = det*(2*e1*e4+4*e1*e5+2*e2*e4+2*e2*e5)+nu*(2*e4+4*e5); erg[2][2][3][1] = tau*(2*e4);
        erg[2][2][4][2] = det*(e1*e4+3*e1*e5+e2*e4+3*e2*e5)+nu*(e4+3*e5); erg[2][2][4][1] = tau*(2*e4+2*e5);
        erg[2][2][5][2] = det*(2*e1*e5+2*e2*e5)+nu*(2*e5); erg[2][2][5][1] = tau*(2*e4+2*e5);
        erg[2][2][6][2] = det*(e1*e5+e2*e5+e4*e5)+nu*(e5); erg[2][2][6][1] = tau*(2*e5); erg[2][2][6][0] = sigma*e4;
        erg[2][2][7][1] = 2*tau*e5;
        erg[2][2][8][0] = sigma*e5;
        
        erg[2][3][0][2] = det*(e1)+nu;
        erg[2][3][1][2] = det*(2*e1)+2*nu;
        erg[2][3][2][2] = det*(2*e1+e2)+2*nu; erg[2][3][2][1] = tau;
        erg[2][3][3][2] = det*(e1+e2)+nu; erg[2][3][3][1] = 2*tau; 
        erg[2][3][4][1] = 2*tau;
        erg[2][3][5][2] = det*(e5); erg[2][3][5][0] = sigma;
        erg[2][3][6][2] = det*e5;
        
        erg[2][4][0][2] = det*(e1)+nu;
        erg[2][4][1][2] = det*(2*e1)+2*nu;
        erg[2][4][2][2] = det*(2*e1+e2)+2*nu; erg[2][4][2][1] = tau;
        erg[2][4][3][2] = det*(2*e1+e2)+2*nu; erg[2][4][3][1] = tau;
        erg[2][4][4][2] = det*(e1+e2)+nu; erg[2][4][4][1] = 2*tau;
        erg[2][4][5][2] = det*(-e4); erg[2][4][5][1] = 2*tau;
        erg[2][4][6][0] = sigma;

        erg[3][3][0][2] = det*(e1*e2+e1*e3+e1*e5)+nu*(e2+e3+e5); erg[3][3][0][0] = e1*sigma+1; 
        erg[3][3][1][2] = det*(2*e1*e3+2*e1*e5)+nu*(2*e3+2*e5); erg[3][3][1][1] = 2*tau*e2;
        erg[3][3][2][2] = det*(e1*e3+3*e1*e5+e2*e3+e2*e5)+nu*(e3+3*e5); erg[3][3][2][1] = 2*tau*e3; erg[3][3][2][0] = e2*sigma;
        erg[3][3][3][2] = det*(4*e1*e5+2*e2*e5)+nu*(4*e5); erg[3][3][3][1] = tau*(2*e3);
        erg[3][3][4][2] = det*(3*e1*e5+3*e2*e5+e3*e5)+nu*(3*e5); erg[3][3][4][1] = tau*(2*e5); erg[3][3][4][0] = sigma*e3;
        erg[3][3][5][2] = det*(2*e1*e5+2*e2*e5+2*e3*e5)+nu*(2*e5); erg[3][3][5][1] = tau*(2*e5);
        erg[3][3][6][2] = det*(e1*e5+e2*e5+e3*e5)+nu*(e5); erg[3][3][6][1] = tau*(2*e5);
        erg[3][3][7][1] = 2*tau*e5;
        erg[3][3][8][0] = sigma*e5;
        
        erg[3][4][0][2] = det*(e1)+nu;
        erg[3][4][1][2] = det*(2*e1)+2*nu;
        erg[3][4][2][2] = det*(3*e1+e2)+3*nu;
        erg[3][4][3][2] = det*(3*e1+2*e2)+3*nu; erg[3][4][3][1] = tau;
        erg[3][4][4][2] = det*(2*e1+2*e2+e3)+2*nu; erg[3][4][4][1] = 2*tau;
        erg[3][4][5][2] = det*(e1+e2+e3)+nu; erg[3][4][5][1] = 2*tau;
        erg[3][4][6][1] = 2*tau;
        erg[3][4][7][0] = sigma;

        erg[4][4][0][2] = det*(e1*e2+e1*e3+e1*e4)+nu*(e2+e3+e4); erg[4][4][0][0] = e1*sigma+1; 
        erg[4][4][1][2] = det*(2*e1*e3+2*e1*e4)+nu*(2*e3+2*e4); erg[4][4][1][1] = 2*tau*e2;
        erg[4][4][2][2] = det*(e1*e3+3*e1*e4+e2*e3+e2*e4)+nu*(e3+3*e4); erg[4][4][2][1] = 2*tau*e3; erg[4][4][2][0] = e2*sigma;
        erg[4][4][3][2] = det*(2*e1*e4+2*e2*e4)+nu*(2*e4); erg[4][4][3][1] = tau*(2*e3+2*e4);
        erg[4][4][4][2] = det*(e1*e4+e2*e4+e3*e4)+nu*(e4); erg[4][4][4][1] = tau*(2*e4); erg[4][4][4][0] = sigma*e3;
        erg[4][4][5][1] = 2*tau*e4;
        erg[4][4][6][0] = sigma*e4;
        
        for (int i=0; i<5; i++)
            for (int k=0; k<9; k++)
            {
                for (int j=i+1; j<5; j++)
                    for (int l=0; l<3; l++)
                        {erg[i][j][k][l] *= -error[i]*error[j]; erg[j][i][k][l] = erg[i][j][k][l];}
                for (int l=0; l<3; l++) erg[i][i][k][l] *= error[i];
            }
    }

    
    /**
     * Errors assumed to be equal in real data and hypothesis.
     * 
     * @param paramsReal
     * @param paramsHypo
     * @param N
     * @param trials
     * @param estimate can be 0 (none estimatet), 1 (beta estimated) or 2 (alpha estimated)
     */
    public static double simulatePower(double alpha, double beta, double sigma, double tau, double nu, double[] error, 
            double alphaH, double betaH, double sigmaH, double tauH, double nuH,  
            int df, int N, int trials, int estimate)
    {
        Random r = new Random();
        
        double[][] lambda = new double[5][2]; lambda[0][0] = 1; lambda[0][1] = 0;
        for (int i=1; i<5; i++) {lambda[i][0] = lambda[i-1][0]*beta; lambda[i][1] = lambda[i-1][1]*beta + alpha;}
        
        double det = numerDet(alpha, beta, sigma, tau, nu, error);
        double[][] sigmaAdj = new double[5][5];
        adjointE(sigmaAdj, alpha, beta, sigma, tau, nu, error);
        
        double detH = numerDet(alphaH, betaH, sigmaH, tauH, nuH, error);
        double[][] sigmaAdjH = new double[5][5];
        adjointE(sigmaAdjH, alphaH, betaH, sigmaH, tauH, nuH, error);
                
        double chol11 = Math.sqrt(sigma), chol12 = tau/chol11, chol22 = Math.sqrt(sigma*nu-tau*tau)/chol11;
        double[] errStdv = new double[5]; for (int i=0; i<5;  i++) errStdv[i] = Math.sqrt(1/error[i]);
        
        double[] data = new double[5];
        double[][] dataMoments = new double[5][5];
        
        // This part only if beta is supposed to be estimated.
        double eps = 0.001;
        double[][][] estimationMatrixBeta = new double[5][5][16], estimationMatrixAlpha = new double[5][5][5];
        double[] estimationPolBeta = new double[16], estimationPolAlpha = new double[5];
        double[] detCo = new double[9], detDerCo = new double[8], work = new double[8], work2 = new double[16], work3 = new double[16];
        double[][][][] coeffMatrix = new double[5][5][9][3];
        adjointEBothCoeffs(coeffMatrix, sigma, tau, nu, error);
        double[][] detCoeffs = new double[9][3];
        numerDetCoeffs(detCoeffs, sigma, tau, nu, error);
        if (estimate==1)
        {
            adjointECoeffsBeta(estimationMatrixBeta, alpha, sigma, tau, nu, error);
            numerDetCoeffsBeta(detCo, alpha, sigma, tau, nu, error);
            for (int i=0; i<8; i++) detDerCo[i] = (i+1)*detCo[i+1];
            for (int i=0; i<5; i++)
                for (int j=0; j<5; j++)
                {
                    for (int k=0; k<8; k++) work[k] = (k+1)*estimationMatrixBeta[i][j][k+1];
                    Statik.polynomialMultiply(estimationMatrixBeta[i][j], detDerCo, work2);
                    Statik.polynomialMultiply(work, detCo, work3);
                    for (int k=0; k<16; k++) estimationMatrixBeta[i][j][k] = - work2[k] + work3[k];
                }
            Statik.polynomialMultiply(detCo, detDerCo, estimationPolBeta);
            for (int i=0; i<16; i++) estimationPolBeta[i] *= N;
        }
        if (estimate==2)
        {
            for (int i=0; i<5; i++)
                for (int j=0; j<5; j++)
                    for (int k=0; k<3; k++) 
                    {
                        double bexp = 1; for (int l=0; l<9; l++) {estimationMatrixAlpha[i][j][k] += bexp * coeffMatrix[i][j][l][k]; bexp *= beta;}
                    }
            for (int k=0; k<3; k++) 
            {
                double bexp = 1; for (int l=0; l<9; l++) {detCo[k] += bexp * detCoeffs[l][k]; bexp *= beta;}
            }
            for (int i=0; i<2; i++) detDerCo[i] = (i+1)*detCo[i+1]; for (int i=2; i<detDerCo.length; i++) detDerCo[i] = 0;
            for (int i=0; i<5; i++)
                for (int j=0; j<5; j++)
                {
                    for (int k=0; k<2; k++) work[k] = (k+1)*estimationMatrixAlpha[i][j][k+1];
                    Statik.polynomialMultiply(estimationMatrixAlpha[i][j], detDerCo, work2);
                    Statik.polynomialMultiply(work, detCo, work3);
                    for (int k=0; k<5; k++) estimationMatrixAlpha[i][j][k] = - work2[k] + work3[k];
                }
            Statik.polynomialMultiply(detCo, detDerCo, estimationPolAlpha);
            for (int i=0; i<5; i++) estimationPolAlpha[i] *= N;
        }
        
        double[][] sigmaAdjE = new double[5][5];
        for (int i=0; i<5; i++)
            for (int j=0; j<5; j++) sigmaAdjE[i][j] = sigmaAdj[i][j];
        double detE = det;
        
        int count = 0; double aveVal = 0;
        lastParameterDistribution = new double[trials];
        for (int trial = 0; trial <trials; trial++)
        {
            for (int j=0; j<5; j++) 
                for (int k=0; k<5; k++) dataMoments[j][k] = 0;
            for (int i=0; i<N; i++)
            {
                double in1 = r.nextGaussian(), in2 = r.nextGaussian();
                double inter = in1*chol11, slope = in1*chol12 + in2*chol22;
                for (int j=0; j<5; j++) data[j] = (error[j]==0?0:inter*lambda[j][0] + slope*lambda[j][1] + r.nextGaussian() * errStdv[j]);
                for (int j=0; j<5; j++) 
                    for (int k=0; k<5; k++) dataMoments[j][k] += data[j]*data[k];
            }
            
            if (estimate==1)
            {
                // This part if beta is to be estimated.
                for (int i=0; i<16; i++) work2[i] = estimationPolBeta[i];
                for (int i=0; i<5; i++) 
                    for (int j=0; j<5; j++) 
                        for (int k=0; k<16; k++) work2[k] += dataMoments[i][j]*estimationMatrixBeta[i][j][k];
                double[] estBeta = Statik.solveReel(work2, eps);
                double betaE = (estBeta.length==0?betaH:estBeta[0]); 
                lastParameterDistribution[trial] = betaE;
                double minVal = Double.MAX_VALUE;
                if (estBeta.length>1)
                {
                    for (int i=0; i<estBeta.length; i++) 
                    {
                        adjointE(sigmaAdjE, alpha, estBeta[i], sigma, tau, nu, error);
                        detE = numerDet(alpha, estBeta[i], sigma, tau, nu, error);
                        // In theory, the value in the log should be divided by the products of inverse errors. Yet, since this part
                        // is only for comparing values, this part cancels out.
                        double val = N*Math.log(Statik.evaluate(detCo, estBeta[i]))*detE;
                        for (int j=0; j<5; j++) 
                            for (int k=0; k<5; k++)
                                val += dataMoments[j][k] * sigmaAdjE[j][k];
                        val /= detE;
                        if (val < minVal) {minVal = val; betaE = estBeta[i];}
                    }
                }
                adjointE(sigmaAdjE, alpha, betaE, sigma, tau, nu, error);
                detE = numerDet(alpha, betaE, sigma, tau, nu, error);
            }
            if (estimate==2)
            {
                // This part if alpha is to be estimated.
                for (int i=0; i<5; i++) work2[i] = estimationPolAlpha[i]; for (int i=5; i<work2.length; i++) work2[i] = 0;
                for (int i=0; i<5; i++) 
                    for (int j=0; j<5; j++) 
                        for (int k=0; k<5; k++) work2[k] += dataMoments[i][j]*estimationMatrixAlpha[i][j][k];
                double[] estAlpha = Statik.solveReel(work2, eps);
                double alphaE = (estAlpha.length==0?alphaH:estAlpha[0]);
                lastParameterDistribution[trial] = alphaE;
                double minVal = Double.MAX_VALUE;
                if (estAlpha.length>1)
                {
                    for (int i=0; i<estAlpha.length; i++) 
                    {
                        adjointE(sigmaAdjE, estAlpha[i], beta, sigma, tau, nu, error);
                        detE = numerDet(estAlpha[i], beta, sigma, tau, nu, error);
                        // In theory, the value in the log should be divided by the products of inverse errors. Yet, since this part
                        // is only for comparing values, this part cancels out.
                        double val = N*Math.log(Statik.evaluate(detCo, estAlpha[i]))*detE;
                        for (int j=0; j<5; j++) 
                            for (int k=0; k<5; k++)
                                val += dataMoments[j][k] * sigmaAdjE[j][k];
                        val /= detE;
                        if (val < minVal) {minVal = val; alphaE = estAlpha[i];}
                    }
                }
                adjointE(sigmaAdjE, alphaE, beta, sigma, tau, nu, error);
                detE = numerDet(alphaE, beta, sigma, tau, nu, error);
            }
            
            double logterm = N*Math.log(detH / detE);
            double dataterm = 0;
            for (int i=0; i<5; i++)
                for (int j=0; j<5; j++)
                    dataterm += dataMoments[i][j] * (sigmaAdjH[i][j] * detE - sigmaAdjE[i][j] * detH);
            
            double val = dataterm / (detE*detH) + logterm;
            if (val > Statik.FIVEPERCENTTHRESHOLD[df-1]) count++;
            aveVal += val;
        }
        System.out.print(aveVal/(double)trials+"\t");
        return (double)count / (double)trials;
    }
    
    public static double simulatePowerWithAttrition(double alpha, double beta, double sigma, double tau, double nu, double[][] error, 
            double alphaH, double betaH, double sigmaH, double tauH, double nuH,  
            int df, int[] N, int trials, int estimate)
    {
        double eps = 0.001;

        Random r = new Random();
        int anzPat = N.length;
        
        double[][] lambda = new double[5][2]; lambda[0][0] = 1; lambda[0][1] = 0;
        for (int i=1; i<5; i++) {lambda[i][0] = lambda[i-1][0]*beta; lambda[i][1] = lambda[i-1][1]*beta + alpha;}
        double chol11 = Math.sqrt(sigma), chol12 = tau/chol11, chol22 = Math.sqrt(sigma*nu-tau*tau)/chol11;
        double[] data = new double[5];
        double[][] errStdv = new double[anzPat][5]; 
        double[] work = new double[8], work2 = new double[16], work3 = new double[16];
        double[][][][] coeffMatrix = new double[5][5][9][3];
        double[][] detCoeffs = new double[9][3];
        
        double[] det = new double[anzPat], detH = new double[anzPat]; 
        double[][][] sigmaAdj = new double[anzPat][5][5], sigmaAdjH = new double[anzPat][5][5], dataMoments = new double[anzPat][5][5];
        double[][][][] estimationPolMatrix = new double[anzPat][5][5][16];
        double[][] estimationPolConstant = new double[anzPat][16];
        double[][] detCo = new double[anzPat][9], detDerCo = new double[anzPat][8];
        double[] errorprod = new double[anzPat];

        for (int pat = 0; pat < anzPat; pat++)
        {
            errorprod[pat] = 1;
            for (int i=0; i<5;  i++) {errStdv[pat][i] = Math.sqrt(1/error[pat][i]); errorprod[pat] *= error[pat][i];}
            
            det[pat] = numerDet(alpha, beta, sigma, tau, nu, error[pat]);
            adjointE(sigmaAdj[pat], alpha, beta, sigma, tau, nu, error[pat]);
            
            detH[pat] = numerDet(alphaH, betaH, sigmaH, tauH, nuH, error[pat]);
            adjointE(sigmaAdjH[pat], alphaH, betaH, sigmaH, tauH, nuH, error[pat]);
                    
            adjointEBothCoeffs(coeffMatrix, sigma, tau, nu, error[pat]);
            numerDetCoeffs(detCoeffs, sigma, tau, nu, error[pat]);
            if (estimate==1)
            {
                adjointECoeffsBeta(estimationPolMatrix[pat], alpha, sigma, tau, nu, error[pat]);
                numerDetCoeffsBeta(detCo[pat], alpha, sigma, tau, nu, error[pat]);
                for (int i=0; i<8; i++) detDerCo[pat][i] = (i+1)*detCo[pat][i+1];
                for (int i=0; i<5; i++)
                    for (int j=0; j<5; j++)
                    {
                        for (int k=0; k<8; k++) work[k] = (k+1)*estimationPolMatrix[pat][i][j][k+1];
                        Statik.polynomialMultiply(estimationPolMatrix[pat][i][j], detDerCo[pat], work2);
                        Statik.polynomialMultiply(work, detCo[pat], work3);
                        for (int k=0; k<16; k++) estimationPolMatrix[pat][i][j][k] = - work2[k] + work3[k];
                    }
                Statik.polynomialMultiply(detCo[pat], detDerCo[pat], estimationPolConstant[pat]);
                for (int i=0; i<16; i++) estimationPolConstant[pat][i] *= N[pat];
            }
            if (estimate==2)
            {
                for (int i=0; i<5; i++)
                    for (int j=0; j<5; j++)
                        for (int k=0; k<3; k++) 
                        {
                            double bexp = 1; for (int l=0; l<9; l++) {estimationPolMatrix[pat][i][j][k] += bexp * coeffMatrix[i][j][l][k]; bexp *= beta;}
                        }
                for (int k=0; k<3; k++) 
                {
                    double bexp = 1; for (int l=0; l<9; l++) {detCo[pat][k] += bexp * detCoeffs[l][k]; bexp *= beta;}
                }
                for (int i=0; i<2; i++) detDerCo[pat][i] = (i+1)*detCo[pat][i+1]; for (int i=2; i<detDerCo.length; i++) detDerCo[pat][i] = 0;
                for (int i=0; i<5; i++)
                    for (int j=0; j<5; j++)
                    {
                        for (int k=0; k<2; k++) work[k] = (k+1)*estimationPolMatrix[pat][i][j][k+1];
                        Statik.polynomialMultiply(estimationPolMatrix[pat][i][j], detDerCo[pat], work2);
                        Statik.polynomialMultiply(work, detCo[pat], work3);
                        for (int k=0; k<5; k++) estimationPolMatrix[pat][i][j][k] = - work2[k] + work3[k];
                    }
                Statik.polynomialMultiply(detCo[pat], detDerCo[pat], estimationPolConstant[pat]);
                for (int i=0; i<5; i++) estimationPolConstant[pat][i] *= N[pat];
            }
        }
        
        lastParameterDistribution = new double[trials];
        int count = 0;
        double mean2lldiff = 0;
        for (int trial = 0; trial <trials; trial++)
        {
            for (int i=0; i<work2.length; i++) work2[i] = 0;
            for (int pat=0; pat < anzPat; pat++)
            {
                for (int j=0; j<5; j++) 
                    for (int k=0; k<5; k++) dataMoments[pat][j][k] = 0;

                // debug only
//                r = new Random(1234567);
                
                for (int i=0; i<N[pat]; i++)
                {
                    double in1 = r.nextGaussian(), in2 = r.nextGaussian();
                    double inter = in1*chol11, slope = in1*chol12 + in2*chol22;
                    for (int j=0; j<5; j++) data[j] = (error[pat][j]==0?0:inter*lambda[j][0] + slope*lambda[j][1] + r.nextGaussian() * errStdv[pat][j]);
                    for (int j=0; j<5; j++) 
                        for (int k=0; k<5; k++) dataMoments[pat][j][k] += data[j]*data[k];
                }
                for (int i=0; i<16; i++) work3[i] = estimationPolConstant[pat][i];
                for (int i=0; i<5; i++) 
                    for (int j=0; j<5; j++) 
                        for (int k=0; k<16; k++) work3[k] += dataMoments[pat][i][j]*estimationPolMatrix[pat][i][j][k];
                // Approximation here: the division should be dependent on the parameter still to be found,
                // and the real value is taken instead.
                for (int i=0; i<16; i++) work2[i] += work3[i] / (det[pat]*det[pat]);
            }
            
//          debug:  double[] ests = Statik.solveReel(work3, eps);
            double[] ests = Statik.solveReel(work2, eps);
            double[][] sigmaAdjE = new double[5][5];
            double detE=0, alphaE = alpha, betaE = beta, best = 0; 
            if (ests.length>0) 
            {
                alphaE = (estimate==2?ests[0]:alpha); betaE = (estimate==1?ests[0]:beta); best = ests[0];
                lastParameterDistribution[trial] = ests[0];
            }
            double minVal = Double.MAX_VALUE;
            if ((ests.length>1) && (estimate!=0))
            {
                for (int i=0; i<ests.length; i++) 
                {
                    alphaE = (estimate==2?ests[i]:alphaH); betaE = (estimate==1?ests[i]:betaH);
                    double val = 0;
                    for (int pat=0; pat<anzPat; pat++)
                    {
                        adjointE(sigmaAdjE, alphaE, betaE, sigma, tau, nu, error[pat]);
                        detE = numerDet(alphaE, betaE, sigma, tau, nu, error[pat]);
                        // In theory, the value in the log should be divided by the products of inverse errors. Yet, since this part
                        // is only for comparing values, this part cancels out.
                        val += N[pat]*Math.log(Statik.evaluate(detCo[pat], ests[i]));
                        double sum = 0;
                        for (int j=0; j<5; j++) 
                            for (int k=0; k<5; k++)
                                sum += dataMoments[pat][j][k] * sigmaAdjE[j][k];
                        val += sum / detE;
                    }
                    if (val < minVal) {minVal = val; best = ests[i];}
                }
                alphaE = (estimate==2?best:alphaH); betaE = (estimate==1?best:betaH);
            }
            
            double val = 0;
            for (int pat=0; pat<anzPat; pat++)
            {
                adjointE(sigmaAdjE, alphaE, betaE, sigma, tau, nu, error[pat]);
                detE = numerDet(alphaE, betaE, sigma, tau, nu, error[pat]);
                double logterm = N[pat]*Math.log(detH[pat] / detE);
                double dataterm = 0;
                for (int i=0; i<5; i++)
                    for (int j=0; j<5; j++)
                        dataterm += dataMoments[pat][i][j] * (sigmaAdjH[pat][i][j] / detH[pat] - sigmaAdjE[i][j] / detE);
                val += dataterm + logterm;
            }
                
            if (val > Statik.FIVEPERCENTTHRESHOLD[df-1]) count++;
            mean2lldiff += val;
        }
        System.out.println("Mean difference = "+(mean2lldiff/(double)trials));
        return (double)count / (double)trials;
    }

    public static double approximatePowerWithAttrition(double alpha, double beta, double sigma, double tau, double nu, double[] error, 
            double alphaH, double betaH, double sigmaH, double tauH, double nuH,  
            int df, int N, int trials, int[][] pattern, double[] patProb, int estimate)
    {
        double eps = 0.001;

        Random r = new Random();
        int anzPat = pattern.length;
        
        double[][] lambda = new double[5][2]; lambda[0][0] = 1; lambda[0][1] = 0;
        for (int i=1; i<5; i++) {lambda[i][0] = lambda[i-1][0]*beta; lambda[i][1] = lambda[i-1][1]*beta + alpha;}
        double chol11 = Math.sqrt(sigma), chol12 = tau/chol11, chol22 = Math.sqrt(sigma*nu-tau*tau)/chol11;
        double[] data = new double[5];
        double[] errStdv = new double[5]; 
        double[] work = new double[8], work2 = new double[16], work3 = new double[16];
        double[][][][] coeffMatrix = new double[5][5][9][3], workCoeffMatrix = new double[5][5][9][3];
        double[][] detCoeffs = new double[9][3], workDetCoeffs = new double[9][3];
        
        double det = 0, detH = 0, detE, alphaE=0, betaE=0, best=0; 
        double[][] sigmaAdj = new double[5][5], sigmaAdjH = new double[5][5], dataMoments = new double[5][5];
        double[][][] estimationPolMatrix = new double[5][5][16];
        double[] estimationPolConstant = new double[16];
        double[] detCo = new double[9], detDerCo = new double[8];
        double[] ests = new double[0];
        double[][] sigmaAdjE = new double[5][5];

        for (int i=0; i<5; i++) errStdv[i] = Math.sqrt(1/error[i]);
        
        for (int i=0; i<5; i++)
            for (int j=0; j<5; j++)
                for (int k=0; k<9; k++)
                    for (int l=0; l<3; l++) coeffMatrix[i][j][k][l] = 0;
        
        for (int k=0; k<9; k++)
            for (int l=0; l<3; l++) detCoeffs[k][l] = 0;

        double[] e = new double[5];
        for (int pat = 0; pat < anzPat; pat++)
        {
            for (int i=0; i<5;  i++) e[i] = error[i] * pattern[pat][i];
            
            numerDetCoeffs(workDetCoeffs, sigma, tau, nu, e);
            adjointEBothCoeffs(workCoeffMatrix, sigma, tau, nu, e);
            
            for (int i=0; i<5; i++)
                for (int j=0; j<5; j++)
                    for (int k=0; k<9; k++)
                        for (int l=0; l<3; l++) coeffMatrix[i][j][k][l] += patProb[pat]*workCoeffMatrix[i][j][k][l];
            
            for (int k=0; k<9; k++)
                for (int l=0; l<3; l++) detCoeffs[k][l] += patProb[pat] * workDetCoeffs[k][l];
        }

        det = Statik.evaluate(detCoeffs, beta, alpha);
        for (int i=0; i<5; i++)
            for (int j=0; j<5; j++) sigmaAdj[i][j] = Statik.evaluate(coeffMatrix[i][j],beta,alpha);
        
        detH = Statik.evaluate(detCoeffs, betaH, alphaH);
        for (int i=0; i<5; i++)
            for (int j=0; j<5; j++) sigmaAdjH[i][j] = Statik.evaluate(coeffMatrix[i][j],betaH,alphaH);

        if (estimate==1)
        {
            for (int i=0; i<5; i++)
                for (int j=0; j<5; j++)
                    for (int k=0; k<9; k++) 
                    {
                        double aexp = 1; for (int l=0; l<3; l++) {estimationPolMatrix[i][j][k] += aexp * coeffMatrix[i][j][k][l]; aexp *= alpha;}
                    }
            for (int k=0; k<9; k++) 
            {
                double aexp = 1; for (int l=0; l<3; l++) {detCo[k] += aexp * detCoeffs[k][l]; aexp *= alpha;}
            }
        }
        if (estimate==2)
        {
            for (int i=0; i<5; i++)
                for (int j=0; j<5; j++)
                    for (int k=0; k<3; k++) 
                    {
                        double bexp = 1; for (int l=0; l<9; l++) {estimationPolMatrix[i][j][k] += bexp * coeffMatrix[i][j][l][k]; bexp *= beta;}
                    }
            for (int k=0; k<3; k++) 
            {
                double bexp = 1; for (int l=0; l<9; l++) {detCo[k] += bexp * detCoeffs[l][k]; bexp *= beta;}
            }
        }
        for (int i=0; i<8; i++) detDerCo[i] = (i+1)*detCo[i+1];
        for (int i=0; i<5; i++)
            for (int j=0; j<5; j++)
            {
                for (int k=0; k<8; k++) work[k] = (k+1)*estimationPolMatrix[i][j][k+1];
                Statik.polynomialMultiply(estimationPolMatrix[i][j], detDerCo, work2);
                Statik.polynomialMultiply(work, detCo, work3);
                for (int k=0; k<16; k++) estimationPolMatrix[i][j][k] = - work2[k] + work3[k];
            }
        Statik.polynomialMultiply(detCo, detDerCo, estimationPolConstant);
        for (int i=0; i<16; i++) estimationPolConstant[i] *= N;
        
        int count = 0;
        for (int trial = 0; trial <trials; trial++)
        {
            for (int j=0; j<5; j++) 
                for (int k=0; k<5; k++) dataMoments[j][k] = 0;
                
            // debug only
//            r = new Random(1234567);
            
            for (int i=0; i<N; i++)
            {
                double in1 = r.nextGaussian(), in2 = r.nextGaussian();
                double inter = in1*chol11, slope = in1*chol12 + in2*chol22;
                for (int j=0; j<5; j++) data[j] = (error[j]==0?0:inter*lambda[j][0] + slope*lambda[j][1] + r.nextGaussian() * errStdv[j]);
                for (int j=0; j<5; j++) 
                    for (int k=0; k<5; k++) dataMoments[j][k] += data[j]*data[k];
            }
            if (estimate != 0)
            {
                for (int i=0; i<16; i++) work2[i] = estimationPolConstant[i];
                for (int i=0; i<5; i++) 
                    for (int j=0; j<5; j++) 
                        for (int k=0; k<16; k++) work2[k] += dataMoments[i][j]*estimationPolMatrix[i][j][k];
                for (int i=0; i<16; i++) work2[i] /= (det*det);
                
                ests = Statik.solveReel(work2, eps);
                detE=0; alphaE = alpha; betaE = beta; best = 0; 
                if (ests.length>0) 
                {
                    alphaE = (estimate==2?ests[0]:alpha); betaE = (estimate==1?ests[0]:beta); best = ests[0];
                }
                else System.out.println("Warning: No zero found.");
            } else {alphaE = alpha; betaE = beta;}
            double minVal = Double.MAX_VALUE;
            if ((ests.length>1) && (estimate!=0))
            {
                for (int i=0; i<ests.length; i++) 
                {
                    alphaE = (estimate==2?ests[i]:alpha); betaE = (estimate==1?ests[i]:beta);
                    detE = Statik.evaluate(detCoeffs, betaE, alphaE);
                    for (int k=0; k<5; k++)
                        for (int j=0; j<5; j++) sigmaAdjE[k][j] = Statik.evaluate(coeffMatrix[k][j],betaE,alphaE);
                    // In theory, the value in the log should be divided by the products of inverse errors. Yet, since this part
                    // is only for comparing values, this cancels out.
                    double val = N*Math.log(Statik.evaluate(detCo, ests[i]));
                    double sum = 0;
                    for (int j=0; j<5; j++) 
                        for (int k=0; k<5; k++)
                            sum += dataMoments[j][k] * sigmaAdjE[j][k];
                    val += sum / detE;

                    if (val < minVal) {minVal = val; best = ests[i];}
                }
                alphaE = (estimate==2?best:alphaH); betaE = (estimate==1?best:betaH);
            }
            
            double val = 0;
            detE = Statik.evaluate(detCoeffs, betaE, alphaE);
            for (int k=0; k<5; k++)
                for (int j=0; j<5; j++) sigmaAdjE[k][j] = Statik.evaluate(coeffMatrix[k][j],betaE,alphaE);

            double logterm = N*Math.log(detH / detE);
            double dataterm = 0;
            for (int i=0; i<5; i++)
                for (int j=0; j<5; j++)
                    dataterm += dataMoments[i][j] * (sigmaAdjH[i][j] / detH - sigmaAdjE[i][j] / detE);
            val += dataterm + logterm;
                
            if (val > Statik.FIVEPERCENTTHRESHOLD[df-1]) count++;
        }
        return (double)count / (double)trials;
    }
        
    public static void simulationNoAttrition()
    {
        double[][] power = new double[100][2];
        int trials = 10000;
        int N = 100;
        double alpha = 1, beta = 2;
        double sigma = 1, tau = 0, nu = 1;
        double[] error = new double[]{4,4,4,4,4};

        long time = System.currentTimeMillis();
        System.out.println("I start.");
        
        beta = 2;
        for (int aIx = 0; aIx<100; aIx++)
        {
            alpha = aIx/(double)200;
            System.out.println("working on alpha = "+alpha+", so far "+(System.currentTimeMillis()-time)+" milliseconds needed.");
            power[aIx][0] = alpha;
            power[aIx][1] = simulatePower(alpha, beta, sigma, tau, nu, error, 0.0, beta, sigma, tau, nu, 1, N, trials, 2);
        }
        Statik.writeMatrix(power, "PowerToRealizeLinear.txt", '\t');
        
        System.out.println("I start beta simulation.");
        alpha = 1;
        for (int bIx = 0; bIx<100; bIx++)
        {
            beta = 1.0 + bIx/(double)400;
            System.out.println("working on beta = "+beta+", so far "+(System.currentTimeMillis()-time)+" milliseconds needed.");
            power[bIx][0] = beta;
            power[bIx][1] = simulatePower(alpha, beta, sigma, tau, nu, error, alpha, 1.0, sigma, tau, nu, 1, N, trials, 1);
        }
        Statik.writeMatrix(power, "PowerToRealizeExponential.txt", '\t');
        
        System.out.println("I finished, took me "+(System.currentTimeMillis()-time)+" milliseconds.");
    }
    
    public static void simulationMovingReliability()
    {
        double[][] power = new double[50][5];
        int trials = 10000;
        int N = 100;
        double alpha = 1, beta = 2;
        double sigma = 1, tau = 0, nu = 1;
        double[] error = new double[]{4,4,4,4,4};

        long time = System.currentTimeMillis();
        System.out.println("Starting on Moving Reliabilities");
        
        System.out.println("I start beta simulation.");
        alpha = 1;
        for (int p=0; p<5; p++)
        {
            error[p] = 0.1;
            for (int bIx = 0; bIx<50; bIx++)
            {
                beta = 1.0 + bIx/(double)400;
                System.out.println("working on beta = "+beta+", so far "+(System.currentTimeMillis()-time)+" milliseconds needed.");
                power[bIx][p] = simulatePower(alpha, beta, sigma, tau, nu, error, alpha, 1.0, sigma, tau, nu, 1, N, trials, 1);
            }
            error[p] = 4;
        }
        Statik.writeMatrix(power, "PowerMovingReliability.txt", '\t');
        
        System.out.println("I finished, took me "+(System.currentTimeMillis()-time)+" milliseconds.");
    }
    
    // Erstellt Pattern fr unabhngige random attrition
    public static void generatePatterns(int[][] ergPat, double[] ergProb, double[] probIn)
    {
        int[] count = new int[]{0,0,0,0};
        boolean stop = false;
        int c = 0, i=0;
        while (!stop)
        {
            for (i=0; i<4; i++) ergPat[c][i+1] = count[i];
            ergPat[c][0] = 1;
            ergProb[c] = 1;
            for (i=0; i<4; i++) ergProb[c] *= (count[i]==1?1-probIn[i]:probIn[i]);
            i=0; while ((i<4) && (count[i]==1)) {count[i] = 0; i++;}
            if (i<4) count[i] = 1; else stop = true;
            c++;
        }
    }

    // TODO
    public static void simulationAttrition()
    {
        double[][] patProb = new double[5][16];
        int[][] pat = new int[16][5];
        generatePatterns(pat, patProb[0], new double[]{0,0,0,0});
        generatePatterns(pat, patProb[1], new double[]{0.01,0.02,0.03,0.04});
        generatePatterns(pat, patProb[2], new double[]{0.05,0.1,0.15,0.2});
        generatePatterns(pat, patProb[3], new double[]{0.1,0.2,0.3,0.4});
        generatePatterns(pat, patProb[4], new double[]{0.2,0.4,0.6,0.8});

        double[][] power = new double[50][5];
        int trials = 10000;
        int N = 100;
        double alpha = 1, beta = 2;
        double sigma = 1, tau = 0, nu = 1;
        double[] error = new double[]{4,4,4,4,4};

        long time = System.currentTimeMillis();
        System.out.println("Starting Simulation Attrition");
        
        System.out.println("I start beta simulation.");
        alpha = 1;
        for (int p=0; p<5; p++)
        {
            for (int bIx = 0; bIx<50; bIx++)
            {
                beta = 1.0 + bIx/(double)400;
                System.out.println("working on beta = "+beta+", so far "+(System.currentTimeMillis()-time)+" milliseconds needed.");
                power[bIx][p] = approximatePowerWithAttrition(alpha, beta, sigma, tau, nu, error, alpha, 1.0, sigma, tau, nu, 1, N, trials, pat, patProb[p], 1);
            }
        }
        Statik.writeMatrix(power, "PowerDifferentAttritions.txt", '\t');
        
        System.out.println("I finished, took me "+(System.currentTimeMillis()-time)+" milliseconds.");
    }

    public static void simulationPlannedAttrition()
    {
        int[][] pat1 = new int[][]{{0,1,2,3,4,5,6,7},{2,3,4,5,6,7,8,9}};
        double[] patProb1 = new double[2]; int N1 = 100*5/4;
        int[][] pat2 = new int[][]{{0,1,2,3,4,5},{2,3,4,5,6,7},{4,5,6,7,8,9}};
        double[] patProb2 = new double[3]; int N2 = 100*5/3;
        int[][] pat3 = new int[][]{{0,1,2,3},{2,3,4,5},{4,5,6,7},{6,7,8,9}};
        double[] patProb3 = new double[4]; int N3 = 100*5/2;
        
        double[] power1 = new double[11];
        double[] power2 = new double[11];
        double[] power3 = new double[11];
        double beta = 1.05;
        double gamma = 0.1;
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};

        long time = System.currentTimeMillis();
        System.out.println("Starting Simulation Attrition");

        for (int i=0; i<11; i++)
        {
            for (int j=0; j<2; j++) patProb1[j] = (i*(50.0/5.0) + j*(5-i)*(100.0/5.0))/100.0;
            for (int j=0; j<3; j++) patProb2[j] = (i*(100.0/15.0) + j*(5-i)*(100.0/15.0))/100.0;
            for (int j=0; j<4; j++) patProb3[j] = (i*5 + j*(5-i)*25.0/7.5)/100.0;
            power1[i] = sarisSatorraPowerAttrition(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0}, err,
                    beta, beta, gamma, 0.0  , facCov, new double[]{0,0,0,0}, err, 
                    pat1, patProb1, 10, N1, 1);
            power2[i] = sarisSatorraPowerAttrition(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0}, err,
                    beta, beta, gamma, 0.0  , facCov, new double[]{0,0,0,0}, err, 
                    pat2, patProb2, 10, N2, 1);
            power3[i] = sarisSatorraPowerAttrition(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0}, err,
                    beta, beta, gamma, 0.0  , facCov, new double[]{0,0,0,0}, err, 
                    pat3, patProb3, 10, N3, 1);
        }
        System.out.println(sarisSatorraPowerAttrition(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0}, err,
                beta, beta, gamma, 0.0  , facCov, new double[]{0,0,0,0}, err, 
                new int[][]{{0,1,2,3,4,5,6,7,8,9}}, new double[]{1.0}, 10, 100, 1));
        Statik.writeMatrix(power1, "powerPlannedAttritions2Pattern.txt");
        Statik.writeMatrix(power2, "powerPlannedAttritions3Pattern.txt");
        Statik.writeMatrix(power3, "powerPlannedAttritions4Pattern.txt");
        
        System.out.println("I finished, took me "+(System.currentTimeMillis()-time)+" milliseconds.");
    }
    
    public static void simulationPlannedAttritionMC()
    {
        int[][] pat1 = new int[][]{{0,1,2,3,4,5,6,7},{2,3,4,5,6,7,8,9}};
        double[] patProb1 = new double[2]; int N1 = 100*5/4;
        int[][] pat2 = new int[][]{{0,1,2,3,4,5},{2,3,4,5,6,7},{4,5,6,7,8,9}};
        double[] patProb2 = new double[3]; int N2 = 100*5/3;
        int[][] pat3 = new int[][]{{0,1,2,3},{2,3,4,5},{4,5,6,7},{6,7,8,9}};
        double[] patProb3 = new double[4]; int N3 = 100*5/2;
        
        double[] power1 = new double[11];
        double[] power2 = new double[11];
        double[] power3 = new double[11];
        double beta = 1.05;
        double gamma = 0.1;
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};
        int trials = 100000;

        long time = System.currentTimeMillis();
        System.out.println("Starting Simulation Attrition");

        for (int i=0; i<11; i++)
        {
            for (int j=0; j<2; j++) patProb1[j] = (i*(50.0/5.0) + j*(5-i)*(100.0/5.0))/100.0;
            for (int j=0; j<3; j++) patProb2[j] = (i*(100.0/15.0) + j*(5-i)*(100.0/15.0))/100.0;
            for (int j=0; j<4; j++) patProb3[j] = (i*5 + j*(5-i)*25.0/7.5)/100.0;
            power1[i] = parameterPowerAttritionFixedPattern(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0},err,
                    0.0, N1, trials, 1, pat1, patProb1, DCSM5.GAMMAY);
            power2[i] = parameterPowerAttritionFixedPattern(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0},err,
                    0.0, N2, trials, 1, pat2, patProb2, DCSM5.GAMMAY);
            power3[i] = parameterPowerAttritionFixedPattern(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0},err,
                    0.0, N3, trials, 1, pat3, patProb3, DCSM5.GAMMAY);
            
            /*
            power1[i] = sarisSatorraPowerAttrition(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0}, err,
                    beta, beta, gamma, 0.0  , facCov, new double[]{0,0,0,0}, err, 
                    pat1, patProb1, 10, N1, 1);
            power2[i] = sarisSatorraPowerAttrition(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0}, err,
                    beta, beta, gamma, 0.0  , facCov, new double[]{0,0,0,0}, err, 
                    pat2, patProb2, 10, N2, 1);
            power3[i] = sarisSatorraPowerAttrition(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0}, err,
                    beta, beta, gamma, 0.0  , facCov, new double[]{0,0,0,0}, err, 
                    pat3, patProb3, 10, N3, 1);
                    */
        }
        System.out.println(parameterPowerAttritionFixedPattern(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0}, err,
                0.0, 100, trials, 1, new int[][]{{0,1,2,3,4,5,6,7,8,9}}, new double[]{1.0}, DCSM5.GAMMAY));
        Statik.writeMatrix(power1, "powerPlannedAttritions2Pattern.txt");
        Statik.writeMatrix(power2, "powerPlannedAttritions3Pattern.txt");
        Statik.writeMatrix(power3, "powerPlannedAttritions4Pattern.txt");
        
        System.out.println("I finished, took me "+(System.currentTimeMillis()-time)+" milliseconds.");
    }

    /**
     *  Simulates unequal attrition on both variables, assuming that each participant assumes one of three possibilities:
     *  Participating in variable one, participating in variable two, or participating in both, or not participating in 
     *  the study.
     */
    public static void simulationUnequalAttrition()
    {
        int[][] pat = new int[][]{{0,1,2,3,4,5,6,7,8,9},
                {1,3,5,7,9},{0,3,5,7,9}};
        double[] patProb = new double[3]; 
        
        double[][] power = new double[26][26];
        double beta = 1.05;
        double gamma = 0.1;
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};
        int N = 100;

        long time = System.currentTimeMillis();
        System.out.println("Starting Simulation Unequal Attrition");

        for (int i=0; i<26; i++)
            for (int j=0; j<26; j++)
            {
                System.out.println("i = "+i+", j = "+j);
                double p = i*0.02, q = j*0.02;
                patProb[0] = (1-p)*(1-q); 
                patProb[1] = p*(1-q);
                patProb[2] = q*(1-p);
                int Noff = (int)Math.round(p*q*N);
                
                // very obviously, this needs a simulation of data. 
                power[i][j] = parameterPowerAttritionFixedPattern(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0},err,
                          0.0, N-Noff, 10000, 1, pat, patProb, DCSM5.GAMMAY);
                //                power[i][j] = parameterPowerAttrition(beta, beta, gamma, gamma, facCov, new double[]{0,0,0,0}, err, 
//                        0.0, N, 1000, 1, double meanReturnProb, double returnProbIncreasePerStdv, int baseNonrandomAttritionOn, int parameter)
                
            }
        Statik.writeMatrix(power, "powerUnequalAttrition.txt", '\t');
        
        System.out.println("I finished, took me "+(System.currentTimeMillis()-time)+" milliseconds.");
    }

    public static void ssPowers()
    {
        double[][][] data = new double[4][100][100];
        double[][] id = new double[][]{{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
        for (int gx=0; gx<100; gx++)
        {
            System.out.println(gx+"% done.");
            for (int gy=0; gy<100; gy++)
            {
                double gaX = gx/(double)1000, gaY = gy/(double)1000;
                data[0][gx][gy] = sarisSatorraPower(1.0, 1.0, gaX, gaY, id, new double[]{0,0,0,0}, new double[]{0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25}, 
                        1.0, 1.0, 0.0, 0.0, id, new double[]{0,0,0,0}, new double[]{0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25}, 10, 100, 2);
                data[1][gx][gy] = sarisSatorraPower(1.0, 1.0, gaX, gaY, id, new double[]{0,0,0,0}, new double[]{0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25}, 
                        1.0, 1.0, 0.1, 0.0, id, new double[]{0,0,0,0}, new double[]{0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25}, 10, 100, 2);
                data[2][gx][gy] = sarisSatorraPower(1.0, 1.0, gaX, gaY, id, new double[]{0,0,0,0}, new double[]{0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25}, 
                        1.0, 1.0, 0.0, 0.1, id, new double[]{0,0,0,0}, new double[]{0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25}, 10, 100, 2);
                data[3][gx][gy] = sarisSatorraPower(1.0, 1.0, gaX, gaY, id, new double[]{0,0,0,0}, new double[]{0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25}, 
                        1.0, 1.0, 0.1, 0.1, id, new double[]{0,0,0,0}, new double[]{0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25}, 10, 100, 2);
            }
        }
        Statik.writeMatrix(data[0], "powerGamma00.txt", '\t');
        Statik.writeMatrix(data[1], "powerGamma01.txt", '\t');
        Statik.writeMatrix(data[2], "powerGamma10.txt", '\t');
        Statik.writeMatrix(data[3], "powerGamma11.txt", '\t');
    }
    
    public static void alphaVsBeta()
    {
        double[][] id = new double[][]{{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
        double[][] erg = new double[51][51];
        for (int a=0; a<51; a++)
            for (int b=0; b<51; b++)
            {
                erg[a][b] = sarisSatorraPower(b*0.04, 1.0, 0.0, 0.0, id, new double[]{0,-0.5+a*0.02,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 
                                              b*0.04, 1.0, 0.0, 0.0, id, new double[]{0,0,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 10, 100, 1);
            }
        Statik.writeMatrix(erg, "powerAlphaVsBeta.txt", '\t');
    }

    public static void betaMovingCorrelation()
    {
        double[][] id = new double[][]{{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
        double[][] erg = new double[51][2];
        for (int a=0; a<51; a++)
        {
            id[0][1] = id[1][0] = -1+0.04*a;
            erg[a][0] = sarisSatorraPower(1.12, 1.0, 0.0, 0.0, id, new double[]{0,0,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 
                                         1, 1.0, 0.0, 0.0, id, new double[]{0,0,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 10, 100, 1);
            erg[a][1] = sarisSatorraPower(0.88, 1.0, 0.0, 0.0, id, new double[]{0,0,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 
                                         1, 1.0, 0.0, 0.0, id, new double[]{0,0,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 10, 100, 1);
        }
        Statik.writeMatrix(erg, "powerBetaOnCov.txt",'\t');
    }

    public static void gammaMovingBothErrors()
    {
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};
        double[][] erg = new double[51][51];
        for (int a=0; a<51; a++)
            for (int b=0; b<51; b++)
            {
                for (int i=0; i<5; i++) {err[2*i] = 0.1+a*0.1; err[2*i+1] = 0.1+b*0.1;}
                erg[a][b] = sarisSatorraPower(1.05, 1.05, 0.1, 0.1, facCov, new double[]{0,0,0,0}, err, 
                                              1.05, 1.05, 0.1, 0.0, facCov, new double[]{0,0,0,0}, err, 10, 100, 1);
            }
          Statik.writeMatrix(erg, "powerGammaInhomogenousErrors.txt", '\t');
    }
    
    public static void gammaOnLevelVariance()
    {
        double[][] id = new double[][]{{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
        double[] erg = new double[51];
        for (int a=0; a<51; a++)
        {
            id[1][1] = 0+a*0.01;
            erg[a] = gammaPower(0.0, 0.0, 0.0, 0.0, id, new double[]{0,1,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 
                                   0.1, 200, 10000, 1);
            System.out.println(erg[a]);
        }
        Statik.writeMatrix(erg, "powerGamma.txt");
    }
    
    public static void gammaMovingAttrition(int start, int end)
    {
        long startTime = System.currentTimeMillis();
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};
        double[][] erg1 = new double[51][51];
        double[][] erg2 = new double[51][51];
        double[][] erg3 = new double[51][51];
        for (int a=start; a<=end; a++) {
            System.out.println("Starting a = "+a+", so far used "+(System.currentTimeMillis()-startTime)+" milliseconds.");
            for (int b=0; b<51; b++)
            {
                erg1[a][b] = parameterPowerAttrition(1.05, 1.05, 0.1, 0.1, facCov, new double[]{0,0,0,0}, err, 
                        0.0, 100, 5000, 1, 0.7+0.25*a/50, 0.2*b/50, ONREAL, GAMMAY);
                erg2[a][b] = parameterPowerAttrition(1.05, 1.05, 0.1, 0.1, facCov, new double[]{0,0,0,0}, err, 
                        0.0, 100, 5000, 1, 0.7+0.25*a/50, 0.2*b/50, ONLATENT, GAMMAY);
                erg3[a][b] = parameterPowerAttrition(1.05, 1.05, 0.1, 0.1, facCov, new double[]{0,0,0,0}, err, 
                        0.0, 100, 5000, 1, 0.7+0.25*a/50, 0.2*b/50, ONCHANGE, GAMMAY);
            }
            Statik.writeMatrix(erg1, "powerGammaOnReal"+start+"to"+end+".txt", '\t');
            Statik.writeMatrix(erg2, "powerGammaOnLatent"+start+"to"+end+".txt", '\t');
            Statik.writeMatrix(erg3, "powerGammaOnChange"+start+"to"+end+".txt", '\t');
        }
        System.out.println("I finished, took me "+(System.currentTimeMillis()-startTime)+" milliseconds.");
    }
    
    public static void gammaTableValuesOnAttrition()
    {
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};
        
        for (int a = 1; a<2; a++)
            for (int b =0 ;b<3; b++)
            {
                double ranAtt = (a==0?0.95:0.875);
                double nonranAtt = (b==0?0:(b==1?0.1:0.2));
                double power = parameterPowerAttrition(1.05, 1.05, 0.1, 0.1, facCov, new double[]{0,0,0,0}, err, 
                        0.0, 100, 100000, 1, ranAtt, nonranAtt, ONLATENT, GAMMAY);
                System.out.println(ranAtt+" & "+Statik.doubleNStellen(nonranAtt,3)+" & "+Statik.doubleNStellen(Statik.mean(lastParameterDistribution),3)+
                        " & "+Statik.doubleNStellen(Statik.stdv(lastParameterDistribution),3)+
                        " & ["+Statik.doubleNStellen(Statik.percentile(lastParameterDistribution, 0.025),3)+";"+Statik.doubleNStellen(Statik.percentile(lastParameterDistribution, 0.975),3)+"]"+
                        " & "+Statik.doubleNStellen(Statik.mean(lastLLRDistribution),3)+
                        " & ["+Statik.doubleNStellen(Statik.percentile(lastLLRDistribution, 0.025),3)+";"+Statik.doubleNStellen(Statik.percentile(lastLLRDistribution, 0.975),3)+"]"+
                        " & "+Statik.doubleNStellen(power, 3)+"\\\\");
            }
    }
    
    public static void gammaNullMovingNonrandomAttrition()
    {
        long startTime = System.currentTimeMillis();
        double[][] id = new double[][]{{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
        double[] erg = new double[51];
        double[] err = new double[]{1,1,1,1,1,1,1,1,1,1};
        for (int a=0; a<51; a++) {
            System.out.println("Starting a = "+a+", so far used "+(System.currentTimeMillis()-startTime)+" milliseconds.");
            erg[a] = parameterPowerAttrition(1.0, 1.0, 0.0, 0.0, id, new double[]{0,0,0,0}, err, 
                    0.0, 100, 10000, 1, 0.7, 0.4*a/50, ONREAL, GAMMAY);
            Statik.writeMatrix(erg, "powerGammaNullOnNonrandom.txt");
        }
        System.out.println("I finished, took me "+(System.currentTimeMillis()-startTime)+" milliseconds.");
    }

    public static void gammaNullMovingNonrandomAttritionAndBias()
    {
        int trials = 10000;
        long startTime = System.currentTimeMillis();
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};
        double[][] erg = new double[51][6];
        for (int a=0; a<51; a++) {
            System.out.println("Starting a = "+a+", so far used "+(System.currentTimeMillis()-startTime)+" milliseconds.");
            erg[a][0] = parameterPowerAttrition(1.05, 1.05, 0.1, 0.0, facCov, new double[]{0,0,0,0}, err, 
                    0.0, 100, trials, 1, 0.7, 0.4*a/50, ONREAL, GAMMAY);
            erg[a][3] = LASTBIAS;
            erg[a][1] = parameterPowerAttrition(1.05, 1.05, 0.1, 0.0, facCov, new double[]{0,0,0,0}, err, 
                    0.0, 100, trials, 1, 0.7, 0.4*a/50, ONLATENT, GAMMAY);
            erg[a][4] = LASTBIAS;
            erg[a][2] = parameterPowerAttrition(1.05, 1.05, 0.1, 0.0, facCov, new double[]{0,0,0,0}, err, 
                    0.0, 100, trials, 1, 0.7, 0.4*a/50, ONCHANGE, GAMMAY);
            erg[a][5] = LASTBIAS;
            Statik.writeMatrix(erg, "powerGammaNullOnNonrandomAndBias.txt", '\t');
        }
        System.out.println("I finished, took me "+(System.currentTimeMillis()-startTime)+" milliseconds.");
    }

    public static void betaNullMovingNonrandomAttritionAndBias()
    {
        int trials = 500;
        long startTime = System.currentTimeMillis();
        double[][] id = new double[][]{{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
        double[][] erg = new double[51][6];
        double[] err = new double[]{1,1,1,1,1,1,1,1,1,1};
        for (int a=0; a<51; a++) {
            System.out.println("Starting a = "+a+", so far used "+(System.currentTimeMillis()-startTime)+" milliseconds.");
            erg[a][0] = parameterPowerAttrition(1.1, 1.0, 0.1, 0.1, id, new double[]{0,0,0,0}, err, 
                    1.1, 100, trials, 1, 0.7, 0.4*a/50, ONREAL, BETAX);
            erg[a][3] = LASTBIAS;
            erg[a][1] = parameterPowerAttrition(1.1, 1.0, 0.1, 0.1, id, new double[]{0,0,0,0}, err, 
                    1.1, 100, trials, 1, 0.7, 0.4*a/50, ONLATENT, BETAX);
            erg[a][4] = LASTBIAS;
            erg[a][2] = parameterPowerAttrition(1.1, 1.0, 0.1, 0.1, id, new double[]{0,0,0,0}, err, 
                    1.1, 100, trials, 1, 0.7, 0.4*a/50, ONCHANGE, BETAX);
            erg[a][5] = LASTBIAS;
            Statik.writeMatrix(erg, "powerBetaNullOnNonrandomAndBias.txt", '\t');
        }
        System.out.println("I finished, took me "+(System.currentTimeMillis()-startTime)+" milliseconds.");
    }

    public static void betaNullMovingAttritionJustBias()
    {
        int trials = 500;
        long startTime = System.currentTimeMillis();
        double[][] id = new double[][]{{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
        double[][] erg = new double[51][6];
        double[] err = new double[]{1,1,1,1,1,1,1,1,1,1};
        for (int a=0; a<21; a++) {
            System.out.println("Starting a = "+a+", so far used "+(System.currentTimeMillis()-startTime)+" milliseconds.");
            parameterPowerAttrition(1.1, 1.0, 0.1, 0.1, id, new double[]{0,0,0,0}, err, 
                    1.1, 100+a*100, trials, 1, 1.0, 0, ONREAL, BETAX);
            erg[a][0] = LASTBIAS;
            parameterPowerAttrition(1.1, 1.0, 0.1, 0.1, id, new double[]{0,0,0,0}, err, 
                    1.1, 100+a*100, trials, 1, 0.8, 0, ONREAL, BETAX);
            erg[a][1] = LASTBIAS;
            parameterPowerAttrition(1.1, 1.0, 0.1, 0.1, id, new double[]{0,0,0,0}, err, 
                    1.1, 100+a*100, trials, 1, 0.6, 0, ONREAL, BETAX);
            erg[a][2] = LASTBIAS;
            Statik.writeMatrix(erg, "biasBetaOnParticipants.txt", '\t');
        }
        System.out.println("I finished, took me "+(System.currentTimeMillis()-startTime)+" milliseconds.");
    }

    public static void gammaMovingError()
    {
        long startTime = System.currentTimeMillis();
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};
        double[] erg = new double[51];
        for (int a=0; a<51; a++) {
            for (int i=0; i<10; i++) err[i] = (1.0 / (0.5+(double)a/(double)100)) - 0.999;
            erg[a] = sarisSatorraPower(1.05, 1.05, 0.1, 0.1, facCov, new double[]{0,0,0,0}, err, 
                                       1.05, 1.05, 0.1, 0.0, facCov, new double[]{0,0,0,0}, err, 10, 100, 1);
        }
        Statik.writeMatrix(erg, "powerGammaOnReliability.txt");
        System.out.println("I finished, took me "+(System.currentTimeMillis()-startTime)+" milliseconds.");
    }

    public static void gammaMovingErrorDistribution()
    {
        long startTime = System.currentTimeMillis();
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};
        double[] erg = new double[51];
        for (int a=0; a<51; a++) {
            double relSlope = ((double)(2*a-50)/100.0)/4.0;
            for (int i=0; i<10; i++) {double rel = 0.65+relSlope*(i/2-2); err[i] = (1.0 / rel) - 0.999;}
            erg[a] = sarisSatorraPower(1.05, 1.05, 0.1, 0.1, facCov, new double[]{0,0,0,0}, err, 
                                       1.05, 1.05, 0.1, 0.0, facCov, new double[]{0,0,0,0}, err, 10, 100, 1);
        }
        Statik.writeMatrix(erg, "powerGammaOnErrorDistribution.txt");
        System.out.println("I finished, took me "+(System.currentTimeMillis()-startTime)+" milliseconds.");
    }

    public static void checkSSvsMCBeta()
    {
        long startTime = System.currentTimeMillis();
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};
        double[][] erg = new double[51][2];

        for (int a=0; a<51; a++) {
            System.out.println("Starting a = "+a+", so far used "+(System.currentTimeMillis()-startTime)+" milliseconds.");
            erg[a][0] = simulatePower( 1, 1 + a*0.2/50,                 1, 0, 0.25, new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0}, 
                                       1, 1,                            1, 0, 0.25, 1, 100, 1000, 1);
            erg[a][1] = sarisSatorraPower(1 + a*0.2/50, 1.0, 0.0, 0.0, facCov, new double[]{0,0.0,0,0}, err, 
                                          1.0         , 1.0, 0.0, 0.0, facCov, new double[]{0,0.0,0,0}, err, 10, 100, 1);
        }
        System.out.println("I finished, took me "+(System.currentTimeMillis()-startTime)+" milliseconds.");
        Statik.writeMatrix(erg, "MCvsSSBeta.txt", '\t');
    }

    public static void checkSSvsMCGamma()
    {
        long startTime = System.currentTimeMillis();
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double[] err = new double[]{1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0,1.0/3.0};
        double[][] erg = new double[51][2];

        for (int a=0; a<51; a++) {
            System.out.println("Starting a = "+a+", so far used "+(System.currentTimeMillis()-startTime)+" milliseconds.");
            erg[a][0] = gammaPower(1.05, 1.05, 0.1, a*0.2/50, facCov, new double[]{0,0,0,0}, err, 0.0, 100, 100000, 1);
            erg[a][1] = sarisSatorraPower(1.05, 1.05, 0.1, a*0.2/50, facCov, new double[]{0,0.0,0,0}, err, 
                                          1.05, 1.05, 0.1, 0, facCov, new double[]{0,0.0,0,0}, err, 10, 100, 1);
        }
        System.out.println("I finished, took me "+(System.currentTimeMillis()-startTime)+" milliseconds.");
        Statik.writeMatrix(erg, "MCvsSSGamma.txt", '\t');
    }

    public static void main(String[] args) 
    {
//        simulationUnequalAttrition();
//        System.exit(0);
        
//        double[] detCoeffs = new double[9];
//        numerDetCoeffsBeta(detCoeffs, 2,3,4,5,new double[]{6,7,8,9,10});
//        double[][] detCoeffs2 = new double[9][3];
//        numerDetCoeffs(detCoeffs2, 3,4,5,new double[]{6,7,8,9,10});
//        for (int i=0; i<9; i++) {
//            double aexp = 1; for (int j=0; j<3; j++) {detCoeffs[i] -= aexp*detCoeffs2[i][j]; aexp *= 2;}
//            System.out.println(detCoeffs[i]);
//        }
//        
//        double[][][] sig = new double[5][5][9];
//        adjointECoeffsBeta(sig, 2,3,4,5,new double[]{6,7,8,9,10});
//        double[][][][] sig2 = new double[5][5][9][3];
//        adjointEBothCoeffs(sig2, 3,4,5,new double[]{6,7,8,9,10});
//        for (int i=0; i<5; i++)
//            for (int j=i; j<5; j++)
//                for (int k=0; k<9; k++)
//                {
//                    double aexp = 1; for (int l=0; l<3; l++) { sig[i][j][k] -= aexp*sig2[i][j][k][l]; aexp *= 2;}
//                    System.out.println((i+1)+", "+(j+1)+", "+k+", "+sig[i][j][k]);
//                }

//        simulationMovingReliability();
//        simulationNoAttrition();
//        simulationAttrition();
        
//            double[] patProb = new double[16];
//            int[][] pat = new int[16][5];
//            generatePatterns(pat, patProb, new double[]{0,0,0,0.8});
//            System.out.println(approximatePowerWithAttrition(1.0, 1.1, 1.0, 0.0, 1.0, new double[]{4,4,4,4,4}, 
//                  1.0, 1.0, 1.0, 0.0, 1.0, 1, 100, 10000, pat, patProb, 0));
//          
//            System.out.println(simulatePowerWithAttrition(1.0, 1.1, 1.0, 0.0, 1.0, new double[][]{{4,4,4,4,0}}, 
//                  1.0, 1.0, 1.0, 0.0, 1.0, 1, new int[]{100}, 1000, 1));
//          System.out.println(simulatePower(1, 1.1, 1, 0, 1, new double[]{4,4,4,4,0.1}, 1, 1.0, 1, 0, 1, 1, 100, 1000, 1));
//        System.out.println(simulatePowerWithAttrition(1.0, 1.05, 1.0, 0.0, 1.0, new double[][]{{4,4,4,4,4},{4,4,4,4,0},{4,4,4,0,4},{4,4,4,0,0}}, 
//                1.0, 1.0, 1.0, 0.0, 1.0, 1, new int[]{70,15,5,10}, 100000, 1));
//        System.out.println(approximatePowerWithAttrition(1.0, 1.05, 1.0, 0.0, 1.0, new double[]{4,4,4,4,4}, 
//                1.0, 1.0, 1.0, 0.0, 1.0, 1, 100, 100000, new int[][]{{1,1,1,1,1},{1,1,1,1,0},{1,1,1,0,1},{1,1,1,0,0}}, new double[]{0.7,0.15,0.05,0.1}, 1));
//        System.out.println(approximatePowerWithAttrition(1.0, 1.05, 1.0, 0.0, 1.0, new double[]{4,4,4,4,4}, 
//                1.0, 1.0, 1.0, 0.0, 1.0, 1, 100, 10000, new int[][]{{1,1,1,1,1}}, new double[]{1}, 1));

        
//        double power = sarisSatorraPower(1.0, 1.0, 0.5, 0.1, id, new double[]{0,0.05,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 
//                                         1.0, 1.0, 0.5, 0.0, id, new double[]{0,0,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 10, 100, 1);
//        System.out.println(power);
//        
//        System.out.println(gammaPowerAttrition(1.0, 1.0, 0.1, 0.1, id, new double[]{0,0,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 
//                0.1, 100, 10000, 1, 0.5, 0.5, ONREAL));
//        System.out.println(sarisSatorraPowerAttrition
//               (1.0, 1.0, 0.1, 0.1, id, new double[]{0,0,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1},
//                1.0, 1.0, 0.1, 0.1, id, new double[]{0,0,0,0}, new double[]{1,1,1,1,1,1,1,1,1,1}, 
//                new int[][]{{0,1},{0,1,2,3},{0,1,2,3,4,5},{0,1,2,3,4,5,6,7},{0,1,2,3,4,5,6,7,8,9}}, 
////                new double[]{0.0,0.0,0.0,0.0,1.0}, 
//                new double[]{0.05,0.0475,0.0451,0.0429,0.8145}, 
//                10, 100, 1));

          // Oertzen, Ghisletta & McArdle Methods
//        checkSSvsMCBeta();                                                                // Figure 1
//        checkSSvsMCGamma();                                                               // Figure 2
//        simulationPlannedAttrition();                                                     // Figure 3 (alt)
        simulationPlannedAttritionMC();                                                     // Figure 3 (neu)
//        gammaMovingAttrition(Integer.parseInt(args[0]),Integer.parseInt(args[1]));        // Figure 4,5,6
//        gammaNullMovingNonrandomAttritionAndBias();                                       // Figure 7, 8
//        gammaMovingError();                                                               // Figure 9
//        gammaMovingErrorDistribution();                                                   // Figure 10 
//        gammaMovingBothErrors();                                                          // Figure 11
//        gammaTableValuesOnAttrition();
                                                                                            // Following Table 2
        /*
        double[][] facCov = new double[][]{{1,0,0,0},{0,0.25,0,0},{0,0,1,0},{0,0,0,0.25}};
        double relX = 0.995, relY = 0.5;
        double[] err = new double[10]; for (int i=0; i<10; i++) err[i] = (1.0/(i%2==0?relX:relY)) - 1.0;
        System.out.println("Power to detect gY = "+
                           sarisSatorraPower(1.05, 1.05, 0.03, 0.03, facCov, new double[]{0,0,0,0}, err, 
                                             1.05, 1.05, 0.03, 0.00, facCov, new double[]{0,0,0,0}, err, 10, 100, 1));
        
        err = new double[10]; for (int i=0; i<10; i++) err[i] = (1.0/(i%2==0?relY:relX)) - 1.0;
        System.out.println("Power to detect gX = "+
                           sarisSatorraPower(1.05, 1.05, 0.03, 0.03, facCov, new double[]{0,0,0,0}, err, 
                                             1.05, 1.05, 0.03, 0.00, facCov, new double[]{0,0,0,0}, err, 10, 100, 1));
        err = new double[10]; for (int i=0; i<10; i++) err[i] = (1.0/(i%2==0?relX:relY)) - 1.0;
        System.out.println("power to detect gY = "+
                           sarisSatorraPower(1.05, 1.05, 0.01, 0.05, facCov, new double[]{0,0,0,0}, err, 
                                             1.05, 1.05, 0.01, 0.00, facCov, new double[]{0,0,0,0}, err, 10, 100, 1));
        err = new double[10]; for (int i=0; i<10; i++) err[i] = (1.0/(i%2==0?relY:relX)) - 1.0;
        System.out.println("power to detect gX= "+
                           sarisSatorraPower(1.05, 1.05, 0.01, 0.05, facCov, new double[]{0,0,0,0}, err, 
                                             1.05, 1.05, 0.01, 0.00, facCov, new double[]{0,0,0,0}, err, 10, 100, 1));
        */
        
        /*       Testing derivatives
        
        double[][][] centMoment = new double[3][10][10]; double[][] sum = new double[3][10];
        computeModelCovAndMeanWRTGamma(centMoment, sum, 0.8, 0.8, 0.2, 0.1, m, new double[]{0,0,0,0}, new double[]{1,2,3,4,5,6,7,8,9,10});
        double[] erg = new double[3], erg2 = new double[3];
        double oldD = 0, oldDD = 0;
        for (int i=0; i<10; i++)
        {
            double gy = 0.3 + eps*i;
            computeModelCovAndMeanWRTGamma(Sigma, mu, 0.8, 0.8, 0.2, gy, m, new double[]{0,0,0,0}, new double[]{1,2,3,4,5,6,7,8,9,10});
            computeMinusTwoLogLikelihoodWRTGamma(erg, Sigma, mu, centMoment[0], sum[0], 1000);
            System.out.println(erg[0]+"\t"+erg[1]+"\t"+(erg[0]-oldD)/eps+"\t"+erg[2]+"\t"+(erg[1]-oldDD)/eps);
            oldD = erg[0]; oldDD = erg[1];
        }
        */
        
//        ssPowers();
//        System.out.println(simulatePower(1.1, 1, 1, 0, 1, new double[]{1,1,1,1,1}, 1, 1.0, 1, 0, 1, 1, 100, 1000, 2));
//        System.out.println(simulatePowerWithAttrition(1.1, 1, 1, 0, 1, new double[][]{{1,1,1,1,1}}, 1, 1, 1, 0, 1, 1, new int[]{100}, 1000, 2));
//        betaMovingCorrelation();
//        gammaMovingRest();
    }
    
}
