/*
 * Created on 09.03.2009
 */
package lgmsimulation;

import silvers.*;

public class DoublePolynomial {

    public final static int NOPARAMETER = -1;
    
    public double[] factor;
    public int[][] parNumber;
    public int[][] exponent;
    
    public DoublePolynomial(double[] factor, int[][] parNumber, int[][] exponent)
    {
        this.factor = new double[factor.length]; for (int i=0; i<factor.length; i++) this.factor[i] = factor[i];
        this.parNumber = new int[parNumber.length][]; 
        for (int i=0; i<parNumber.length; i++) {
            this.parNumber[i] = new int[parNumber[i].length]; for (int j=0; j<parNumber[i].length; j++) this.parNumber[i][j] = parNumber[i][j];
        }
        this.exponent = new int[exponent.length][]; 
        for (int i=0; i<exponent.length; i++) {
            this.exponent[i] = new int[exponent[i].length]; for (int j=0; j<exponent[i].length; j++) this.exponent[i][j] = exponent[i][j];
        }
    }
    
    public DoublePolynomial(double factor, int[] parNumber, int[] exponent)
    {
        this.factor = new double[]{factor};
        this.parNumber = new int[1][parNumber.length]; for (int i=0; i<parNumber.length; i++) this.parNumber[0][i] = parNumber[i];
        this.exponent = new int[1][exponent.length]; for (int i=0; i<exponent.length; i++) this.exponent[0][i] = exponent[i];
    }
    
    public DoublePolynomial(double value)
    {
        this.factor = new double[]{value};
        this.parNumber = new int[1][0];
        this.exponent = new int[1][0];
    }
    
    public DoublePolynomial()
    {
        this.factor = new double[0];
        this.parNumber = new int[0][];
        this.exponent = new int[0][];
    }
    
    public DoublePolynomial(DoublePolynomial in)
    {
        this(in.factor, in.parNumber, in.exponent);
    }
    
    protected void prune()
    {
        for (int i=0; i<factor.length; i++)
        {
            int anzExps = 0;
            for (int j=0; j<exponent[i].length; j++) if (exponent[i][j]==0) anzExps++;
            if (anzExps > 0)
            {
                int[] nexps = new int[exponent[i].length-anzExps];
                int[] npars = new int[exponent[i].length-anzExps];
                int c = 0;
                for (int j=0; j<npars.length; j++) {if (exponent[i][j]==0) c++; npars[j] = parNumber[i][j+c];} 
                for (int j=0; j<nexps.length; j++) {if (exponent[i][j]==0) c++; nexps[j] = exponent[i][j+c];}
                parNumber[i] = npars;
                exponent[i] = nexps;
            }
        }
        int anzMonoms = 0;
        for (int i=0; i<factor.length; i++)
        {
            for (int j=i+1; j<factor.length; j++)
            {
                boolean equal = (exponent[i].length == exponent[j].length);
                int k=0;
                while ((equal) && (k<exponent[i].length)) {equal = (parNumber[i][k]==parNumber[j][k]) && (exponent[i][k] == exponent[j][k]); k++;}
                if (equal) {factor[i] += factor[j]; factor[j] = 0;}
            }
            if (factor[i] == 0.0) anzMonoms++;
        }
        
        if (anzMonoms > 0)
        {
            int[][] nexps = new int[factor.length-anzMonoms][];
            int[][] npars = new int[factor.length-anzMonoms][];
            double[] nfacs = new double[factor.length-anzMonoms];
            int c = 0;
            for (int i=0; i<nfacs.length; i++)
            {
                if (factor[i]==0.0) c++;
                nexps[i] = new int[exponent[i+c].length]; for (int j=0; j<nexps[i].length; j++) nexps[i][j] = exponent[i+c][j];
                npars[i] = new int[parNumber[i+c].length]; for (int j=0; j<npars[i].length; j++) npars[i][j] = parNumber[i+c][j];
                nfacs[i] = factor[i+c]; 
            }
            exponent = nexps; parNumber = npars; factor = nfacs;
        }
    }
    
    public void substituteValue(int par, double value)
    {
        for (int i=0; i<factor.length; i++)
        {
            int pruneout = -1;
            for (int j=0; j<parNumber[i].length; j++)
                if (parNumber[i][j] == par) {factor[i] *= Math.pow(value,exponent[i][j]); pruneout = j;}
            if (pruneout != -1) {
                int[] nexponent = new int[exponent[i].length-1]; 
                for (int j=0; j<nexponent.length; j++) nexponent[j] = (j<pruneout?exponent[i][j]:exponent[i][j+1]); 
                int[] nparNumber = new int[parNumber[i].length-1];
                for (int j=0; j<nparNumber.length; j++) nparNumber[j] = (j<pruneout?parNumber[i][j]:parNumber[i][j+1]);
                parNumber[i] = nparNumber; exponent[i] = nexponent;
            }
        }
        prune();
    }
    
    /**
     * Evaluates the polnyomial at values as given. If parameter Numbers are given, the polynomial is first differentiated wrt these numbers.
     * @param values
     * @return
     */
    public double evaluate(double[] values) {return evaluate(values, NOPARAMETER, NOPARAMETER);}
    public double evaluate(double[] values, int first) {return evaluate(values, first, NOPARAMETER);}
    public double evaluate(double[] values, int first, int second)
    {
        double erg = 0;
        for (int i=0; i<factor.length; i++)
        {
            double zwerg = factor[i]; int exp1st = 0; int exp2nd = 0;
            for (int j=0; j<parNumber[i].length; j++)
            {
                if (parNumber[i][j] == first) exp1st = exponent[i][j]; else
                if (parNumber[i][j] == second) exp2nd = exponent[i][j]; else 
                    zwerg *= Math.pow(values[parNumber[i][j]],exponent[i][j]);
            }
            if (first==second) {
                if (exp1st >= 2) zwerg *= exp1st * (exp1st-1) * Math.pow(values[first],exp1st-2);
                else if (first != NOPARAMETER) zwerg = 0;
            } else {
                if (exp1st >= 1) zwerg *= exp1st * Math.pow(values[first],exp1st-1); else if (first != NOPARAMETER) zwerg = 0;
                if (exp2nd >= 1) zwerg *= exp2nd * Math.pow(values[second],exp2nd-1); else if (second != NOPARAMETER) zwerg = 0;
            }
            erg += zwerg;
        }
        return erg;
    }
    
    /**
     * Substitutes value for the indicated parameter and afterwards reduces all higher parameter numbers by one.
     * 
     * @param par
     * @param value
     */
    public void fixParameter(int par, double value)
    {
        substituteValue(par, value);
        for (int i=0; i<factor.length; i++)
            for (int j=0; j<parNumber[i].length; j++)
                if (parNumber[i][j] > par) parNumber[i][j]--;
    }
    
    public String toString()
    {
        String erg = "";
        for (int i=0; i<factor.length; i++)
        {
            erg += Statik.doubleNStellen(factor[i], 2);
            if (exponent[i].length>0) erg+="*";
            for (int j=0; j<exponent[i].length; j++)
            {
                erg += "["+parNumber[i][j]+"]";
                if (exponent[i][j]>1) erg += "^"+exponent[i][j];
            }
            if (i<factor.length-1) erg += " + ";
        }
        return erg;
    }
    
    public int maxParNumber()
    {
        int erg = NOPARAMETER;
        for (int i=0; i<factor.length; i++)
            for (int j=0; j<parNumber[i].length; j++) erg = Math.max(parNumber[i][j], erg);
        return erg;
    }
    
    public static void main(String[] args)
    {
        DoublePolynomial p = new DoublePolynomial(new double[]{1.0,2.0,1.0}, new int[][]{{},{0},{0}}, new int[][]{{},{1},{2}});
        double v = p.evaluate(new double[]{2}), vd = p.evaluate(new double[]{2},0), vdd = p.evaluate(new double[]{2},0,0);
        System.out.println(p);
        System.out.println("Value = "+v+", vd = "+vd+", vdd = "+vdd);
        p = new DoublePolynomial(new double[]{1.0,2.0,1.0}, new int[][]{{0},{1},{0,1}}, new int[][]{{3},{3},{3,2}});
        v = p.evaluate(new double[]{1,2}); vd = p.evaluate(new double[]{1,2},0); 
        vdd = p.evaluate(new double[]{1,2},0,1); double vdd2 = p.evaluate(new double[]{2,1},0,1);
        System.out.println(p);
        System.out.println("Value = "+v+", vd = "+vd+", vdd = "+vdd+", vdd2 = "+vdd2);
        p.substituteValue(1, 2.0);
        System.out.println(p);
    }
}
