/*
 * Created on 11.09.2009
 */
package lgmsimulation;

import java.io.PrintStream;
import java.util.Arrays;
import java.util.Random;

import silvers.Statik;

public abstract class Model 
{
      /** Indicates whether computeLogLikelihoodDerivatives and computeLeastSquaresDerivatives compute the first
       *  and second derivatives wrt to lambda, where lambda 
       */
      static public final boolean supportsPathDerivatives = false; 
    
      static public final int NOPARAMETER = -1;
      static public final int MISSING = -999;
      static public final double LNTWOPI = Math.log(2*Math.PI);
//      static protected Random rand = new Random(20398746);
      static protected final double suggestedEPS = 0.0001;
    static Random rand = new Random();
    
      public static enum warningFlagTypes {OK, SUSPICIOUS, FAILED};
      public warningFlagTypes warningFlag;
      
      /** if null, no messages will be issued **/
      public PrintStream logStream = null;
    
      public int anzPar;
      public int anzVar;
      
      public int anzPer;
      public double[][] data;
      public double[] dataMean;
      public double[] xsum;
      public double[][] dataCov;
      public double[][] xBiSum;
    
      public double[][] sigma;
      public double[] mu;
      public double sigmaDet;
      
      public double ll;
      public double[] llD;
      public double[][] llDD;
      public double ls;
      public double[] lsD;
      public double[][] lsDD;
      public double pathLambda;
      public double llPathD;
      public double[] llPathDD;
      public double[] pathDirection;
      
      // Some working variables
      private double[][] modelSigmaWork, modelSigmaWorkPD, modelSigmaWork2PD, modelWork1, modelWork2;
      private double[] p1w, p2w, cew, a1w, a2w, modelPos, modelMove, modelMuWork;
      
      double[][] sigInv;
      double[] lastEstimate;
      double convergenceIndex; // <=1 no convergence, ca. above 1.3 good convergence
      double lastGain, lastSteplength;
      int steps;
      
      public double[][] fisherInformationMatrix;

      public int getAnzPar() {return anzPar;}
      public int getAnzVar() {return anzVar;}

      /** copies the model at hand */
      public abstract Model copy(); 
      /** Computes matrix multiplied by the derivative of sigma wrt. par, matrix = null computes only the derivative. */
      protected abstract void computeMatrixTimesSigmaDev(int par, double[][] matrix, double[][] erg);
      /** Computes matrix multiplied by the 2nd derivative of sigma wrt. par1, par2, matrix = null computes only the derivative. */
      protected abstract void computeMatrixTimesSigmaDevDev(int par1, int par2, double[][] matrix, double[][] erg);
      /** Computes matrix multiplied by the derivative of mu wrt. par, matrix = null computes only the derivative. */
      protected abstract void computeMatrixTimesMuDev(int par, double[][] matrix, double[] erg);

      /** Computes the inverse of sigma multiplied by the derivative of sigma wrt. par. */
      protected void computeSigmaInvDev(int par, double[][] erg) {computeMatrixTimesSigmaDev(par, sigInv, erg);}
      /** Computes the inverse of sigma multiplied by the 2nd derivative of sigma wrt. par1, par2. */
      protected void computeSigmaInvDevDev(int par1, int par2, double[][] erg) {computeMatrixTimesSigmaDevDev(par1, par2, sigInv, erg);}
      /** Computes the inverse of sigma multiplied by the derivative of mu wrt. par */
      protected void computeSigmaInvMuDev(int par, double[][] matrix, double[] erg) {computeMatrixTimesMuDev(par, sigInv, erg);}
      
      public void setData(double[][] data)
      {
          this.data = data;
          this.anzPer = data.length;
          computeMoments();
      }
    
      public void computeMoments()
      {
          if (data==null) return;
          xsum = new double[anzVar];
          dataMean = new double[anzVar];
          for (int i=0; i<anzVar; i++) xsum[i] = 0.0;
          for (int i=0; i<anzPer; i++)
              for (int j=0; j<anzVar; j++) xsum[j] += data[i][j];
          for (int i=0; i<anzVar; i++) dataMean[i] = xsum[i] / (double)anzPer;
            
          xBiSum = new double[anzVar][anzVar];
          dataCov = new double[anzVar][anzVar];
          for (int i=0; i<anzVar; i++)
              for (int j=0; j<=i; j++)
              {
                  xBiSum[i][j] = 0.0;
                  for (int k=0; k<anzPer; k++) xBiSum[i][j] += data[k][i] * data[k][j];
                  for (int k=0; k<anzPer; k++) dataCov[i][j] += (data[k][i]-dataMean[i]) * (data[k][j]-dataMean[j]);
                  xBiSum[j][i] = xBiSum[i][j];
                  dataCov[i][j] /= ((double)anzPer-1.0);
                  dataCov[j][i] = dataCov[i][j];
              }
      }
      
      public void computeMomentsFromDataCovarianceAndMean()
      {
          for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) xBiSum[i][j] = (anzPer-1)*dataCov[i][j] + anzPer*dataMean[i]*dataMean[j];
          for (int i=0; i<anzVar; i++) xsum[i] = anzPer*dataMean[i];
      }

      /** Sets all parameter to the given values. */
      public void setParameter(double[] values) {for (int i=0; i<values.length; i++) setParameter(i,values[i]);}
      /** Sets parameter to the given value. */
      public abstract void setParameter(int nr, double value);
    
      public double[] getParameter() {double[] erg = new double[getAnzPar()]; for (int i=0; i<erg.length; i++) erg[i] = getParameter(i); return erg;}
      public abstract double getParameter(int nr);
      /** Returns true if the parameter is only involved in the diagonal of sigma. */
      public abstract boolean isErrorParameter(int nr);
    
      /** Fixes the parameters to their values, reduces higher parameter numbers accordingly. */
      public void fixParameter(int[] nrs) {Arrays.sort(nrs); for (int i=nrs.length-1; i>=0; i--) fixParameter(nrs[i]);}
      /** Fixes the paramter to its value, reduces higher parameter numbers accordingly. */
      public void fixParameter(int nr) {fixParameter(nr, getParameter(nr));}
      /** Fixes the parameter to the given values, reduces higher parameter numbers accordingly. */
      public abstract void fixParameter(int nr, double value);
    
      /** Creates a data set according to model sigma. */
      public double[][] createData (int anzPersons, double[] values) {setParameter(values); return createData(anzPersons);}
      public double[][] createData (int anzPersons)
      {
          this.anzPer = anzPersons;
          data = new double[anzPersons][anzVar];
          
          evaluateMuAndSigma(null);
          double[][] cholesky = Statik.choleskyDecompose(sigma);

          double[] m = new double[anzVar], m2 = new double[anzVar];
          for (int i=0; i<anzPersons; i++)
          {
              for (int j=0; j<anzVar; j++) m[j] = rand.nextGaussian();
              Statik.multiply(cholesky,m,m2);
              for (int j=0; j<anzVar; j++) data[i][j] = m2[j]+mu[j];
          }
          computeMoments();
          return data;
      }
          
      /** Recomputes mu and Sigma of the model. */
      public void evaluateMuAndSigma() {evaluateMuAndSigma(null);}
      /** Recomputes mu and Sigma, applying the given values for the parameter. valus = null means using the actual parameter values. */
      public abstract void evaluateMuAndSigma(double[] values);
    
      /** Computes the least Squares index. */
      public double getLeastSquares() {return getLeastSquares(null);}
      /** Applies the parameter values (null => actual values) and computes the least Squares index. */
      public double getLeastSquares(double[] values)
      {
          evaluateMuAndSigma(values);
          ls = 0;
          for (int i=0; i<anzVar; i++)
              for (int j=0; j<anzVar; j++) ls += 0.5*(sigma[i][j] - dataCov[i][j])*(sigma[i][j] - dataCov[i][j]);
          for (int i=0; i<anzVar; i++) ls += 0.5*(mu[i] - dataMean[i])*(mu[i] - dataMean[i]);
          return ls;
      }
        
      /** Computes the -2 log likelihood index. */
      public double getMinusTwoLogLikelihood() {return getMinusTwoLogLikelihood(null, true);}
      /** Applies the parameter values (null => actual values) and computes the -2 log likelihood index. */
      public double getMinusTwoLogLikelihood(double[] value) {return getMinusTwoLogLikelihood(value, true);}
      /** Applies the parameter values (null => actual values) and computes the -2 log likelihood index. 
       * If recomputeMuAndSigma is false, these values will not be recomputed. */
      public double getMinusTwoLogLikelihood(double[] value, boolean recomputeMuAndSigma) 
      {
          if ((modelSigmaWork==null) || (modelSigmaWork.length!=anzVar) || (modelSigmaWork[0].length!=anzVar)) modelSigmaWork = new double[anzVar][anzVar];
          if ((sigInv==null) || (sigInv.length!=anzVar) || (sigInv[0].length!=anzVar)) sigInv = new double[anzVar][anzVar];

          if (value != null) setParameter(value);
          if (recomputeMuAndSigma) evaluateMuAndSigma(value);
        
        try {sigmaDet = Statik.invert(sigma,sigInv, modelSigmaWork);} catch(RuntimeException e) {sigmaDet = -1;}
        if ((Statik.determinantOfPositiveDefiniteMatrix(sigma)==-1) && (sigmaDet > 0))
        {
//            if (logStream != null) logStream.println("Warning: Not positive defnite Sigma, I set ll to NaN.");
            sigmaDet = Double.NaN;
        }
        
        ll = anzPer*anzVar*LNTWOPI + anzPer*Math.log(sigmaDet);
        for (int i=0; i<anzVar; i++) 
            for (int j=0; j<anzVar; j++)
                ll += sigInv[i][j]*(xBiSum[i][j] - anzPer*mu[i]*dataMean[j] - anzPer*mu[j]*dataMean[i] + anzPer*mu[i]*mu[j]);
        
        if ((logStream!=null) && (ll < 0)) logStream.println("Warning: encountered negative Log Likelihood (possible, but very rare). I accept it as such.");
        
        return ll;
    }

      /**
       * Computes the least square index and for each parameter its derivative and second derivative. Parameters are expected to be numbered
       * from 0 to anzPar-1. If values is not null, these parameter values are applied; otherwise, the method assumes all values are correct.
       */
      public void computeLeastSquaresDerivatives(double[] value) {computeLeastSquaresDerivatives(value,true);}
      /**
       * Computes the least square index and for each parameter its derivative and second derivative. Parameters are expected to be numbered
       * from 0 to anzPar-1. If values is not null, these parameter values are applied; otherwise, the method assumes all values are correct.
       * If recomputeMuAndSigma is false, these values will not be recomputed.
       */
      public abstract void computeLeastSquaresDerivatives(double[] value, boolean recomputeMuAndSigma);

      /** Finds arbitrary starting values */
    public double[] getArbitraryStartingValues()
    {
        double[] erg = new double[anzPar];
        
        for (int i=0; i<anzPar; i++)
        {
            erg[i] = 1.0;
            if (isErrorParameter(i)) erg[i] = 3.0;
        }
        
        return erg;
    }

    /** Computes the minus two log likelihood index and its first and second derivatives wrt to all parameters. 
     * If value is null, actual parameter values are assumed */
    public void computeLogLikelihoodDerivatives(double[] value) {computeLogLikelihoodDerivatives(value,true);}
    /** Computes the minus two log likelihood index and its first and second derivatives wrt to all parameters. 
     * If value is null, actual parameter values are assumed.
     * If recomputeMuAndSigma is false, these values will not be recomputed. */
    public abstract void computeLogLikelihoodDerivatives(double[] value, boolean recomputeMuAndSigma);

    /**
     * Computes a single step in the newton method, using optimal damping in the direction suggested. Returns false if any problems occured.
     * If true is returned, convergence measures are computed.
     * 
     * @param useMaxLikelihood      If true, the maximum likelihood is optimized, otherwise the least squares.
     * @return                      true if successful, false if stuck or negative result
     */
    public boolean moveWithOptimalDamping(double EPS, boolean useMaximumLikelihood) {return moveWithOptimalDampingPregivenStep(EPS, useMaximumLikelihood, true);}

    /**
     * Computes a single step in the pregiven direction if computeNewtonStep is false and in the newton method if true.
     * 
     * @param EPS
     * @param useMaximumLikelihood
     * @param computeNewtonStep
     * @return
     */
    public boolean moveWithOptimalDampingPregivenStep(double EPS, boolean useMaximumLikelihood, boolean computeNewtonStep)
    {
        final int MAXINNERRUNS = 100;
        double[] pos = modelPos, move = modelMove;      // just simpler naming
        
        steps++;
        
        if ((p1w==null) || (p1w.length!=pos.length)) p1w = new double[pos.length];
        if ((p2w==null) || (p2w.length!=pos.length)) p2w = new double[pos.length];
        if ((cew==null) || (cew.length!=pos.length)) cew = new double[pos.length];
        if ((a1w==null) || (a1w.length!=pos.length)) a1w = new double[pos.length];
        if ((a2w==null) || (a2w.length!=pos.length)) a2w = new double[pos.length];

        if (computeNewtonStep)
        {
            try {Statik.invert((useMaximumLikelihood?llDD:lsDD), modelWork1, modelWork2);} catch (Exception e) {return false;}
            Statik.multiply(modelWork1, (useMaximumLikelihood?llD:lsD), modelMove);
        }
        
        double[] t; double td;
        double startValue = (useMaximumLikelihood?getMinusTwoLogLikelihood(pos):getLeastSquares(pos));
        double p1v = startValue;

        int n = pos.length;
        for (int i=0; i<n; i++) p1w[i] = pos[i];
        
        double derivative = Statik.multiply((useMaximumLikelihood?llD:lsD), move);
        Statik.multiply((int)Math.signum(-derivative), move, p2w);
        Statik.add(pos,p2w,p2w);
        double p2v = (useMaximumLikelihood?getMinusTwoLogLikelihood(p2w):getLeastSquares(p2w));
        
        int inRuns = 0;
        while (((Double.isNaN(p2v)) || (Double.isInfinite(p2v))) && (inRuns<MAXINNERRUNS)) {
            for (int i=0; i<n; i++) p2w[i] = (p1w[i]+p2w[i])/2.0; 
            p2v = (useMaximumLikelihood?getMinusTwoLogLikelihood(p2w):getLeastSquares(p2w));
            inRuns++;
        }
        
        double cev = p2v+EPS*EPS, a1v, a2v, dist;
        
        // p1 will not be manipulated in this loop.
        while ((p2v < p1v) && (p2v < cev))
        {
            for (int i=0; i<n; i++) cew[i] = p2w[i]; cev = p2v;
            for (int i=0; i<n; i++) p2w[i] = 2 * cew[i] - p1w[i];
            p2v = (useMaximumLikelihood?getMinusTwoLogLikelihood(p2w):getLeastSquares(p2w));
            while (((Double.isNaN(p2v)) || (Double.isInfinite(p2v))) && (inRuns<MAXINNERRUNS)) {
                for (int i=0; i<n; i++) p2w[i] = (cew[i]+p2w[i])/2.0; 
                p2v = (useMaximumLikelihood?getMinusTwoLogLikelihood(p2w):getLeastSquares(p2w));
                inRuns++;
            }

            if (inRuns == 30 && logStream!=null) logStream.println("I seem to approach an infinite minimum, best value = "+
                    p2v+", previous was = "+cev+", start = "+startValue+", I'll try some more moves.");
            if (inRuns > 100) {
                if (logStream != null) logStream.println("I hit a boundary or infinite minimum, I give up the inner iteration."); 
                return false;
            }
            inRuns++;
        }

        for (int i=0; i<n; i++) cew[i] = (p1w[i]+p2w[i])/2.0; cev = (useMaximumLikelihood?getMinusTwoLogLikelihood(cew):getLeastSquares(cew));
        // at this point, it should be guaranteed that a minimum lies between p1w and p2w. 
        if (p2v < p1v) {t = p1w; p1w = p2w; p2w = t; td = p1v; p1v = p2v; p2v = td;}
        boolean goon = true; inRuns = 0;
        while ((goon) && (inRuns < MAXINNERRUNS))
        {
            while ((((cev>p1v) && (cev>p2v)) || (Double.isInfinite(cev)) || (Double.isNaN(cev))) && (inRuns < MAXINNERRUNS))
            {
                // This case means a maximum between p1v and p2v, or a illegal area, which is bad; 
                // taking the lower point is the best guess we can make. We will converge out of the prohibited area, but possibly
                // to a stretch without a minimum.
                t = p2w; p2w = cew; cew = t; td = p2v; p2v = cev; cev = td;
                for (int i=0; i<n; i++) cew[i] = (p1w[i]+p2w[i])/2.0; cev = (useMaximumLikelihood?getMinusTwoLogLikelihood(cew):getLeastSquares(cew));
                inRuns++;
            }
            if ((cev>p1v) || (cev>p2v)) {
                if (cev>p1v) {t = p2w; p2w = cew; cew = t; p2v = cev;} else {t = p1w; p1w = cew; cew = t; p1v = cev;}  
                for (int i=0; i<n; i++) cew[i] = (p1w[i]+p2w[i])/2.0; cev = (useMaximumLikelihood?getMinusTwoLogLikelihood(cew):getLeastSquares(cew));
            } else {
                // This is the case that center is the smaller than both p1 and p2. Quarters need to be computed to make sure
                // that a minimum lies in the part.
                for (int i=0; i<n; i++) a1w[i] = (p1w[i]+cew[i])/2.0; a1v = (useMaximumLikelihood?getMinusTwoLogLikelihood(a1w):getLeastSquares(a1w)); 
                for (int i=0; i<n; i++) a2w[i] = (p2w[i]+cew[i])/2.0; a2v = (useMaximumLikelihood?getMinusTwoLogLikelihood(a2w):getLeastSquares(a2w)); 
                if (a1v < cev) {t = p2w; p2w = cew; cew = t; p2v = cev; t = cew; cew = a1w; a1w = t; cev = a1v;}
                else if (a2v < cev) {t = p1w; p1w = cew; cew = t; p1v = cev; t = cew; cew = a2w; a2w = t; cev = a2v;}
                else {
                    // This is the case that both quarters are higher than the center.
                    t = p2w; p2w = a2w; a2w = t; p2v = a2v;
                    t = p1w; p1w = a1w; a1w = t; p1v = a1v;
                }
            }
            if (p2v < p1v) {t = p1w; p1w = p2w; p2w = t; td = p1v; p1v = p2v; p2v = td;}
            dist = 0; for (int i=0; i<n; i++) dist += (p1w[i]-p2w[i])*(p1w[i]-p2w[i]);
            if (dist < EPS*EPS) goon = false;
            
            if ((inRuns == 30) && (logStream!=null)) 
                logStream.println("It takes me some time to converge, best value so far = "+p1v+", alternative = "+p2v+", start = "+startValue);
            if (inRuns > MAXINNERRUNS) {
                if (logStream!=null) logStream.println("I give up at inner iteration now."); 
                return false;
            }
            inRuns++;
        }
        for (int i=0; i<n; i++) move[i] = p1w[i] - pos[i];
        for (int i=0; i<n; i++) pos[i] = p1w[i];

        if (p1v < 0) {
            if (logStream!=null) System.out.println("It seems I'm heading towards an infinite minimum: "+p1v); 
            return false;
        }

        // the return true block, computes some statistics about the convergence process and updates the second derivatives.
        if (useMaximumLikelihood) computeLogLikelihoodDerivatives(modelPos); else computeLeastSquaresDerivatives(modelPos);
        double steplength = Statik.abs(modelMove); 
        convergenceIndex = (steplength < 1.0?(steplength > EPS?Math.log(steplength)/Math.log(lastSteplength):2):1);
        lastSteplength = steplength;
        lastGain = startValue - (useMaximumLikelihood?ll:ls);
        if ((Double.isInfinite(lastGain)) || (Double.isNaN(lastGain))) return false;
        return true;
    }    

    /**
     * Computes a single step in the newton method, using optimal damping in the direction suggested. Returns false if any problems occured.
     * If true is returned, convergence measures are computed.
     * 
     * @param useMaxLikelihood      If true, the maximum likelihood is optimized, otherwise the least squares.
     * @return                      true if successful, false if stuck or negative result
     */
    public boolean moveNewtonStep(double EPS, boolean useMaximumLikelihood)
    {
        double[] pos = modelPos, move = modelMove;      // just simpler naming
        
        steps++;
        
        double lastValue = (useMaximumLikelihood?ll:ls);
        Statik.invert((useMaximumLikelihood?llDD:lsDD), modelWork1, modelWork2);
        Statik.multiply(modelWork1, (useMaximumLikelihood?llD:lsD), move);
        
        double derivative = Statik.multiply((useMaximumLikelihood?llD:lsD), move);
        Statik.multiply((int)Math.signum(-derivative), move, move);
        Statik.add(pos,move,pos);
        
        // the return true block, computes some statistics about the convergence process and updates the second derivatives.
        if (useMaximumLikelihood) computeLogLikelihoodDerivatives(pos); else computeLeastSquaresDerivatives(pos);
        double steplength = Statik.abs(move); 
        convergenceIndex = (lastSteplength < 1.0?(steplength > EPS?Math.log(steplength)/Math.log(lastSteplength):2):1);
        lastSteplength = steplength;
        lastGain = lastValue - (useMaximumLikelihood?ll:ls);
        return true;
    }    

    private void initEstimation(double[] starting, boolean useMaximumLikelihood)
    {
        warningFlag = warningFlagTypes.OK;
        if ((modelMove==null) || (modelMove.length!=anzPar)) modelMove = new double[anzPar];
        if ((modelPos==null) || (modelPos.length!=anzPar)) modelPos = new double[anzPar];
        if ((modelWork1==null) || (modelWork1.length!=anzPar) || (modelWork1[0].length!=anzPar)) modelWork1 = new double[anzPar][anzPar];
        if ((modelWork2==null) || (modelWork2.length!=anzPar) || (modelWork2[0].length!=anzPar)) modelWork2 = new double[anzPar][anzPar];
        
        for (int i=0; i<anzPar; i++) modelPos[i] = (starting==null?getParameter(i):starting[i]);
        
        steps = 0; lastSteplength = Double.POSITIVE_INFINITY; lastGain = Double.POSITIVE_INFINITY;
        if (useMaximumLikelihood) computeLogLikelihoodDerivatives(modelPos); else computeLeastSquaresDerivatives(modelPos);
    }
    
    /** Estimates Parameters by minimizing the least squares distance. */
    public double[] estimateLS() {return estimateLS(getArbitraryStartingValues(),suggestedEPS);}
    /** Estimates Parameters by minimizing the least squares distance, EPS gives the precision */
    public double[] estimateLS(double EPS) {return estimateLS(getArbitraryStartingValues(),EPS);}
    /** Estimates Parameters by minimizing the least squares distance, EPS gives the precision, with the pregiven starting values. */
    public double[] estimateLS(double[] starting, double EPS) {return estimateMLOrLS(starting, EPS, false);}
    /** Estimates Parameters by minimizing the -2 log likelihood index with the least squares estimate as starting values */
    public double[] estimateML() {return estimateML(estimateLS(getArbitraryStartingValues(),suggestedEPS), suggestedEPS);}
    /** Estimates Parameters by minimizing the -2 log likelihood index, EPS gives the precision, with the least squares estimate as starting values */
    public double[] estimateML(double EPS) {return estimateML(estimateLS(getArbitraryStartingValues(),EPS), EPS);}
    /** Estimates Parameters by minimizing the -2 log likelihood index, with the pregiven starting values. */
    public double[] estimateML(double[] starting) {return estimateML(starting, suggestedEPS);}
    /** Estimates Parameters by minimizing the -2 log likelihood index, EPS gives the precision, with the pregiven starting values. */
    public double[] estimateML(double[] starting, double EPS) {return estimateMLOrLS(starting, EPS, true);}
    /** Estimates Parameters by minimizing either the -2 log likelihood or the least squares index */
    private double[] estimateMLOrLS(double[] starting, double EPS, boolean useMaximumLikelihood)
    {
        final int MAXRUNS = 50; 

        initEstimation(starting, useMaximumLikelihood);
        
        double startValue = (useMaximumLikelihood?ll:ls);
        if (Double.isNaN(startValue)) {
            if (logStream!=null) logStream.println("The index of the starting values is not defined, I can't optimize!"); 
            warningFlag = warningFlagTypes.FAILED; return starting;
        }
        
//        while ((runs<=MAXRUNS) && (steplength > EPS) && (lastGain > EPS))
        while ((steps<=MAXRUNS) && (lastSteplength > EPS))
        {
            boolean success = moveWithOptimalDamping(EPS, useMaximumLikelihood);

            if (! success) 
            {
                for (int i=0; i<anzPar; i++) modelPos[i] = starting[i];
                if (useMaximumLikelihood) computeLogLikelihoodDerivatives(modelPos); else computeLeastSquaresDerivatives(modelPos);
                try {
                    Statik.invert((useMaximumLikelihood?llDD:lsDD), modelWork1, modelWork2);
                } catch (RuntimeException e) {System.out.println("Matrix in Optimization became singular."); return null;}
                Statik.multiply(modelWork1, (useMaximumLikelihood?llD:lsD), modelMove);
                double v;
                do {
                    for (int i=0; i<anzPar; i++) {
                        modelPos[i] = starting[i] + rand.nextGaussian() * Math.abs(modelMove[i]) / 3.0;
                        if (isErrorParameter(i)) modelPos[i] += Math.abs(modelPos[i]+2*starting[i]);
                    }
                    v = (useMaximumLikelihood?getMinusTwoLogLikelihood(modelPos):getLeastSquares(modelPos));
                } while ((Double.isNaN(v)) || (Double.isInfinite(v)));
                if (useMaximumLikelihood) computeLogLikelihoodDerivatives(modelPos); else computeLeastSquaresDerivatives(modelPos);
                if (logStream!=null) logStream.println("Headed towards boundary minimum, I restarted with perturbed values and higher error.");
                warningFlag = warningFlagTypes.SUSPICIOUS;
            }
        }

        if (steps > MAXRUNS) 
        {
            if (logStream!=null) logStream.println("Warning: More than 50 runs needed, final steplength = "+lastSteplength); 
            warningFlag = warningFlagTypes.FAILED;
        } else {
            double value = (useMaximumLikelihood?ll:ls); 
            if ((value > startValue) || (Double.isNaN(value)) || (Double.isInfinite(value))) {
                if (logStream!=null) logStream.println("I was unable to improve the starting -2ll ("+startValue+"), converged on "+value+"."); 
                warningFlag = warningFlagTypes.FAILED;
            }  
            else if (warningFlag == warningFlagTypes.SUSPICIOUS) 
            {
                if (logStream!=null) logStream.println("It seemed I was able to solve out the problem, convergences reached on -2ll = "+
                        value+", better than starting ("+startValue+".)");
            }
        }
        
        if ((lastEstimate == null) || (lastEstimate.length != anzPar)) lastEstimate = new double[anzPar];
        double[] erg = new double[anzPar];
        for (int i=0; i<anzPar; i++) lastEstimate[i] = erg[i] = modelPos[i];
        return erg;
    }
    
    /** Sets sigma to a diagonal matrix with the data variances on the diagonal. */
    public void independentSigma() 
    {
        if ((sigma==null) || (sigma.length!=anzVar) || (sigma[0].length!=anzVar)) sigma = new double[anzVar][anzVar];
        for (int i=0; i<sigma.length; i++)
            for (int j=0;j<sigma.length ;j++)
                sigma[i][j] = (i==j?this.dataCov[i][j]:0);
    }
    
    /** Computes the degrees of freedom between the actual and a saturated model. */
    public int restrictedDF()
    {
        return anzVar*(anzVar+1)/2 - anzPar;
    }
    
    /** Computes the -2 log likelihood for the data set assuming the data covariance matrix and data mean as mu and sigma. */
    public double getSaturatedLL() 
    {
        if ((sigma==null) || (sigma.length!=anzVar) || (sigma[0].length!=anzVar)) sigma = new double[anzVar][anzVar];
        if ((mu==null) || (mu.length!=anzVar)) mu = new double[anzVar];
        double myll = ll;
        for (int i=0; i<anzVar; i++)
            for (int j=0; j<anzVar; j++)
                sigma[i][j] = dataCov[i][j];
        for (int i=0; i<anzVar; i++)
            mu[i] = dataMean[i];
        
        double sat = getMinusTwoLogLikelihood(null, false);
        ll = myll;
        return sat;
    }

    /** Computes the -2 log likelihood for the data set assuming the independent sigma as sigma. */
    public double getIndependentLL() 
    {
        if ((sigma==null) || (sigma.length!=anzVar) || (sigma[0].length!=anzVar)) sigma = new double[anzVar][anzVar];
        if ((mu==null) || (mu.length!=anzVar)) mu = new double[anzVar];

        double myll = ll;
        for (int i=0; i<anzVar; i++)
            for (int j=0; j<anzVar; j++)
                sigma[i][j] = (i==j?this.dataCov[i][j]:0);
        for (int i=0; i<anzVar; i++)
            mu[i] = dataMean[i];
        
        double sat = getMinusTwoLogLikelihood(null, false);
        ll = myll;
        return sat;
    }

    /** Computes the RMSEA fit index for this model. */
    public double getRMSEA()
    {
        return Math.sqrt(Math.max((ll - restrictedDF()),0) / (restrictedDF()*(anzPer-1)));
    }
    
    /** Computes the comparative RMSEA fit index for this model compared to an restricted model. Both this model and the restricted
     * mode must have the correct -2 log likelihood value in the ll field.  */
    public double getDeltaRMSEA(Model restrictedModel)
    {
        return Math.sqrt(((restrictedModel.ll - ll) - (restrictedModel.anzPar - anzPar))/((restrictedModel.anzPar - anzPar)*(anzPer-1)));
    }
    
    /** Computes the CFI (comparative fit index) for this model compared to the independent Model. */
    public double getCFI()
    {
        double myll = ll;
        independentSigma(); 
        this.computeLogLikelihoodDerivatives(null,false);
        double indepLL = ll;
        ll = myll;
        int restrictedDF = restrictedDF();
        int indepRestrictedDF = anzVar*(anzVar-1)/2;
        if (myll <= restrictedDF) return 1.0;
        return 1.0-((myll-restrictedDF) / Math.max(myll-restrictedDF,indepLL - indepRestrictedDF));
    }

    /** Computes the CFI (comparative fit index) for this model compared to a restricted model. Both this model and the restricted
     * model must have the correct -2 log likelihood value in the ll field. */
    public double getCFI(Model restrictedModel)
    {
        if (ll <= restrictedDF()) return 1.0;
        return 1.0-((ll-restrictedDF()) / Math.max(ll-restrictedDF(),restrictedModel.ll - restrictedModel.restrictedDF()));
    }
    
    /** Computes the delta CFI (difference of comparative fit indices) for this model compared to a restricted model. Both this model and the restricted
     * model must have the correct -2 log likelihood value in the ll field. */
    public double getDeltaCFI(Model restrictedModel)
    {
        return this.getCFI() - restrictedModel.getCFI();
    }
    
    /** Computes the McDonald Centrality Index. */
    public double getMcDonaldsCentralityIndex()
    {
        return Math.exp(-0.5*(ll-restrictedDF())/(anzPer-1));
    }
    
    /**
     * Computes the derivative of ML with respect to the path coefficient lambda. The actual position is in lambda.
     *   
     * @param baseSigma
     * @param diffSigma
     * @param baseMu
     * @param diffMu
     */
    protected void computePathDerivatives(double lambda, double[][] baseSigma, double[][] diffSigma, 
                                          double[] baseMu, double[] diffMu, boolean computePathDirection)
    {
        if ((modelSigmaWorkPD==null) || (modelSigmaWorkPD.length!=anzVar) || (modelSigmaWorkPD[0].length!=anzVar)) modelSigmaWorkPD = new double[anzVar][anzVar];
        if ((modelMuWork==null) || (modelMuWork.length!=anzVar)) modelMuWork = new double[anzVar];
        if ((modelSigmaWork2PD==null) || (modelSigmaWork2PD.length!=anzVar) || (modelSigmaWork2PD[0].length!=anzVar)) modelSigmaWork2PD = new double[anzVar][anzVar];
        if ((llPathDD==null) || (llPathDD.length!=anzPar)) llPathDD = new double[anzPar];
        
        llPathD = 0;
        for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) llPathD += diffSigma[i][j]*sigInv[i][j];
        llPathD *= anzPer-1;
        double v = 0;
        for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) v += diffMu[i]*(baseMu[j]+lambda*diffMu[j])*sigInv[i][j];
        llPathD += 2*v*anzPer;
        
        for (int p=0; p<anzPar; p++)
        {
            computeSigmaInvDev(p, modelSigmaWorkPD);
            Statik.multiply(modelSigmaWorkPD, sigInv, modelSigmaWork2PD);
            v = 0;
            for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) v += diffSigma[i][j]*modelSigmaWork2PD[i][j];
            llPathDD[p] = -v*(anzPer-1);
            v = 0;
            for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) v += diffMu[i]*modelSigmaWork2PD[i][j]*(baseMu[j]+lambda*diffMu[j]);
            llPathDD[p] -= 2*v*anzPer;
            v = 0;
            computeMatrixTimesMuDev(p, sigInv, modelMuWork);
            for (int i=0; i<anzVar; i++) v += modelMuWork[i]*diffMu[i];
            llPathDD[p] -= 2*anzPer*v;
            
        }
        if (computePathDirection)
        {
            if ((pathDirection==null) || (pathDirection.length!=anzPar)) pathDirection = new double[anzPar];
            if ((modelWork1==null) || (modelWork1.length!=anzPar) || (modelWork1[0].length!=anzPar)) modelWork1 = new double[anzPar][anzPar];
            if ((modelWork2==null) || (modelWork2.length!=anzPar) || (modelWork2[0].length!=anzPar)) modelWork2 = new double[anzPar][anzPar];
            
            Statik.invert(llDD, modelWork1, modelWork2);
            Statik.multiply(modelWork1, llPathDD, pathDirection);
            for (int i=0; i<anzPar; i++) pathDirection[i] = -pathDirection[i];
        }
    }

    public double[] estimateMLByPathtracking(double[] startPos, double EPS)
    {
        final int MAXNEWTONSTEPS = 1000;
//        final double INITNOISE = 0.01;
        final double INITNOISE = 0.001;
        final int ANZCAMPS = 2;
        final int ANZSTEPS = 4;
        final double BONUSSTEPS = 0.01;
        final int MAXSTEPS = 7;
        final int ABSOLUTEMAXSTEPS = 50;
        final int TOPCAMPMAXSTEPS = 5;
        // Results that are better than ACCLEVEL times the difference between basecamp and the linear
        // interpolated ll - outcome are accepted as successfully converged. A higher acceptance factor leads to more errors on the path
        // tracking, but less intermediate steps. 
        final double ACCEPTANCEFACTOR = 1.5;        
        final int LASTTRIES = 50;
                                                    
        if (startPos == null) startPos = getArbitraryStartingValues();

        if ((modelSigmaWorkPD==null) || (modelSigmaWorkPD.length!=anzVar) || (modelSigmaWorkPD[0].length!=anzVar)) modelSigmaWorkPD = new double[anzVar][anzVar];
        if ((modelSigmaWork2PD==null) || (modelSigmaWork2PD.length!=anzVar) || (modelSigmaWork2PD[0].length!=anzVar)) modelSigmaWork2PD = new double[anzVar][anzVar];

        double[][] diffSigma = new double[anzVar][anzVar];
        double[] diffMu = new double[anzVar];
        
        this.setParameter(startPos); getMinusTwoLogLikelihood();
        Model baseCamp = this.copy();
        baseCamp.modelPos = Statik.copy(startPos);
//        for (int i=0; i<anzPar; i++) baseCamp.modelPos[i] += rand.nextGaussian()*INITNOISE;     // TODO reinsert; random removed for DEBUG
        baseCamp.setParameter(baseCamp.modelPos);
        baseCamp.pathLambda = 0.0;
        baseCamp.computeLogLikelihoodDerivatives(null, true);
        double startLS = baseCamp.getLeastSquares();
        for (int i=0; i<anzVar; i++) baseCamp.dataMean[i] = baseCamp.mu[i];
        for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) baseCamp.dataCov[i][j] = baseCamp.sigma[i][j];
        for (int i=0; i<anzVar; i++) diffMu[i] = dataMean[i] - baseCamp.dataMean[i];
        for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) diffSigma[i][j] = dataCov[i][j] - baseCamp.dataCov[i][j];
        baseCamp.computeMomentsFromDataCovarianceAndMean();
        baseCamp.getMinusTwoLogLikelihood(null, false);
        baseCamp.computePathDerivatives(0.0, baseCamp.dataCov, diffSigma, baseCamp.dataMean, diffMu, true);
        
        Model[] spawn = new Model[ANZCAMPS];
        double[] acceptanceLevel = new double[ANZCAMPS];
        
        int anzSpawns = 0;

        this.setParameter(baseCamp.modelPos);
        this.initEstimation(null, true);
        double startValue = ll;
        if (Double.isNaN(ll)) this.warningFlag = warningFlagTypes.FAILED;
        double epssqr = EPS*EPS;
        
        int newtonSteps = 0, spawningRounds = 0;
        if (logStream!=null) logStream.println("(lambda, steps, ll, convergenceIndex, acceptanceLevel, alive)\r\nFunction Distance = "+startLS);
        while ((newtonSteps<=MAXNEWTONSTEPS) && ((anzSpawns==0) || (spawn[0].pathLambda<1.0) || 
                                                 (spawn[0].lastSteplength > EPS) || (spawn[0].warningFlag != warningFlagTypes.OK)))
        {
            int newCamp = -1;
            int localMaxSteps = Math.min(MAXSTEPS + (int)Math.round(Math.pow(2,spawningRounds) * BONUSSTEPS),ABSOLUTEMAXSTEPS);
            int localAnzSteps = Math.max(ANZSTEPS, localMaxSteps/ANZCAMPS);
            int runs = 0;
            while ((newCamp==-1) && (runs < localAnzSteps) && (newtonSteps <= MAXNEWTONSTEPS))
            {
                for (int spnr = 0; (newCamp==-1 && spnr < anzSpawns); spnr++)
                    if ((spawn[spnr].warningFlag == warningFlagTypes.OK) && (spawn[spnr].lastSteplength > epssqr) && 
                       ((spawn[spnr].pathLambda<1.0) || (spawn[spnr].steps < TOPCAMPMAXSTEPS)) && ((spawn[spnr].steps < localMaxSteps)))
                    {
                        if (!spawn[spnr].moveWithOptimalDamping(epssqr, true)) spawn[spnr].warningFlag = warningFlagTypes.FAILED;
                        else if ((spawn[spnr].lastSteplength < epssqr) && (spawn[spnr].ll<acceptanceLevel[spnr])) newCamp= spnr;
                        newtonSteps++;
                    }
                for (int i=0; i<anzSpawns; i++)
                    if (logStream!=null) logStream.print("("+Statik.doubleNStellen(spawn[i].pathLambda,2)+","+
                                         spawn[i].steps+","+
                                         Statik.doubleNStellen(spawn[i].ll,2)+","+
                                         Statik.doubleNStellen(spawn[i].lastSteplength,2)+","+
                                         Statik.doubleNStellen(acceptanceLevel[i],2)+","+
                            (spawn[i].warningFlag != warningFlagTypes.OK?"DEAD":(spawn[i].lastSteplength>epssqr?"RUNS":"DONE"))+") - ");
                if (logStream!=null) logStream.println();
                runs++;
            }

            if ((anzSpawns > 0) && (spawn[0].pathLambda==1.0) && (spawn[0].ll < this.ll)) {this.setParameter(spawn[0].modelPos); this.ll = spawn[0].ll;}
            if (newCamp == -1)
            {
                // new spawn
                spawningRounds++;
                if (anzSpawns == ANZCAMPS) {
                    anzSpawns--; 
                    Model t = spawn[0]; for (int i=0; i<anzSpawns; i++) {spawn[i] = spawn[i+1]; acceptanceLevel[i] = acceptanceLevel[i+1];}
                    spawn[anzSpawns] = t;                    
                }

                if (spawn[anzSpawns]==null) spawn[anzSpawns] = this.copy();
                spawn[anzSpawns].pathLambda = (anzSpawns==0?1.0:spawn[anzSpawns-1].pathLambda/2.0);
                for (int i=0; i<anzVar; i++) spawn[anzSpawns].dataMean[i] = baseCamp.dataMean[i] + spawn[anzSpawns].pathLambda*diffMu[i];
                for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) 
                    spawn[anzSpawns].dataCov[i][j] = baseCamp.dataCov[i][j] + spawn[anzSpawns].pathLambda*diffSigma[i][j];
                spawn[anzSpawns].computeMomentsFromDataCovarianceAndMean();
                
                spawn[anzSpawns].initEstimation(baseCamp.modelPos, true);
                for (int i=0; i<anzPar; i++) spawn[anzSpawns].modelMove[i] = spawn[anzSpawns].pathLambda * baseCamp.pathDirection[i];
                spawn[anzSpawns].warningFlag = warningFlagTypes.OK;
                spawn[anzSpawns].moveWithOptimalDampingPregivenStep(epssqr, true, false);  // Optimaler Step in Richtung des Pfades
                newtonSteps++;
                
                acceptanceLevel[anzSpawns] = baseCamp.ll + spawn[anzSpawns].pathLambda*baseCamp.llPathD *
                                               (baseCamp.llPathD<0?1.0/ACCEPTANCEFACTOR:ACCEPTANCEFACTOR);
                anzSpawns++;
            } else {
                // new basecamp
                try {
                    // These two lines correct the result to the parameter result to reinsure global minimum, on the risk of not
                    // moving forward anymore.
//                    for (int i=0; i<anzVar; i++) spawn[newCamp].dataMean[i] = spawn[newCamp].mu[i];
//                    for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) spawn[newCamp].dataCov[i][j] = spawn[newCamp].sigma[i][j];
                    
                    for (int i=0; i<anzVar; i++) diffMu[i] = dataMean[i] - spawn[newCamp].dataMean[i];
                    for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) diffSigma[i][j] = dataCov[i][j] - spawn[newCamp].dataCov[i][j];
                    spawn[newCamp].computePathDerivatives(0.0, spawn[newCamp].dataCov, diffSigma, spawn[newCamp].dataMean, diffMu, true);
                    spawningRounds = 0;
                    Model t = baseCamp; baseCamp = spawn[newCamp]; spawn[newCamp] = t;
                    anzSpawns = 0;
                    double ls = 0;
                    for (int i=0; i<anzVar; i++) ls += (dataMean[i] - baseCamp.dataMean[i])*(dataMean[i] - baseCamp.dataMean[i]);
                    for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) 
                        ls += (dataCov[i][j] - baseCamp.dataCov[i][j])*(dataCov[i][j] - baseCamp.dataCov[i][j]);
                    if (logStream!=null) logStream.println("Establishing new basecamp at lambda = "+baseCamp.pathLambda+", function distance = "+ls);
                    baseCamp.pathLambda = 0;
//                    if (logStream != null) logStream.println("New Basecamp Variance of Slope = "+((CholeskyModel)baseCamp).facCov[1][1]);
                    if (logStream != null) logStream.println("New Basecamp Parameters = "+Statik.matrixToString(baseCamp.modelPos));
                } catch (Exception e)
                {
                    // new camp was not fit for basecamp, possilby because of singular Hessian. 
                    spawn[newCamp].warningFlag = warningFlagTypes.FAILED;
                }
            }
        }

        this.initEstimation(null, true);
        while ((this.lastSteplength > EPS) && (this.steps<LASTTRIES)) this.moveWithOptimalDamping(EPS, true);
        if (newtonSteps > MAXNEWTONSTEPS) 
        {
            
            if (logStream!=null) logStream.println("It took me too many iterations in the pathtracking, the last steplength for "+
                    "the original model was "+lastSteplength); 
            warningFlag = warningFlagTypes.FAILED;
        } else {
            double value = ll; 
            if ((value > startValue) || (Double.isNaN(value)) || (Double.isInfinite(value))) {
                if (logStream!=null) logStream.println("I was unable to improve the starting -2ll ("+startValue+"), converged on "+value+"."); 
                warningFlag = warningFlagTypes.FAILED;
            }  
            else if (warningFlag == warningFlagTypes.SUSPICIOUS) 
            {
                if (logStream!=null) logStream.println("It seemed I was able to solve out the problem, convergences reached on -2ll = "+
                        value+", better than starting ("+startValue+".)");
            }
        }
        if (logStream!=null) logStream.println(newtonSteps+" steps needed overall.");
        if ((lastEstimate == null) || (lastEstimate.length != anzPar)) lastEstimate = new double[anzPar];
        double[] erg = new double[anzPar];
        for (int i=0; i<anzPar; i++) lastEstimate[i] = erg[i] = modelPos[i];
        steps = newtonSteps;
        return erg;
    }
}
