/*
 * Created on 26.04.2006
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package lgmsimulation;

import java.io.*;
import java.util.Arrays;
import java.util.Vector;

import silvers.Statik;

/**
 * @author timo
 */
public class MultipleGroupFactorModel 
{
    static int NOPARAMETER = -1, NOTCOMPUTED = -1;
    static final double LNTWOPI = Math.log(2*Math.PI);

      int anzFac;
      int anzPar;
      int anzVar;
      int anzGroups;
      
      double ll;
      double[] llD;
      double[][] llDD;
      
      FactorModel[] submodel;
      
      // Parameter number maps to downscale the submodels. Each Model has a list of global parameter numbers
      // corresponding to the index, which is the local parameter number
      int[][] parmap;
      
      // Some working variables
      double[][] sigInv;
      
      public double[][] fisherInformationMatrix;
      
      public double lastSaturated;
      public double lastIndependend;
      public double lastEqualCovariancesSaturated;
      public double[] lastEstimate;
      

      public MultipleGroupFactorModel(FactorModel[] submodel)
      {
          this.submodel = submodel;
          anzGroups = submodel.length;
          parmap = new int[anzGroups][];
          anzPar = 0; for (int i=0; i<anzGroups; i++) anzPar = Math.max(anzPar, submodel[i].maxParNumber()+1);
          for (int i=0; i<anzGroups; i++) 
          {
              int[] activePars = submodel[i].getActiveParameters();
              parmap[i] = new int[activePars.length];
              for (int j=0; j<parmap[i].length; j++) parmap[i][j] = activePars[j];              
          }
          
          anzFac = 0; for (int i=0; i<anzGroups; i++) anzFac += submodel[i].anzFac;
          anzVar = 0; for (int i=0; i<anzGroups; i++) anzVar += submodel[i].anzVar;

          fisherInformationMatrix = new double[anzPar][anzPar];
      }

      public MultipleGroupFactorModel(MultipleGroupFactorModel toCopy)
      {
          anzFac = toCopy.anzFac;
          anzVar = toCopy.anzVar;
          anzPar = toCopy.anzPar;
          anzGroups = toCopy.anzGroups;
          parmap = new int[toCopy.parmap.length][];
          for (int i=0; i<parmap.length; i++)
          {
              parmap[i] = new int[toCopy.parmap[i].length];
              for (int j=0; j<parmap[i].length; j++) parmap[i][j] = toCopy.parmap[i][j]; 
          }
          
          submodel = new FactorModel[anzGroups];
          for (int i=0; i<submodel.length; i++) submodel[i] = new FactorModel(toCopy.submodel[i]);
            
          computeMoments();
          fisherInformationMatrix = new double[anzPar][anzPar];
      }
      
      public int maxParNumber()
      {
          return anzPar-1;
      }
    
    public int getAnzFac() {
        return anzFac;
    }
    public int getAnzPar() {
        return anzPar;
    }
    public int getAnzVar() {
        return anzVar;
    }
    
    public double[][] getData()
    {
        Vector<double[]> erg = new Vector<double[]>();
        for (int i=0; i<anzGroups; i++) 
        {
            double[][] data = submodel[i].data;
            for (int j=0; j<data.length; j++)
            {
                double[] zwerg = new double[data[j].length+1];
                zwerg[0] = i;
                for (int k=0; k<data[i].length; k++) zwerg[k+1] = data[j][k];
                erg.add(zwerg);
            }
        }
        double[][] toterg = new double[erg.size()][];
        for (int i=0; i<toterg.length; i++) toterg[i] = (double[])erg.elementAt(i);
        return toterg;
    }
    
    public void setData(double[][] data)
    {
        Vector[] erg = new Vector[anzGroups];
        for (int i=0; i<anzGroups; i++) erg[i] = new Vector<double[]>();
        
        for (int i=0; i<data.length; i++) 
        {
            int group = (int)Math.round(data[i][0]);
            double[] dataset = new double[data[i].length-1];
            for (int j=0; j<dataset.length; j++) dataset[j] = data[i][j+1];
            erg[group].add(dataset);
        }
        for (int i=0; i<anzGroups; i++)
        {
            double[][] dat = new double[erg[i].size()][submodel[i].anzVar];
            for (int j=0; j<dat.length; j++)
                dat[j] = (double[])erg[i].elementAt(j);
            submodel[i].data = dat;
            submodel[i].anzPer = dat.length;
        }
        computeMoments();
    }

    public int getAnzPer() {int erg = 0; for (int i=0; i<anzGroups; i++) erg += submodel[i].anzPer; return erg;}

    public void applyParameterValues(double[] vals)
    {
        for (int i=0; i<anzGroups; i++)
            submodel[i].applyParameterValues(vals);
    }

    public void createData (int anzPersons) {for (int i=0; i<anzGroups; i++) submodel[i].createData(anzPersons);}
    public void createData (int anzPersons, double[] values) {for (int i=0; i<anzGroups; i++) submodel[i].createData(anzPersons, values);}
    
    public boolean factorPositiveDefinite(double[] values)
    {
        for (int i=0; i<anzGroups; i++) if (!submodel[i].factorPositiveDefinite(values)) return false;
        return true;
    }
    
    public double euclididanDistance(double[] values)
    {
        double erg = 0;
        for (int i=0; i<anzGroups; i++) erg += submodel[i].euclididanDistance(values);
        return erg;
    }
    
    public void evaluateSigma(double[] values) {for (int i=0; i<anzGroups; i++) submodel[i].evaluateSigma(values);}
    
    /**
     * Reduces the parameter numbers of all submodels so that no non-active parameters remain. Only known call is in computerLogLikelihoodDerivatives.
     */
    private void downgradeSubmodels()
    {
        for (int i=0; i<anzGroups; i++)
            for (int j=0; j<parmap[i].length; j++) submodel[i].renumberParameter(parmap[i][j], j);
    }
    /**
     * Increases the parameter numbers of all submodels so that submodels use the same parameter space. Only known call is in computerLogLikelihoodDerivatives.
     */
    private void upgradeSubmodels()
    {
        for (int i=0; i<anzGroups; i++)
            for (int j=parmap[i].length-1; j>=0; j--) submodel[i].renumberParameter(j, parmap[i][j]);
    }
    
    public void computeMoments() {for (int i=0; i<anzGroups; i++) submodel[i].computeMoments();}
    
    public double getMinusTwoLogLikelihood(double[] value) {return getMinusTwoLogLikelihood(value, true);}
    public double getMinusTwoLogLikelihood(double[] value, boolean recomputeSigma)
    {
        ll = 0;
        for (int i=0; i<anzGroups; i++) ll += submodel[i].getMinusTwoLogLikelihood(value, recomputeSigma);
        return ll;        
    }
    
    public void computeLogLikelihoodDerivatives(double[] value)
    {
        if (value != null) applyParameterValues(value);
        evaluateSigma(value);
        
        if ((llD==null) || (llD.length!=anzPar)) llD = new double[anzPar];
        if ((llDD==null) || (llDD.length!=anzVar) || (llDD[0].length!=anzVar)) llDD = new double[anzPar][anzPar];
        
        ll = 0; for (int i=0; i<anzPar; i++) {llD[i] = 0; for (int j=0; j<anzPar; j++) llDD[i][j] = 0;} 
        
        downgradeSubmodels();
        for (int i=0; i<anzGroups; i++)
        {
            submodel[i].computeLogLikelihoodDerivatives(null, false);
            ll += submodel[i].ll;
            for (int j=0; j<submodel[i].anzPar; j++) llD[parmap[i][j]] += submodel[i].llD[j]; 
            for (int j=0; j<submodel[i].anzPar; j++) 
                for (int k=0; k<submodel[i].anzPar; k++) 
                    llDD[parmap[i][j]][parmap[i][k]] += submodel[i].llDD[j][k]; 
        }
        upgradeSubmodels();
    }
    
    public int restrictedDF()
    {
        int df = 0;
        for (int i=0; i<anzGroups; i++) df += submodel[i].anzVar*(submodel[i].anzVar+1)/2;
        return df - anzPar;
    }
    
    public double getSaturatedLL()
    {
        double saturatedLL = 0;
        for (int i=0; i<anzGroups; i++) saturatedLL += submodel[i].getSaturatedLL();
        return saturatedLL;
    }
    
    public double getEqualCovarianceMatrixNullModellLL()
    {
        double myll = ll;
        int anzVar = submodel[0].anzVar;
        for (int i=0; i<anzGroups; i++) if (submodel[i].anzVar != anzVar) {System.out.println("Error in Equal Covariance Null Model: Groups have different number of variables."); return 0;}
        double[][] centralSigma = new double[anzVar][anzVar]; 
        int totPer = 0; for (int k=0; k<anzGroups; k++) totPer += submodel[k].anzPer;
        for (int i=0; i<anzVar; i++)
            for (int j=0; j<anzVar; j++)
            {
                centralSigma[i][j] = 0;
                for (int k=0; k<anzGroups; k++) centralSigma[i][j] += submodel[k].dataCov[i][j] * submodel[k].anzPer;
                centralSigma[i][j] /= (double)totPer;
            }
        double ecmnmll = 0;
        for (int k=0; k<anzGroups; k++) 
        {
            if (submodel[k].sigma==null) submodel[k].sigma = new double[anzVar][anzVar];
            for (int i=0; i<anzVar; i++) for (int j=0; j<anzVar; j++) submodel[k].sigma[i][j] = centralSigma[i][j];
            ecmnmll += submodel[k].getMinusTwoLogLikelihood(null, false);
        }
        ll = myll;
        return ecmnmll;
    }
    
    public double getChisquareIndex()
    {
        return ll - getSaturatedLL();
    }
    
    public double getRMSEA()
    {
        return Math.sqrt(Math.max((getChisquareIndex() - restrictedDF())/(getAnzPer()-1),0) / restrictedDF());
    }
    
    public double getDeltaRMSEA(MultipleGroupFactorModel restrictedModel)
    {
        return Math.sqrt(((restrictedModel.ll - ll) - (restrictedModel.anzPar - anzPar))/((restrictedModel.anzPar - anzPar)*(getAnzPer()-1)));
    }
    
    public double getCFI()
    {
        double myll = ll;
        double satLL = getSaturatedLL();
        double mychi = ll-satLL;
        independentSigma();
        double indepLL = this.getMinusTwoLogLikelihood(null, false);
        double indepChi = indepLL - satLL;
        int restrictedDF = restrictedDF();
        int indepRestDF = restrictedDF + anzPar - anzVar;
        if (mychi <= restrictedDF) return 1.0;
        ll = myll;
        return 1.0-((mychi-restrictedDF) / Math.max(mychi-restrictedDF,indepChi - indepRestDF));
    }

    public double getCFI(MultipleGroupFactorModel restrictedModel)
    {
        double mychi = getChisquareIndex();
        if (mychi <= restrictedDF()) return 1.0;
        return 1.0-((mychi-restrictedDF()) / Math.max(mychi-restrictedDF(),restrictedModel.getChisquareIndex() - restrictedModel.restrictedDF()));
    }
    
    public double getDeltaCFI(MultipleGroupFactorModel restrictedModel)
    {
        return this.getCFI() - restrictedModel.getCFI();
    }
    
    public double getMcDonaldsCentralityIndex()
    {
        return Math.exp(-0.5*(getChisquareIndex()-restrictedDF())/(getAnzPer()-1));
    }
    public double getIndependentFit() 
    {
        double myll = ll;
        independentSigma();
        double erg = this.getMinusTwoLogLikelihood(null,false);
        ll = myll;
        return erg;
    }
    private void independentSigma() 
    {
        for (int i=0; i<anzGroups; i++) submodel[i].independentSigma();
    }
    
    /**
     * Computes the optimal damping in the direction of move from pos. Result is returned in pos, move is destroyed. 
     * 
     * @param pos   Initially position, return value
     * @param move  Initially direction, is destroyed in process
     * @return      true if successfull, false if stuck or negative result
     */
    private boolean moveWithOptimalDamping(double[] pos, double[] move, double EPS)
    {
        double[] t;
        double start = getMinusTwoLogLikelihood(pos);
        double min = start; // Value at pos;
        double alt;      // Value at move;
        double pre;      // Value at last position of move.
        
//        System.out.println("POS  = "+Statik.matrixToString(pos)+"\t\t"+start);
//        System.out.println("MOVE = "+Statik.matrixToString(move));
        
        double derivative = Statik.multiply(llD, move);
        Statik.multiply((int)Math.signum(-derivative), move, move);
        Statik.add(pos,move,move);
        alt = getMinusTwoLogLikelihood(move);
        pre = min;

        int inRuns = 0;
        while ((!Double.isNaN(alt)) && (!Double.isInfinite(alt)) && (alt < pre))
        {
            pre = alt;
            for (int i=0; i<move.length; i++) move[i] = 2*move[i] - pos[i];
            alt = getMinusTwoLogLikelihood(move);
            inRuns++;
            if (inRuns > 20)
            {
                System.out.println("Got stuck in increasing steps, min = "+alt+", start = ");
                if (inRuns > 100) {System.out.println("Giving up."); return false;}
            }
        }
        
        if ((pre != min) && (alt > min)) {for (int i=0; i<move.length; i++) move[i] = (move[i] + pos[i]) / 2.0; alt = pre;}
        pre = alt+EPS*EPS;
        
        inRuns = 0;
        while ((Double.isNaN(alt)) || (Double.isInfinite(alt)) || (Double.isNaN(pre)) || (Double.isInfinite(pre)) || (alt < pre) || (start < alt)) {
            pre = alt;
            if (pre < min) {min = pre; t = pos; pos = move; move = t;}
            for (int i=0; i<move.length; i++) move[i] = (move[i] + pos[i]) / 2.0;
            alt = getMinusTwoLogLikelihood(move);
            inRuns++;
            if (inRuns > 20)
            {
                System.out.println("Got stuck in interval search, min = "+min+", alternative = "+alt+", start = "+start);
                if (inRuns > 100) {System.out.println("Giving up."); return false;}
            }
        }
        for (int i=0; i<move.length; i++) move[i] = pos[i]; // Since pos and move may have been exchanged.
        if (min < 0) {System.out.println("Heading towards boundary minimum: "+min); return false;}
        
//        System.out.println("OPT  = "+Statik.matrixToString(pos)+"\t\t"+min);
//        System.out.println();
        return true;
    }
    
    public double[] estimateML(double[] starting) {return estimateML(starting, 0.0001);}
    public double[] estimateML(double[] starting, double EPS)
    {
        final int MAXRUNS = 50; 

        double[] move = new double[anzPar];
        double[] pos = new double[anzPar];
        double[][] work1 = new double[anzPar][anzPar], work2 = new double[anzPar][anzPar];
        
        for (int i=0; i<anzPar; i++) pos[i] = starting[i];
        
        int runs = 0; double steplength = Double.POSITIVE_INFINITY, gain = Double.POSITIVE_INFINITY;
        computeLogLikelihoodDerivatives(pos);
        double DEBUGstart = ll;
        while ((runs<=MAXRUNS) && (steplength > EPS) && (gain > EPS))
        {
            double lastLL = ll;
            Statik.invert(llDD, work1, work2);
            Statik.multiply(work1, llD, move);
            boolean success = moveWithOptimalDamping(pos, move, EPS);

            if (success)
            {
                computeLogLikelihoodDerivatives(pos);
                steplength = Statik.abs(move);
                gain = lastLL - ll;
                runs++;
            } else {
                for (int i=0; i<anzPar; i++) pos[i] = starting[i];
                computeLogLikelihoodDerivatives(pos);
                Statik.invert(llDD, work1, work2);
                Statik.multiply(work1, llD, move);
                double abs = Statik.abs(move);
                for (int i=0; i<anzPar; i++) pos[i] += starting[i] + submodel[0].rand.nextDouble() * abs / 10.0;
                for (int i=0; i<anzGroups; i++) 
                    if (submodel[i].errPar != FactorModel.NOPARAMETER) pos[submodel[i].errPar] += abs / 2.0;
                System.out.println("Headed towards boundary minimum, I restarted with perturbed values with higher error.");
            }
        }

//        if (runs > MAXRUNS) System.out.println("Warning: More than 50 runs needed, final steplength = "+steplength);
//        System.out.println("Runs       = "+runs);
//        System.out.println("Steplength = "+steplength);
//        System.out.println("Last gain  = "+gain);
//        System.out.println("Total gain = "+(DEBUGstart - ll));
//        System.out.println("Final LL   = "+ll);
        
        if ((lastEstimate == null) || (lastEstimate.length != anzPar)) lastEstimate = new double[anzPar];
        for (int i=0; i<anzPar; i++) lastEstimate[i] = pos[i];
        return pos;
    }
    
    public void setDataAsIn(MultipleGroupFactorModel model) {
        for (int i=0; i<anzGroups; i++) submodel[i].setData(model.submodel[i].data);
    }

    public static void bigSimulation(int simulation, int startCond, int endCond, int anzRep)
    {
        final double EPS = 0.001;
        final int anzItems = 5;
        final double facVar = 0.64;
        final double facCovVar = facVar*facVar*0.3;
        final double usualLoading = 0.6 / Math.sqrt(facVar);

        String header = "CONDITION\tTRIAL\tFULL_FIT\tFULL_INDEPENDENT_FIT\tFULL_SATURATED_FIT\tFULL_EQUALCOVARIANCEMATRICES_FIT\tFULL_RESTRICTED_DFS\t";
        header += "RES1_FIT\tRES1_INDEPENDENT_FIT\tRES1_SATURATED_FIT\tRES1_EQUALCOVARIANCEMATRICES_FIT\tRES1_RESTRICTED_DFS\t";
        header += "RES2_FIT\tRES2_INDEPENDENT_FIT\tRES2_SATURATED_FIT\tRES2_EQUALCOVARIANCEMATRICES_FIT\tRES2_RESTRICTED_DFS\r\n";
        
        try {
            BufferedWriter w = new BufferedWriter(new FileWriter("MultiGroupFactorCond"+startCond+"TO"+endCond+".txt"));
            w.write(header); w.flush();
            long startTime = System.currentTimeMillis();
            if (simulation == 1) // conditions range from 0 to 119
            {
                for (int condNr = startCond; condNr <= endCond; condNr++)
                {
                    int cond = condNr;
                    long time = System.currentTimeMillis() - startTime;
                    double percentageFinished = ((double)cond/(double)(endCond-startCond+1));
                    System.out.println("Starting condition "+condNr+" ("+Statik.doubleNStellen(100*percentageFinished,2)+"% finished, "+time+" ms needed, "+Statik.doubleNStellen(time * (1.0-percentageFinished) / percentageFinished, 0)+" ms to go)");
                    int smpC = cond % 3; cond /= 3;
                    int effC = cond % 5; cond /= 5;
                    int nfaC = cond % 4; cond /= 4;
                    int equC = cond % 2;
                    
                    int N = (smpC==0?100:(smpC==1?500:1000));
                    double manip = effC*0.05;
                    int nfac = 1+nfaC;
                    int[][] facCovPar = new int[nfac][nfac];
    
                    int pnr = 1;
                    for (int i=0; i<nfac; i++) 
                        for (int j=i+1; j<nfac; j++) facCovPar[i][j] = facCovPar[j][i] = pnr++;
                    for (int i=0; i<nfac; i++) facCovPar[i][i] = -1;
                    
                    int[][] loadingsParA = new int[nfac][anzItems], loadingsParB = new int[nfac][anzItems], 
                            loadingsParB1 = new int[nfac][anzItems], loadingsParB2 = new int[nfac][anzItems];
                    for (int i=0; i<nfac; i++)
                        for (int j=0; j<anzItems; j++) loadingsParA[i][j] = pnr++;
                    for (int i=0; i<nfac; i++)
                        for (int j=0; j<anzItems; j++) loadingsParB[i][j] = pnr++;
                    for (int i=0; i<nfac; i++)
                        for (int j=0; j<anzItems; j++) loadingsParB1[i][j] = (i==0?loadingsParA[i][j]:loadingsParB[i][j]-anzItems);
                    for (int i=0; i<nfac; i++)
                        for (int j=0; j<anzItems; j++) loadingsParB2[i][j] = loadingsParA[i][j];
                    
                    double[][] loadingsValA = new double[nfac][anzItems], loadingsValB = new double[nfac][anzItems];
                    for (int i=0; i<nfac; i++)
                        for (int j=0; j<anzItems; j++) 
                        {
                            loadingsValA[i][j] = usualLoading + ((((i==0) || (equC==0)) && (j<2))? (j==0?+manip:-manip) : 0.0);   
                            loadingsValB[i][j] = usualLoading + ((((i==0) || (equC==0)) && (j<2))? (j==0?-manip:+manip) : 0.0);
                        }
                    
                    double[][] facCov = new double[nfac][nfac];
                    for (int i=0; i<nfac; i++) for (int j=0; j<nfac; j++) facCov[i][j] = (i==j?facVar:facCovVar);
    
                    int anzParFull = 1 + nfac*(nfac+1)/2 + 2 * nfac * anzItems;   
                    double[] startingFull = new double[anzParFull], 
                             startingRes1 = new double[anzParFull - anzItems], 
                             startingRes2 = new double[anzParFull - nfac*anzItems];
                    startingFull[0] = startingRes1[0] = startingRes2[0] = 1.0;
                    pnr = 1;
                    for (int i=0; i<nfac; i++) for (int j=i+1; j<nfac; j++) {startingFull[pnr] = startingRes1[pnr] = startingRes2[pnr] = facCov[i][j]; pnr++;} 
                    for (int i=0; i<nfac; i++)
                        for (int j=0; j<anzItems; j++)
                        {
                            startingFull[pnr+i*anzItems+j] = loadingsValA[i][j]; startingFull[pnr+(i+nfac)*anzItems+j] = loadingsValB[i][j];
                            if (i>0) {startingRes1[pnr+i*anzItems+j] = loadingsValA[i][j]; startingRes1[pnr+(i-1+nfac)*anzItems+j] = loadingsValB[i][j];}
                            else startingRes1[pnr+j] = (loadingsValA[i][j] + loadingsValB[i][j]) / 2.0;
                            startingRes2[pnr+i*anzItems+j] = (loadingsValA[i][j] + loadingsValB[i][j]) / 2.0;
                        }
                    
                    
                    FactorModel A = new FactorModel(loadingsParA, loadingsValA, facCovPar, facCov, 0, 1.0);
                    FactorModel B = new FactorModel(loadingsParB, loadingsValB, facCovPar, facCov, 0, 1.0);
                    FactorModel B1 = new FactorModel(loadingsParB1, loadingsValB, facCovPar, facCov, 0, 1.0);
                    FactorModel B2 = new FactorModel(loadingsParB2, loadingsValB, facCovPar, facCov, 0, 1.0);
                    MultipleGroupFactorModel model = new MultipleGroupFactorModel(new FactorModel[]{A,B});
                    MultipleGroupFactorModel modelRes1 = new MultipleGroupFactorModel(new FactorModel[]{A,B1});
                    MultipleGroupFactorModel modelRes2 = new MultipleGroupFactorModel(new FactorModel[]{A,B2});
    
                    for (int rep=0; rep<anzRep; rep++)
                    {
                        model.applyParameterValues(startingFull);
                        model.createData(N);
                        modelRes1.setDataAsIn(model);
                        modelRes2.setDataAsIn(model);
                    
                        try {
                            double[] est1 = model.estimateML(startingFull, EPS);
                            double[] est2 = modelRes1.estimateML(startingRes1, EPS);
                            double[] est3 = modelRes2.estimateML(startingRes2, EPS);
                            
    //                        System.out.println(Statik.matrixToString(est1));
    //                        System.out.println(Statik.matrixToString(est2));
    //                        System.out.println(Statik.matrixToString(est3));
    
                            w.write(condNr+"\t"+rep+"\t");
                            w.write(model.ll+"\t"+model.getIndependentFit()+"\t"+model.getSaturatedLL()+"\t"+model.getEqualCovarianceMatrixNullModellLL()+"\t"+model.restrictedDF()+"\t");
                            w.write(modelRes1.ll+"\t"+modelRes1.getIndependentFit()+"\t"+modelRes1.getSaturatedLL()+"\t"+modelRes1.getEqualCovarianceMatrixNullModellLL()+"\t"+modelRes1.restrictedDF()+"\t");
                            w.write(modelRes2.ll+"\t"+modelRes2.getIndependentFit()+"\t"+modelRes2.getSaturatedLL()+"\t"+modelRes2.getEqualCovarianceMatrixNullModellLL()+"\t"+modelRes2.restrictedDF()+"\r\n");
                            
                        } catch (Exception e)
                        {
                            System.out.println("I hit exception "+e+", returning -999 as missing indicator.");
                            w.write(cond+"\t"+rep+"\t");
                            for (int i=0; i<15; i++) w.write("-999"+(i==14?"\r\n":"\t"));
                        }
                    }
                    w.flush();
                }
            }
            long time = System.currentTimeMillis() - startTime;
            System.out.println("Finished, "+time+" ms needed");
            w.close();
        } catch (Exception e)
        {
            System.out.println("I hit an unexpected exception, abnormal termination: "+e);
        }
    }
    
    /**
     * deprecated.
     * @param erg
     */
    public void fillCompleteDescription(double[] erg)
    {
        int p = 0;
        for (int i=0; i<lastEstimate.length; i++) erg[p++] = lastEstimate[i];
        erg[p++] = ll;
        erg[p++] = getIndependentFit();
        erg[p++] = getSaturatedLL();
        erg[p++] = getEqualCovarianceMatrixNullModellLL();
        erg[p++] = restrictedDF();
    }
    
    /**
     * deprecated
     * @return
     */
    public int getDescriptionLength() {return lastEstimate.length + 5;}
    public String getHeaderLine() 
    {
        String erg = "";
        for (int i=0; i<lastEstimate.length; i++) erg += "ESTIMATE"+i+"\t";
        erg += "FIT\t";
        erg += "INDEPENDENT_FIT\t";
        erg += "SATURATED_FIT\t";
        erg += "EQUALCOVARIANCEMATRICES_FIT\t";
        erg += "RESTRICTED_DFS\t";
        return erg;
    }
    
    /*
    public double getRMSEA()
    {
        return Math.sqrt(Math.max((getChisquareIndex() - restrictedDF())/(getAnzPer()-1),0) / restrictedDF());
    }
    
    public double getDeltaRMSEA(MultipleGroupFactorModel restrictedModel)
    {
        return Math.sqrt(((restrictedModel.ll - ll) - (restrictedModel.anzPar - anzPar))/((restrictedModel.anzPar - anzPar)*(getAnzPer()-1)));
    }
    
    public double getCFI()
    {
        double myll = ll;
        double satLL = getSaturatedLL();
        double mychi = ll-satLL;
        independentSigma();
        double indepLL = this.getMinusTwoLogLikelihood(null, false);
        double indepChi = indepLL - satLL;
        int restrictedDF = restrictedDF();
        int indepRestDF = restrictedDF + anzPar - anzVar;
        if (mychi <= restrictedDF) return 1.0;
        ll = myll;
        return 1.0-((mychi-restrictedDF) / Math.max(mychi-restrictedDF,indepChi - indepRestDF));
    }

    public double getCFI(MultipleGroupFactorModel restrictedModel)
    {
        double mychi = getChisquareIndex();
        if (mychi <= restrictedDF()) return 1.0;
        return 1.0-((mychi-restrictedDF()) / Math.max(mychi-restrictedDF(),restrictedModel.getChisquareIndex() - restrictedModel.restrictedDF()));
    }
    
    public double getDeltaCFI(MultipleGroupFactorModel restrictedModel)
    {
        return this.getCFI() - restrictedModel.getCFI();
    }
    
    public double getMcDonaldsCentralityIndex()
    {
        return Math.exp(-0.5*(getChisquareIndex()-restrictedDF())/(getAnzPer()-1));
    }
    public double getIndependentFit() 
    {
        double myll = ll;
        independentSigma();
        double erg = this.getMinusTwoLogLikelihood(null,false);
        ll = myll;
        return erg;
    }

    String header = "CONDITION\tTRIAL\tFULL_FIT\tFULL_INDEPENDENT_FIT\tFULL_SATURATED_FIT\tFULL_EQUALCOVARIANCEMATRICES_FIT\tFULL_RESTRICTED_DFS\t";
    header += "RES1_FIT\tRES1_INDEPENDENT_FIT\tRES1_SATURATED_FIT\tRES1_EQUALCOVARIANCEMATRICES_FIT\tRES1_RESTRICTED_DFS\t";
    header += "RES2_FIT\tRES2_INDEPENDENT_FIT\tRES2_SATURATED_FIT\tRES2_EQUALCOVARIANCEMATRICES_FIT\tRES2_RESTRICTED_DFS\r\n";
    */
    
    public static void evaluateResults(String path)
    {
        File p = new File(path);
        String[] files = p.list();
        double[] cfimean= new double[120], cfisquare= new double[120], cfipow1= new double[120], cfipow2= new double[120], cfipow3= new double[120],
                 mdcmean= new double[120], mdcsquare= new double[120], mdcpow1= new double[120], mdcpow2= new double[120], mdcpow3= new double[120],
                 drmmean= new double[120], drmsquare= new double[120], drmpow1= new double[120], drmpow2= new double[120],
                 crmmean= new double[120], crmsquare= new double[120], crmpow1= new double[120],
                 chipow1= new double[120];
        
        Vector[] cfiMdcDrmCrmChi = new Vector[120]; // deltacfi1 und 2, deltatmci1 und 2, deltarmsea1 und 2, crmsea1 und 2
        for (int i=0; i<120; i++) cfiMdcDrmCrmChi[i] = new Vector();
        for (int f=0; f<files.length; f++)
            if (files[f].contains("MultiGroupFactor"))
            {
                try {
                    BufferedReader b = new BufferedReader(new FileReader(new File(p, files[f])));
                    b.readLine();
                    while (b.ready()) 
                    {
                        String line = Statik.loescheRandWhitespaces(b.readLine());
                        while ((line.length()==0) && (b.ready())) line = Statik.loescheRandWhitespaces(b.readLine());
                        if (line.length()>0)
                        {
                            String[] content = line.split("\t");
                            int condition = Integer.parseInt(content[0]);
                            int cond = condition; 
                            int smpC = cond % 3; cond /= 3;
                            int effC = cond % 5; cond /= 5;
                            int nfaC = cond % 4; cond /= 4;
                            int equC = cond % 2;
                            int N = (smpC==0?100:(smpC==1?500:1000));

                            double llFull = Double.parseDouble(content[2]), llIndep= Double.parseDouble(content[3]), 
                                   llSat = Double.parseDouble(content[4]), llEqCovMat = Double.parseDouble(content[5]), 
                                   llRes1 = Double.parseDouble(content[7]), llRes2 = Double.parseDouble(content[12]);
                            int dfFull = Integer.parseInt(content[6]), dfRes1 = Integer.parseInt(content[11]), dfRes2 = Integer.parseInt(content[16]);
                            int dfIndep = (1+nfaC)*10;
                            double fullChi = llFull - llSat, res1Chi = llRes1 - llSat, res2Chi = llRes2 - llSat;
                            double cfiFull = (fullChi<dfFull?1.0:1.0-((fullChi-dfFull) / Math.max(fullChi-dfFull,llIndep - llSat - dfIndep)));
                            double cfiRes1 = (res1Chi<dfRes1?1.0:1.0-((res1Chi-dfRes1) / Math.max(res1Chi-dfRes1,llIndep - llSat - dfIndep)));
                            double cfiRes2 = (res2Chi<dfRes2?1.0:1.0-((res2Chi-dfRes2) / Math.max(res2Chi-dfRes2,llIndep - llSat - dfIndep)));
                            double mciFull = Math.exp(-0.5*(fullChi-dfFull)/(N-1));
                            double mciRes1 = Math.exp(-0.5*(res1Chi-dfRes1)/(N-1));
                            double mciRes2 = Math.exp(-0.5*(res2Chi-dfRes2)/(N-1));
                            double rmseaFull = Math.sqrt(Math.max((fullChi - dfFull)/(N-1),0) / dfFull);
                            double rmseaRes1 = Math.sqrt(Math.max((res1Chi - dfRes1)/(N-1),0) / dfRes1);
                            double rmseaRes2 = Math.sqrt(Math.max((res2Chi - dfRes2)/(N-1),0) / dfRes2);
                            double crmseaRes1 = Math.sqrt(Math.max(((llRes1 - llFull) - (dfRes1 - dfFull))/((dfRes1- dfFull)*(N-1)),0));
                            double crmseaRes2 = Math.sqrt(Math.max(((llRes2 - llFull) - (dfRes2 - dfFull))/((dfRes2- dfFull)*(N-1)),0));
                            double[] cmdcc = new double[10];
                            cmdcc[0] = cfiFull-cfiRes1; 
                            cmdcc[1] = cfiFull-cfiRes2; 
                            cmdcc[2] = mciFull-mciRes1; 
                            cmdcc[3] = mciFull-mciRes2; 
                            cmdcc[4] = rmseaRes1-rmseaFull; 
                            cmdcc[5] = rmseaRes2-rmseaFull; 
                            cmdcc[6] = crmseaRes1; 
                            cmdcc[7] = crmseaRes2; 
                            cmdcc[8] = llRes1 - llFull - Statik.FIVEPERCENTTHRESHOLD[dfRes1-dfFull-1]; 
                            cmdcc[9] = llRes2 - llFull - Statik.FIVEPERCENTTHRESHOLD[dfRes2-dfFull-1]; 
                            cfiMdcDrmCrmChi[condition].add(cmdcc); 
                        }
                    }
                    b.close();
                } catch (IOException e) 
                {
                    System.out.println("Error while reading files: "+e);
                }
            }
        double[][][] value = new double[120][10][];
        for (int i=0; i<120; i++)
        {
            for (int j=0; j<10; j++) value[i][j] = new double[cfiMdcDrmCrmChi[i].size()];
            for (int k=0; k<value[i][0].length; k++)
                for (int j=0; j<10; j++) value[i][j][k] = ((double[])cfiMdcDrmCrmChi[i].elementAt(k))[j];
        }
        double[][][][] dynamicFivePercent = new double[3][4][2][10];      // cfi, mdc, drmsea, crmsea
        for (int i=0; i<120; i++)
            for (int j=0; j<10; j++)
            {
                int cond = i; 
                int smpC = cond % 3; cond /= 3;
                int effC = cond % 5; cond /= 5;
                int nfaC = cond % 4; cond /= 4;
                int equC = cond % 2;
                 Arrays.sort(value[i][j]);
                 if (effC ==0) dynamicFivePercent[smpC][nfaC][equC][j] = value[i][j][(int)Math.round(value[i][j].length*0.95)];
            }
        /*
         *  The 36 results per cell are: 5 values for CFI (Mean, STDV, 3 power values),
         *                               5 for MDC (mean, stdv, 3 power values), 
         *                               4 for Delta-RMSEA (mean, stdv, 2 power values), 
         *                               3 for comparative RMSEA (mean, stdv, power) 
         *                               1 for the chisquare (power)
         *  and then all repeated for the second restricted model.
         */  
            
        // 1% cutoff reported by meadeEtAl08 for the mci with 1,2,3,4,5 factors and 5,10,15,20,25 items. Notice the non-monotonicity.
        final double[] MEADECRITICALMCI = new double[]{0.0065, 0.0080, 0.0058, 0.0058, 0.0086};
        double[][][][][] result = new double[3][5][4][2][36];
        for (int i=0; i<3; i++)
            for (int j=0; j<5; j++)
                for (int k=0; k<4; k++)
                    for (int l=0; l<2; l++)
                    {
                        int cond = i + 3*j + 15*k + 60*l;
                        int cfiPow11 = 0, cfiPow12 = 0, cfiPow21 = 0, cfiPow22 = 0, cfiPow31 = 0,cfiPow32 = 0,
                            mdcPow11 = 0, mdcPow12 = 0, mdcPow21 = 0, mdcPow22 = 0, mdcPow31 = 0,mdcPow32 = 0,
                            drmPow11 = 0, drmPow12 = 0, drmPow21 = 0, drmPow22 = 0, 
                            crmPow11 = 0, crmPow12 = 0, chiPow11 = 0, chiPow12 = 0;
                        for (int m=0; m<value[cond][0].length; m++)
                        {
                            result[i][j][k][l][0] += value[cond][0][m]; result[i][j][k][l][18] += value[cond][1][m];
                            result[i][j][k][l][1] += value[cond][0][m]*value[cond][0][m]; result[i][j][k][l][19] += value[cond][1][m]*value[cond][1][m];
                            result[i][j][k][l][5] += value[cond][2][m]; result[i][j][k][l][23] += value[cond][3][m];
                            result[i][j][k][l][6] += value[cond][2][m]*value[cond][2][m]; result[i][j][k][l][24] += value[cond][3][m]*value[cond][3][m];
                            result[i][j][k][l][10] += value[cond][4][m]; result[i][j][k][l][28] += value[cond][5][m];
                            result[i][j][k][l][11] += value[cond][4][m]*value[cond][4][m]; result[i][j][k][l][29] += value[cond][5][m]*value[cond][5][m];
                            result[i][j][k][l][14] += value[cond][6][m]; result[i][j][k][l][32] += value[cond][7][m];
                            result[i][j][k][l][15] += value[cond][6][m]*value[cond][6][m]; result[i][j][k][l][33] += value[cond][7][m]*value[cond][7][m];
                            if ((cfiPow11==0) && (value[cond][0][m] > 0.002)) cfiPow11 = value[cond][0].length-m;  
                            if ((cfiPow12==0) && (value[cond][1][m] > 0.002)) cfiPow12 = value[cond][0].length-m;  
                            if ((cfiPow21==0) && (value[cond][0][m] > 0.01)) cfiPow21 = value[cond][0].length-m;  
                            if ((cfiPow22==0) && (value[cond][1][m] > 0.01)) cfiPow22 = value[cond][0].length-m;  
                            if ((cfiPow31==0) && (value[cond][0][m] > dynamicFivePercent[i][k][l][0])) cfiPow31 = value[cond][0].length-m;  
                            if ((cfiPow32==0) && (value[cond][1][m] > dynamicFivePercent[i][k][l][1])) cfiPow32 = value[cond][0].length-m;  
                            if ((mdcPow11==0) && (value[cond][2][m] > MEADECRITICALMCI[k])) mdcPow11 = value[cond][0].length-m;            // Tab. 12 in Meade et al.
                            if ((mdcPow12==0) && (value[cond][3][m] > MEADECRITICALMCI[k])) mdcPow12 = value[cond][0].length-m;  
                            if ((mdcPow21==0) && (value[cond][2][m] > 0.02)) mdcPow21 = value[cond][0].length-m;  
                            if ((mdcPow22==0) && (value[cond][3][m] > 0.02)) mdcPow22 = value[cond][0].length-m;  
                            if ((mdcPow31==0) && (value[cond][2][m] > dynamicFivePercent[i][k][l][2])) mdcPow31 = value[cond][0].length-m;  
                            if ((mdcPow32==0) && (value[cond][3][m] > dynamicFivePercent[i][k][l][3])) mdcPow32 = value[cond][0].length-m;  
                            if ((drmPow11==0) && (value[cond][4][m] > 0.07)) drmPow11 = value[cond][0].length-m;  
                            if ((drmPow12==0) && (value[cond][5][m] > 0.07)) drmPow12 = value[cond][0].length-m;  
                            if ((drmPow21==0) && (value[cond][4][m] > dynamicFivePercent[i][k][l][4])) drmPow21 = value[cond][0].length-m;  
                            if ((drmPow22==0) && (value[cond][5][m] > dynamicFivePercent[i][k][l][5])) drmPow22 = value[cond][0].length-m;  
                            if ((crmPow11==0) && (value[cond][6][m] > dynamicFivePercent[i][k][l][6])) crmPow11 = value[cond][0].length-m;  
                            if ((crmPow12==0) && (value[cond][7][m] > dynamicFivePercent[i][k][l][7])) crmPow12 = value[cond][0].length-m;  
                            if ((chiPow11==0) && (value[cond][8][m] > 0)) chiPow11 = value[cond][0].length-m;  
                            if ((chiPow12==0) && (value[cond][9][m] > 0)) chiPow12 = value[cond][0].length-m;  
                        }
                        double anz = (double)value[cond][0].length;
                        result[i][j][k][l][0] /= anz; result[i][j][k][l][18] /= anz;
                        result[i][j][k][l][1] = Math.sqrt(result[i][j][k][l][1] /anz - result[i][j][k][l][0]*result[i][j][k][l][0]);
                        result[i][j][k][l][19] = Math.sqrt(result[i][j][k][l][19] /anz - result[i][j][k][l][18]*result[i][j][k][l][18]);
                        result[i][j][k][l][5] /= anz; result[i][j][k][l][23] /= anz;
                        result[i][j][k][l][6] = Math.sqrt(result[i][j][k][l][6] /anz - result[i][j][k][l][5]*result[i][j][k][l][5]);
                        result[i][j][k][l][24] = Math.sqrt(result[i][j][k][l][24] /anz - result[i][j][k][l][23]*result[i][j][k][l][23]);
                        result[i][j][k][l][10] /= anz; result[i][j][k][l][28] /= anz;
                        result[i][j][k][l][11] = Math.sqrt(result[i][j][k][l][11] /anz - result[i][j][k][l][10]*result[i][j][k][l][10]);
                        result[i][j][k][l][29] = Math.sqrt(result[i][j][k][l][29] /anz - result[i][j][k][l][28]*result[i][j][k][l][28]);
                        result[i][j][k][l][14] /= anz; result[i][j][k][l][32] /= anz;
                        result[i][j][k][l][15] = Math.sqrt(result[i][j][k][l][15] /anz - result[i][j][k][l][14]*result[i][j][k][l][14]);
                        result[i][j][k][l][33] = Math.sqrt(result[i][j][k][l][33] /anz - result[i][j][k][l][32]*result[i][j][k][l][32]);
                        result[i][j][k][l][2] = cfiPow11 / anz; result[i][j][k][l][20] = cfiPow12 / anz;
                        result[i][j][k][l][3] = cfiPow21 / anz; result[i][j][k][l][21] = cfiPow22 / anz;
                        result[i][j][k][l][4] = cfiPow31 / anz; result[i][j][k][l][22] = cfiPow32 / anz;
                        result[i][j][k][l][7] = mdcPow11 / anz; result[i][j][k][l][25] = mdcPow12 / anz;
                        result[i][j][k][l][8] = mdcPow21 / anz; result[i][j][k][l][26] = mdcPow22 / anz;
                        result[i][j][k][l][9] = mdcPow31 / anz; result[i][j][k][l][27] = mdcPow32 / anz;
                        result[i][j][k][l][12] = drmPow11 / anz; result[i][j][k][l][30] = drmPow12 / anz;
                        result[i][j][k][l][13] = drmPow21 / anz; result[i][j][k][l][31] = drmPow22 / anz;
                        result[i][j][k][l][16] = crmPow11 / anz; result[i][j][k][l][34] = crmPow12 / anz;
                        result[i][j][k][l][17] = chiPow11 / anz; result[i][j][k][l][35] = chiPow12 / anz;
                    }
        double[][] ergMatrix = new double[36*2][3*5*4];
        for (int i=0; i<3; i++)
            for (int j=0; j<5; j++)
                for (int k=0; k<4; k++)
                    for (int l=0; l<2; l++)
                        for (int m=0; m<36; m++)
                            ergMatrix[m+36*l][i+3*j+15*k] = result[i][j][k][l][m];
        Statik.writeMatrix(ergMatrix, path+"\\resultTable.txt", '\t');
                                                
    }
    
    public static void main(String args[])
    {
//        if (args.length==0) bigSimulation(1,0,10,100);
//        else bigSimulation(Integer.parseInt(args[0]),Integer.parseInt(args[1]),Integer.parseInt(args[2]),Integer.parseInt(args[3]));

        evaluateResults("C:\\Dokumente und Einstellungen\\timo\\Eigene Dateien\\Projekte\\Simson\\AFI\\Simulation\\Simulation210409");
        
//        FactorModel A = new FactorModel(new int[][]{{0,1},{2,3}}, new double[][]{{1,1},{1,1}}, 
//                new int[][]{{-1,4},{4,-1}}, new double[][]{{1.0,0.3},{0.3,1.0}}, 5, 0.2);
//        FactorModel B = new FactorModel(new int[][]{{0,1},{2,3}}, new double[][]{{1,1},{1,1}}, 
//                new int[][]{{-1,4},{4,-1}}, new double[][]{{1.0,0.3},{0.3,1.0}}, 5, 0.2);
//        MultipleGroupFactorModel model = new MultipleGroupFactorModel(new FactorModel[]{A,B});
////        model.createData(100);
////        Statik.writeMatrix(model.getData(), "ToPaoloTestData.txt", '\t');
//        model.setData(Statik.loadMatrix("DataSet2ViolatedNull.txt", '\t'));
//        double[] pop = new double[]{1,1.5,1,1.5,0.3,0.2};
////        double[] pop = new double[]{1,1.5,1,1.5,0.3,0.2};
//        double[] est = model.estimateML(pop, 0.001);
//        System.out.println("Population     = "+Statik.matrixToString(pop));
//        System.out.println("Estimates      = "+Statik.matrixToString(est));
//        System.out.println("Estimated   LL = "+model.ll);
//        System.out.println("Independent LL = "+model.getIndependentFit());
//        System.out.println("Saturated   LL = "+model.getSaturatedLL());
//        System.out.println("Equal Cov.  LL = "+model.getEqualCovarianceMatrixNullModellLL());
//        System.out.println("est - sat   LL = "+model.getChisquareIndex());
//        System.out.println("restricted dfs = "+model.restrictedDF());
//        System.out.println("CFI            = "+model.getCFI());
//        System.out.println("RMSEA          = "+model.getRMSEA());
        
        
        
//        double[] middleInitial = new double[]{1.5,0.4,0.25,2.0,0.2,2.5};
//        double[] middleAcute = new double[6];
//        int pnr = 0;
//        for (int i=0; i<3; i++) for (int j=i; j<3; j++) middleInitial[pnr++] = (A.facCovVal[i][j]+B.facCovVal[i][j])/2.0;
//        double[] est = model.estimateML(middleInitial);
//        pnr = 0;
//        for (int i=0; i<3; i++) for (int j=i; j<3; j++) middleAcute[pnr++] = (A.facCovVal[i][j]+B.facCovVal[i][j])/2.0;
//        System.out.println("Middle Population = "+Statik.matrixToString(middleInitial));
//        System.out.println("Middle Estimate   = "+Statik.matrixToString(middleAcute));
//        System.out.println("Overal Estimate   = "+Statik.matrixToString(est));
    }
    
}
