/*
 * Created on 20.05.2008
 */
package lgmsimulation;

import silvers.Statik;
import java.util.Random;

public class SteveTestclass 
{
    public static Random random = new Random();
    public static boolean FAILEDFLAG = false;
    
    public static double[][] fisherInformationMatrix(double[][] data, double[] c, double f, double err)
    {
        double[][] erg = new double[c.length+1][c.length+1];
        for (int i=0; i<erg.length; i++) for (int j=0; j<erg.length; j++) erg[i][j] = 0.0;
        for (int i=0; i<c.length; i++) // through each data row
        {
            for (int j=0; j<data[i].length; j++) 
            {
                // 2nd derivative of f
                double sinv = Math.sin(j*f+c[i]);
                double cosv = Math.cos(j*f+c[i]);
                double v = (1/err)*(sinv*(sinv - data[i][j]) - cosv*cosv);
                erg[0][0] += j*j*v;

                // 2nd derivative of each c
                erg[i+1][i+1] += v;
                
                // 2nd derivative f with corresponding c
                erg[0][i+1] += j*v;
            }
            erg[i+1][0] = erg[0][i+1];
        }
        return erg;
    }
    
    public static double estimatedSDOfFrequencyEstimate(double[][] data, double[] c, double f, double errorVariance)
    {
        double[][] fisherInformation = fisherInformationMatrix(data, c, f, errorVariance);
        double v = fisherInformation[0][0];
        for (int i=1; i<fisherInformation.length; i++) v -= fisherInformation[0][i]*fisherInformation[0][i]/fisherInformation[i][i];
        v = 1/v;    // computes the entry [0][0] of the inverse.
        return Math.sqrt(v);
    }
    
    public static double minus2llDifferenceShort(double[][] data, double[] c, double f, double fhyp, double[] chyp, double errStdv)
    {
        double erg = 0;
        for (int i=0; i<data.length; i++)
        {
            for (int j=0; j<data[i].length; j++) erg += (data[i][j]-Math.sin(j*fhyp+chyp[i]))*(data[i][j]-Math.sin(j*fhyp+chyp[i]));
            for (int j=0; j<data[i].length; j++) erg -= (data[i][j]-Math.sin(j*f+c[i]))*(data[i][j]-Math.sin(j*f+c[i]));
        }
        return erg / (errStdv*errStdv);
    }
    
    public static double optimizeF(double[][] data, double[] c, double starting, double eps)
    {
        final int MAXITER = 30;
        double pos = starting;
        double step = Double.MAX_VALUE;
        int iter = 0; 
        while ((Math.abs(step) > eps) && (iter < MAXITER))
        {
            double val = 0, der = 0; 
            for (int i=0; i<data.length; i++)
                for (int j=0; j<data[i].length; j++)
                {
                    double sinv = Math.sin(j*pos+c[i]);
                    double cosv = Math.cos(j*pos+c[i]);
                    val += j*cosv*(data[i][j] - sinv);
                    der += j*j*(sinv*(sinv - data[i][j]) - cosv*cosv);
                }
            step = -val / der;
            pos += (iter < 3?0.8:1)*step;
            while (pos > Math.PI) pos -= 2*Math.PI;
            if (pos<0) pos = -pos;
            iter++;
        }
        return pos;
    }
    
    public static double squaredDistance(double[] data, double f, double c)
    {
        double erg = 0;
        for (int i=0; i<data.length; i++) {double t = data[i]-Math.sin(f*i+c); erg += t*t;}
        return erg;
    }
    
    public static double optimizeCAlternative(double[] data, double f, double eps)
    {
        double b = Math.asin(data[0]); if (Double.isNaN(b)) if (data[0]<-1) b = -Math.PI/2; else b = Math.PI/2;
        double bval = squaredDistance(data, f,b);
        double balt = (b>0?Math.PI - b:-Math.PI - b), baltval = squaredDistance(data, f, balt);
        if (baltval  < bval) {bval = baltval; b = balt;} 
        double a = b - 0.3, aval = squaredDistance(data,f,a);
        double c = b + 0.3, cval = squaredDistance(data,f,c);
        while (aval < bval) {c = b; cval = bval; b = a; bval = aval; a -= 0.2; aval = squaredDistance(data, f, a);}
        while (cval < bval) {a = b; aval = bval; b = c; bval = cval; c += 0.2; cval = squaredDistance(data, f, c);}
        if ((aval < bval) || (cval < bval)) 
            System.out.println("Problem in Alternative Optimization of phase, initial value not minimal.");
        while (c - a > eps)
        {
            double t1 = (a+b)/2, t1v = squaredDistance(data, f, t1);
            double t2 = (b+c)/2, t2v = squaredDistance(data, f, t2);
            if ((t1v>bval) && (t2v>bval)) {a=t1; aval = t1v; c=t2; cval = t2v;}
            else if (t1v < t2v) {c=b; cval = bval; b=t1; bval = t1v;}
            else {a=b; aval = bval; b=t2; bval = t2v;}
        }
        while (b < -Math.PI) b += 2*Math.PI;
        while (b > Math.PI)  b -= 2*Math.PI;

        balt = (b>0?Math.PI - b - (data.length-1)*f:-Math.PI-b - (data.length-1)*f);
        baltval = squaredDistance(data, f, balt);
        if (baltval < bval) return balt;
        return b;
    }
    
    public static double optimizeC(double[] data, double f, double starting, double eps)
    {
        final int MAXITER = 30;
        double pos = starting;
        double step = Double.MAX_VALUE;
        int iter = 0;
        while ((Math.abs(step) > eps) && (iter < MAXITER))
        {
            double val = 0, der = 0;
            for (int i=0; i<data.length; i++)
            {
                double sinv = Math.sin(i*f+pos);
                double cosv = Math.cos(i*f+pos);
                val += cosv * (data[i] - sinv);
                der += sinv * (sinv - data[i]) - cosv*cosv;
            }
            step = -val / der;
            while (step > Math.PI) step -= 2*Math.PI;
            while (step < -Math.PI) step += 2*Math.PI;
            pos += (iter < 3?0.8:1)*step;
            while (pos > Math.PI) pos -= 2*Math.PI;
            while (pos < -Math.PI) pos += 2*Math.PI;
            iter++;
            
            // check for the local minimum at Pi - pos
            if (Math.abs(step) <= eps)
            {
                double vHere = 0, vThere = 0; 
                double tPos = (pos>0?Math.PI - pos - (data.length-1)*f:-Math.PI-pos - (data.length-1)*f); 
                for (int i=0; i<data.length; i++) 
                {
                    double sH = Math.sin(i*f+pos), sT = Math.sin(i*f+tPos); 
                    vHere += (sH-data[i])*(sH-data[i]); vThere += (sT-data[i])*(sT-data[i]);
                }
                if (vThere < vHere) {pos = tPos; step = Double.MAX_VALUE;}
            }
        }
        if (iter>=MAXITER) return optimizeCAlternative(data, f, eps);
        return pos;
    }
    
    public static void optimizeFAndC(double[][] data, double[] cerg, double[] ferg, double eps)
    {
        FAILEDFLAG = false;
        final int MAXITER = 50;
//        final double cuberooteps = Math.pow(eps, 1.0/3.0);
        final double squarerooteps = Math.pow(eps, 1.0/2.0);
        double FStep = Double.MAX_VALUE;
        int iter = 0;
        while ((Math.abs(FStep) > eps) && (iter < MAXITER))
        {
            for (int i=0; i<data.length; i++)
            {
                double nc = optimizeC(data[i], ferg[0], cerg[i], squarerooteps); cerg[i] = nc;
//                double nc = optimizeCAlternative(data[i], ferg[0], squarerooteps); cerg[i] = nc;
            }
            double nf = optimizeF(data, cerg, ferg[0], eps); FStep = ferg[0] - nf; ferg[0] = nf;
            iter++; 
        }
        if (iter >= MAXITER) FAILEDFLAG = true;
    }
    
    public static double[] createRawData(int c, double f, int length)
    {
        double[] data = new double[length];
        for (int i=0; i<length; i++) data[i] = Math.sin(i*f + c);
        return data;
    }
    
    public static double[][] createChopsRandom(double f, int anz, int length, double errStdv) {double[][] erg = new double[anz][length]; createChopsRandom(f, anz, length, errStdv, erg); return erg;}
    public static double[][] createChopsRandom(double f, int anz, int length, double errStdv, double[][] erg)
    {
        for (int i=0; i<anz; i++)
        {
            double c = random.nextDouble() * 2 * Math.PI;
            for (int j=0; j<length; j++) erg[i][j] = Math.sin(c + j*f) + random.nextGaussian()*errStdv;
        }
        return erg;
    }
    
    public static double[][] createChopsSubsequently(double f, int anz, int length, double errStdv) {double[][] erg = new double[anz][length]; createChopsSubsequently(f, anz, length, errStdv, erg); return erg;}
    public static double[][] createChopsSubsequently(double f, int anz, int length, double errStdv, double[][] erg)
    {
        double c = random.nextDouble() * 2 * Math.PI;
        for (int j=0; j<length; j++) erg[0][j] = Math.sin(c + j*f) + random.nextGaussian()*errStdv;
        for (int i=1; i<anz; i++)
        {
            for (int j=0; j<length-1; j++) erg[i][j] = erg[i-1][j+1];
            erg[i][length-1] = Math.sin(c + (i+length-1)*f) + random.nextGaussian()*errStdv;
        }
        return erg;
    }
    
    public static double computeFisherSTDVOfFrequencyParameter(double f, int anz, int length, double errStdv, boolean timeDelayEmbedding)
    {
        f *= 2*Math.PI;
        double[] c = new double[anz];
        double[][] data = new double[anz][anz];
        c[0] = random.nextDouble() * 2 * Math.PI;
        for (int j=0; j<length; j++) data[0][j] = Math.sin(c[0] + j*f) + random.nextGaussian()*errStdv;
        for (int i=1; i<anz; i++)
        {
            if (timeDelayEmbedding) {
                for (int j=0; j<length-1; j++) data[i][j] = data[i-1][j+1];
                data[i][length-1] = Math.sin(c[0] + (i+length-1)*f) + random.nextGaussian()*errStdv;
                c[i] = c[i-1] + f;
            } else {
                c[i] = random.nextDouble() * 2 * Math.PI;
                for (int j=0; j<length; j++) data[i][j] = Math.sin(c[i] + j*f) + random.nextGaussian()*errStdv;
            }
        }
        return estimatedSDOfFrequencyEstimate(data, c, f, errStdv*errStdv) / (2*Math.PI); 
    }

    public static double[][] selectChopsRandom(double[] data, int anz, int length) {double[][] erg = new double[anz][length]; selectChopsRandom(data, anz, length, erg); return erg;}
    public static double[][] selectChopsRandom(double[] data, int anz, int length, double[][] erg)
    {
        if (length >= data.length) throw new RuntimeException("Cannot pick chops of length "+length+" from a data stream of length "+data.length);
        for (int i=0; i<anz; i++)
        {
            int start = random.nextInt(data.length - length);
            for (int j=0; j<length; j++) erg[i][j] = data[start+j];
        }
        return erg;
    }
    
    public static double[][] selectChopsSubsequently(double[] data, int anz, int length) {double[][] erg = new double[anz][length]; selectChopsSubsequently(data, anz, length, erg); return erg;}
    public static double[][] selectChopsSubsequently(double[] data, int anz, int length, double[][] erg)
    {
        if (anz+length >= data.length) throw new RuntimeException("Cannot pick a chop of length "+(anz+length)+" from a data stream of length "+data.length);
        int start = random.nextInt(data.length - length - anz);
        for (int i=0; i<anz; i++)
        {
            for (int j=0; j<length; j++) erg[i][j] = data[start+i+j];
        }
        return erg;
    }
    
    /**
     * 
     * @param frequency
     * @param choplength
     * @param chopanz
     * @param errStdv
     * @param trials
     * @param eps
     * @return a 8-tuple of: mean Frequency, STDV frequency, outlayers in frequency with random bursts, 
     *                       mean Frequency, STDV frequency, outlayers in frequency with time delay embedding,
     *                       power with random bursts,
     *                       power with time delay embedding.
     *                       
     */
    public static double[] compareChoppings(double frequency, int choplength, int chopanz, double errStdv, int trials, double eps)
    { return compareChoppings(frequency, 0.0, choplength, chopanz, errStdv, trials, eps);}
    public static double[] compareChoppings(double frequency, double freqHyp, int choplength, int chopanz, double errStdv, int trials, double eps)
    {
        if (freqHyp==0) freqHyp = eps; freqHyp *= 2*Math.PI;
        double[] erg = new double[8];
        double freq = 2*Math.PI*frequency;
        double[] f = new double[]{freq}, c = new double[chopanz], chyp = new double[chopanz];
        double[][] data = new double[chopanz][choplength];
        double mean = 0, var = 0, lokMean = 0, lokVar = 0, stdvMean = 0, stdvVar = 0; int lokLen = 30;
        int outlayer = 0; int valid = 0; double averageNoncentrality = 0.0;
        int reject = 0;
        boolean stopsign = false;
        for (int i=0; (i<trials) && (!stopsign); i++)
        {
            f[0] = freq; for (int j=0; j<chopanz; j++) c[j] = 0;
            createChopsRandom(freq, chopanz, choplength, errStdv, data);
            optimizeFAndC(data, c, f, eps);
            if ((FAILEDFLAG) || (Math.abs(f[0]-freq)>0.5)) outlayer++;
            else 
            {
//                lokMean += Math.abs(f[0]-freq); 
                lokMean += f[0]; 
                lokVar += f[0]*f[0];
                for (int j=0; j<data.length; j++) chyp[j] = optimizeC(data[j], freqHyp, c[j], Math.pow(eps,1.0/2.0));
                double m2diff = minus2llDifferenceShort(data, c, f[0], freqHyp, chyp, errStdv); valid++;
                averageNoncentrality += m2diff;
                if (m2diff > Statik.FIVEPERCENTTHRESHOLD[0]) reject++;
//                System.out.println(i+"\t"+m2diff+"\t"+f[0]);
            }
            if (valid % lokLen == 0) 
            {
                mean += lokMean; var += lokVar; 
                lokMean /= lokLen; lokVar /= lokLen; stdvMean += Math.sqrt(lokVar - lokMean*lokMean); stdvVar += lokVar - lokMean*lokMean;
                lokMean = 0; lokVar = 0;
                double stdv = Math.sqrt( (var - mean*mean/valid)/valid);
                int anzLok = valid / lokLen;
                double stdvStdv = Math.sqrt((stdvVar - stdvMean*stdvMean/anzLok)/(double)(anzLok*valid));
                if ((stdvStdv / stdv < 0.01) && (i>5*lokLen)) {stopsign = true; System.out.println("stopping at trial "+i);} 
            }
        }
//      erg[0] = ((mean / valid) - freq)/(2*Math.PI); erg[1] = Math.sqrt( (var - mean*mean/valid)/valid)/(2*Math.PI); erg[2] = outlayer; 
        erg[0] = freq/(2*Math.PI) + ((mean / valid) - freq)/(2*Math.PI) * (valid/200.0); erg[1] = Math.sqrt( (var - mean*mean/valid)/valid)/(2*Math.PI); erg[2] = outlayer; 
//        erg[0] = (mean / valid)/(2*Math.PI); erg[1] = Math.sqrt( (var - mean*mean/valid)/valid)/(2*Math.PI); erg[2] = outlayer; 
        erg[6] = (double)reject/(double)valid;
        mean = 0; var = 0; lokMean = 0; lokVar = 0; stdvMean = 0; stdvVar = 0;
        outlayer = 0; valid = 0; averageNoncentrality = 0.0; reject = 0;
        stopsign = false;
        for (int i=0; (i<trials) && (!stopsign); i++)
        {
            f[0] = freq; for (int j=0; j<chopanz; j++) c[j] = 0;
            createChopsSubsequently(freq, chopanz, choplength, errStdv, data);
            optimizeFAndC(data, c, f, eps);
            if ((FAILEDFLAG) || (Math.abs(f[0]-freq)>0.5)) outlayer++;
            else 
            {
//                lokMean += Math.abs(f[0]-freq); 
              lokMean += f[0]; 
                lokVar += f[0]*f[0]; 
                for (int j=0; j<data.length; j++) chyp[j] = optimizeC(data[j], freqHyp, c[j], Math.pow(eps,1.0/2.0));
                double m2diff = minus2llDifferenceShort(data, c, f[0], freqHyp, chyp, errStdv); valid++;
                if (m2diff > Statik.FIVEPERCENTTHRESHOLD[0]) reject++;
                averageNoncentrality += m2diff;
//                System.out.println(i+"\t"+m2diff+"\t"+f[0]);
            }
            if (valid % lokLen == 0) 
            {
                mean += lokMean; var += lokVar; 
                lokMean /= lokLen; lokVar /= lokLen; stdvMean += Math.sqrt(lokVar - lokMean*lokMean); stdvVar += lokVar - lokMean*lokMean;
                lokMean = 0; lokVar = 0;
                double stdv = Math.sqrt( (var - mean*mean/valid)/valid); 
                int anzLok = valid / lokLen;
                double stdvStdv = Math.sqrt((stdvVar - stdvMean*stdvMean/anzLok)/(double)(anzLok*valid));
                if ((stdvStdv / stdv < 0.01) && (i>5*lokLen)) {stopsign = true; System.out.println("stopping at trial "+i);} 
            }
        }
//        erg[3] = (mean / valid)/(2*Math.PI); erg[4] = Math.sqrt( (var - mean*mean/valid)/valid)/(2*Math.PI); erg[5] = outlayer;
//      erg[3] = ((mean / valid) - freq)/(2*Math.PI); erg[4] = Math.sqrt( (var - mean*mean/valid)/valid)/(2*Math.PI); erg[5] = outlayer;
        erg[3] = freq/(2*Math.PI) + ((mean / valid) - freq)/(2*Math.PI) * (valid / 200.0); erg[4] = Math.sqrt( (var - mean*mean/valid)/valid)/(2*Math.PI); erg[5] = outlayer;
        erg[7] = (double)reject/(double)valid;
        
        return erg;
    }
    
    public static void someCrossings(int trials)
    {
        double[][] out1 = new double[11][11], out2 = new double[11][11];
        for (int a=0; a<=10; a++)
        {
            for (int b=0; b<=10; b++)
            {
                System.out.println("a = "+a+", b = "+b);
                int chunklength = 2*a+2;
                int chunkanz = 10*b+10;
                double[] zwerg = compareChoppings(0.05, chunklength, chunkanz, 0.3, trials, 0.00001);
                out1[a][b] = (Math.log(zwerg[1])-Math.log(zwerg[4]))/Math.log(2);
                if (Math.abs(zwerg[1]-zwerg[4]) < 0.0001) System.out.println("Warning: Epsilon less than ten times smaller than difference = "+Math.abs(zwerg[1]-zwerg[4]));
                if (Math.abs(zwerg[1]-zwerg[4]) < 0.00001) System.out.println("Warning: Epsilon too big");
                double freq = 0.01 + 0.19*a/10;
                double err = 0.1*b;
                zwerg = compareChoppings(freq, 10, 100, err, trials, 0.00001);
                out2[a][b] = (Math.log(zwerg[1])-Math.log(zwerg[4]))/Math.log(2);
                if (Math.abs(zwerg[1]-zwerg[4]) < 0.0001) System.out.println("Warning: Epsilon less than ten times smaller than difference = "+Math.abs(zwerg[1]-zwerg[4]));
                if (Math.abs(zwerg[1]-zwerg[4]) < 0.00001) System.out.println("Warning: Epsilon too big");
            }
            Statik.writeMatrix(out1, "chunkSizeVsLength.txt", '\t');
            Statik.writeMatrix(out2, "FrequencyVsError.txt", '\t');
        }
    }
    
    public static void simulateStevesData(double errStdv, int N)
    {
        double eps = 0.00001;
        int tau = 4;
        int K = 5, trials = 400;
        double[][] org = Statik.loadMatrix("SingleOscMultiFreqMultiDamp1Cut.dat", ' ', true);
        double[][] data = new double[N][K];
        double[] f = new double[]{0.01}, c = new double[N];
        int outlayer1 = 0, outlayer2 = 0; double mean1 = 0.0, var1 = 0.0, mean2 = 0.0, var2 = 0.0;
        for (int trial = 0; trial<trials; trial++)
        {
//            if (trial % 10 == 0) System.out.println("Trial = "+trial+", Mean(seq) = "+mean1/(trial-outlayer1)+", STDV(seq) = "+
//                    Math.sqrt(var1 / (trial-outlayer1) - mean1*mean1/((trial-outlayer1)*(trial-outlayer1)))+", Mean(ran) = "+mean2/(trial-outlayer2)+", STDV(ran) = "+
//                    Math.sqrt(var2 / (trial-outlayer2) - mean2*mean2/((trial-outlayer2)*(trial-outlayer2))));
            f[0] = 0.06; for (int i=0; i<c.length; i++) c[i] = 0.0;
            int start = 2+random.nextInt(org[0].length-2 - N - tau*K);
            for (int i=0; i<tau; i++) 
                for (int j=0; j<K; j++) data[i][j] = org[0][start+ i + tau*j]/10.0 + random.nextGaussian()*errStdv;
            for (int i=tau; i<N; i++) {
                for (int j=0; j<K-1; j++) data[i][j] = data[i-tau][j+1];
                data[i][K-1] = org[0][start + i + tau*(K-1)]/10.0 + random.nextGaussian()*errStdv;
            }
            optimizeFAndC(data, c, f, eps);
            if ((FAILEDFLAG) || (Math.abs(f[0]-0.06)>0.2)) outlayer1++;
            else 
            {
                mean1 += f[0]; 
                var1 += f[0]*f[0]; 
            }
            
            for (int i=0; i<N; i++)
            {
                start = 2+random.nextInt(org[0].length-2 - tau*K);
                for (int j=0; j<K; j++) 
                    data[i][j] = org[0][start + tau*j]/10.0 + random.nextGaussian()*errStdv;
            }
            optimizeFAndC(data, c, f, eps);
            if ((FAILEDFLAG) || (Math.abs(f[0]-0.06)>0.2)) outlayer2++;
            else 
            {
                mean2 += f[0]; 
                var2 += f[0]*f[0]; 
            }
            
        }
        mean1 /= (trials - outlayer1);
        var1 /= (trials - outlayer1); var1 -=mean1*mean1;
        mean2 /= (trials - outlayer2);
        var2 /= (trials - outlayer2); var2 -=mean2*mean2;
        System.out.println(mean1+"\t"+Math.sqrt(var1)+"\t"+mean2+"\t"+Math.sqrt(var2));
//        System.out.println("Overall "+(org[0].length-2)+" data points, "+outlayer1+" outlayer(seq), "+outlayer2+" outlayer(ran)");
    }

    public static void main(String[] args)
    {
        random = new Random();
//        System.out.println("Mean sequential   \tSTDV sequential  \tmean random        \tSTDV random");
//        simulateStevesData(0.01, 40);
//        simulateStevesData(0.01, 200);
//        simulateStevesData(0.01, 1000);
//        simulateStevesData(0.05, 40);
//        simulateStevesData(0.05, 200);
//        simulateStevesData(0.05, 1000);
        
        
        double freq = 0.01, hfreq = 0.01;
        for (int i=0; i<10; i++) System.out.println(computeFisherSTDVOfFrequencyParameter(freq, 40, 5, 0.01, false));        
        double[] zwerg = compareChoppings(freq, hfreq, 5, 40, 0.01, 1000, 0.0000001);
        for (int i=0; i<zwerg.length; i++) System.out.println(zwerg[i]+"\t");
//        zwerg = compareChoppings(freq, hfreq, 5, 200, 0.01, 1000, 0.0000001);
//        for (int i=0; i<zwerg.length; i++) System.out.println(zwerg[i]+"\t");
//        zwerg = compareChoppings(freq, hfreq, 5, 1000, 0.01, 1000, 0.0000001);
//        for (int i=0; i<zwerg.length; i++) System.out.println(zwerg[i]+"\t");
//        zwerg = compareChoppings(freq, hfreq, 5, 40, 0.05, 1000, 0.0000001);
//        for (int i=0; i<zwerg.length; i++) System.out.println(zwerg[i]+"\t");
//        zwerg = compareChoppings(freq, hfreq, 5, 200, 0.05, 1000, 0.0000001);
//        for (int i=0; i<zwerg.length; i++) System.out.println(zwerg[i]+"\t");
//        zwerg = compareChoppings(freq, hfreq, 5, 1000, 0.05, 1000, 0.0000001);
//        for (int i=0; i<zwerg.length; i++) System.out.println(zwerg[i]+"\t");
//        
//        double f = 0.06; int count = 1; double erg = 0;
//        for (double phase = -2*f; phase < Math.PI/2.0 - 2*f; phase += 0.01)
//        {
//            count++;
//            double mean = 0, var = 0;
//            for (int i=0; i<5; i++) 
//            {
//                double dat = Math.sin(phase + i*f);
//                mean += dat; var += dat*dat;
//            }
//            mean /= 5.0; var = (var/5.0) - mean*mean;
//            System.out.println(count+"\t"+phase+"\t"+var);
//            erg += var;
//        }
//        System.out.println(erg / (double)count);
        
        System.exit(0);

        //        someCrossings((args.length>0?Integer.parseInt(args[0]):1000));
        /*
        double[][] out = new double[89][4];
        for (int a=0; a<=88; a++)
        {
            System.out.println("a = "+a);
            int chunklength = a+2;
            int chunkanz = 80;
            double[] zwerg = compareChoppings(0.05, 0.045, chunklength, chunkanz, 1.0, 1000, 0.000001);
            out[a][0] = zwerg[1]; out[a][1] = zwerg[4]; out[a][2] = zwerg[6]; out[a][3] = zwerg[7];
            if (Math.abs(zwerg[1]-zwerg[4]) < 0.00001) System.out.println("Warning: Epsilon less than ten times smaller than difference = "+Math.abs(zwerg[1]-zwerg[4]));
            if (Math.abs(zwerg[1]-zwerg[4]) < 0.000001) System.out.println("Warning: Epsilon too big");
            Statik.writeMatrix(out, "chunkSizeOnly.txt", '\t');
        }
        */
        // real Frequency, choplength, chopnumber, error, trials
//        double[] res = compareChoppings(0.01, 10, 100, 0.05, 500, 0.00001);
//        for (int i=0; i<6; i++) System.out.println(res[i]);
    }
}
